#ifndef DUNE_MICROSTRUCTURE_EFFECTIVEQUANTITIESCOMPUTER_HH
#define DUNE_MICROSTRUCTURE_EFFECTIVEQUANTITIESCOMPUTER_HH

#include <dune/common/parametertree.hh>
#include <dune/istl/eigenvalue/test/matrixinfo.hh> // TEST: compute condition Number 
#include <dune/istl/io.hh>
#include <dune/istl/matrix.hh>
#include <dune/microstructure/matrix_operations.hh>
#include <dune/microstructure/CorrectorComputer.hh>

using namespace MatrixOperations;
using std::shared_ptr;
using std::make_shared;
using std::string;
using std::cout;
using std::endl;


template <class Basis, class Material>
class EffectiveQuantitiesComputer {

public:
	static const int dimworld = 3;
	static const int dim = Basis::GridView::dimension;
	using Domain = typename CorrectorComputer<Basis,Material>::Domain; 
	using VectorRT = typename CorrectorComputer<Basis,Material>::VectorRT;
	using MatrixRT = typename CorrectorComputer<Basis,Material>::MatrixRT;
	using Func2Tensor = typename CorrectorComputer<Basis,Material>::Func2Tensor;
	using FuncVector = typename CorrectorComputer<Basis,Material>::FuncVector;
	using VectorCT = typename CorrectorComputer<Basis,Material>::VectorCT;
    using Func2int = std::function< int(const Domain&) >;
    using MatrixPhase = std::function< MatrixRT(const int&) >;

protected:
	// CorrectorComputer<Basis,Material>& correctorComputer_; 
    std::shared_ptr<CorrectorComputer<Basis,Material>> correctorComputer_; 
    // const Material& material_;  //Get this from correctorComputer_ this is now an std::shared_ptr.

public:

	MatrixRT Qeff_;	    // Effective moduli COEFFICIENTS
	VectorRT Bhat_;		
	VectorRT Beff_;	    // Effective prestrain COEFFICIENTS


	///////////////////////////////
	// constructor
	///////////////////////////////
	// EffectiveQuantitiesComputer(CorrectorComputer<Basis,Material>& correctorComputer, 
    //                             const Material& material)
    //                             : correctorComputer_(correctorComputer), 
    //                             material_(material)
    // {} 
	// EffectiveQuantitiesComputer(CorrectorComputer<Basis,Material>& correctorComputer)
    //                             : correctorComputer_(correctorComputer)
    // {} 
	EffectiveQuantitiesComputer(std::shared_ptr<CorrectorComputer<Basis,Material>> correctorComputer)
                                : correctorComputer_(correctorComputer)
    {} 

    ///////////////////////////////
    // Getter
    ///////////////////////////////
	// CorrectorComputer<Basis,Material> getCorrectorComputer(){return correctorComputer_;}
    CorrectorComputer<Basis,Material> getCorrectorComputer(){return *correctorComputer_;}


	const shared_ptr<Basis> getBasis() {return *correctorComputer_.getBasis();}

    auto getQeff(){return Qeff_;}
    auto getBeff(){return Beff_;}


    void updateCorrectorComputer(std::shared_ptr<CorrectorComputer<Basis,Material>> correctorComputer)
    {
        correctorComputer_ = correctorComputer;
    }

  // -----------------------------------------------------------------
  // --- Compute Effective Quantities
    void computeEffectiveQuantities()
    {
        Dune::Timer effectiveQuantitiesTimer;
        Dune::ParameterTree parameterSet = correctorComputer_->getParameterSet();

        if (parameterSet.get<bool>("printMicroOutput ", false))
            std::cout << "starting effective quantities computation..." << std::endl;
        

        auto MContainer = correctorComputer_->getMcontainer();
        auto MatrixBasisContainer = correctorComputer_->getMatrixBasiscontainer();
        auto x3MatrixBasisContainer = correctorComputer_->getx3MatrixBasiscontainer();

        auto gamma = correctorComputer_->getGamma();
        auto basis = *correctorComputer_->getBasis();
        

		shared_ptr<VectorCT> phiBasis[3] = {correctorComputer_->getCorr_phi1(), 
                                            correctorComputer_->getCorr_phi2(),
                                            correctorComputer_->getCorr_phi3()
										    };

        auto localIndicatorFunction = (correctorComputer_->material_)->getLocalIndicatorFunction();
        
        Qeff_ = 0 ;
        Bhat_ = 0;
    
        for(size_t a=0; a < 3; a++)
        for(size_t b=0; b < 3; b++)
        {
            double energy = 0.0;
            double prestrain = 0.0;
            auto localView = basis.localView();

            auto GVFunc_a = derivative(Dune::Functions::makeDiscreteGlobalBasisFunction<VectorRT>(basis,*phiBasis[a]));
            auto localfun_a = localFunction(GVFunc_a);

            auto matrixFieldG1GVF  = Dune::Functions::makeGridViewFunction(x3MatrixBasisContainer[a], basis.gridView());
            auto matrixFieldG1 = localFunction(matrixFieldG1GVF);
            auto matrixFieldG2GVF  = Dune::Functions::makeGridViewFunction(x3MatrixBasisContainer[b], basis.gridView());
            auto matrixFieldG2 = localFunction(matrixFieldG2GVF);

            for (const auto& element : elements(basis.gridView()))
            {
                localView.bind(element);
                matrixFieldG1.bind(element);
                matrixFieldG2.bind(element);
                localfun_a.bind(element);

                localIndicatorFunction.bind(element);

                double elementEnergy = 0.0;
                double elementPrestrain = 0.0;

                auto geometry = element.geometry();
                const auto& localFiniteElement = localView.tree().child(0).finiteElement();

                int orderQR = 2*(dim*localFiniteElement.localBasis().order()-1);

                const auto& quad = Dune::QuadratureRules<double, dim>::rule(element.type(), orderQR);

                for (const auto& quadPoint : quad) 
                {
                    const auto& quadPos = quadPoint.position();
                    const double integrationElement = geometry.integrationElement(quadPos);
                    
                    auto Chi1 = sym(crossSectionDirectionScaling(1.0/gamma, localfun_a(quadPos))) + MContainer[a];
                    
                    auto G1 = matrixFieldG1(quadPos);
                    auto G2 = matrixFieldG2(quadPos);
                    
                    auto X1 = matrixToVoigt(G1 + Chi1);
                    
                    double energyDensity= voigtScalarProduct((correctorComputer_->material_)->applyElasticityTensor(X1,localIndicatorFunction(quadPos)), matrixToVoigt(sym(G2)));

                    elementEnergy += energyDensity * quadPoint.weight() * integrationElement;      // quad[quadPoint].weight() ???
                    if (b==0)
                    {
                        auto prestrainPhaseValue = (correctorComputer_->material_)->prestrain(localIndicatorFunction(quadPos),element.geometry().global(quadPos));
                        auto value = voigtScalarProduct((correctorComputer_->material_)->applyElasticityTensor(X1,localIndicatorFunction(quadPos)),matrixToVoigt(sym(prestrainPhaseValue)));

                        elementPrestrain += value * quadPoint.weight() * integrationElement;
                    }
                }
                energy += elementEnergy;
                prestrain += elementPrestrain;
            
            }
            Qeff_[a][b] = energy;    
            if (b==0)
                Bhat_[a] = prestrain;
        }
        if(parameterSet.get<bool>("print_debug", false))
        {
            printmatrix(std::cout, Qeff_, "Qeff_", "--");
            printvector(std::cout, Bhat_, "Bhat_", "--");
        }

        ///////////////////////////////
        // Compute effective Prestrain B_eff (by solving linear system)
        //////////////////////////////
         
        // std::cout << "------- Information about Q matrix -----" << std::endl;  // TODO
        // MatrixInfo<MatrixRT> matrixInfo(Qeff_,true,2,1);
        // std::cout << "----------------------------------------" << std::endl;
        Qeff_.solve(Beff_,Bhat_);
        if(parameterSet.get<bool>("print_debug", false))
            printvector(std::cout, Beff_, "Beff_", "--");
        
        //LOG-Output
        // auto& log = *(correctorComputer_.getLog());
        // log << "--- Effective moduli --- " << std::endl;
        // log << "Qeff_: \n" << Qeff_ << std::endl;
        // log << "------------------------ " << std::endl;
        // log << "--- Prestrain Output --- " << std::endl;
        // log << "Bhat_: " << Bhat_ << std::endl;
        // log << "Beff_: " << Beff_ <<  " (Effective Prestrain)" << std::endl;
        // log << "------------------------ " << std::endl;

        //   std::cout << std::setprecision(std::numeric_limits<float_50>::digits10) << higherPrecEnergy << std::endl;
        if (parameterSet.get<bool>("printMicroOutput ", false))
            std::cout << "Time required to solve for effective quantities: " << effectiveQuantitiesTimer.elapsed() << std::endl;
        return ;
    }


  // -----------------------------------------------------------------
  // --- write Data to .Txt-File for Matlab / Optimization-Code
    void writeEffectiveQuantitiesToTxt(std::string outputPath)
    {
        std::cout << "write effective quantities as .txt files to output folder..." << std::endl;
        writeMatrixToMatlab(Qeff_, outputPath + "/QMatrix.txt");
        // write effective Prestrain in Matrix for Output
        Dune::FieldMatrix<double,1,3> BeffMat;
        BeffMat[0] = Beff_;
        writeMatrixToMatlab(BeffMat, outputPath + "/BMatrix.txt");
        return;
    }

    template<class MatrixFunction>
    double energySP(const MatrixFunction& matrixFieldFuncA,
                    const MatrixFunction& matrixFieldFuncB)
    {
        double energy = 0.0;
        auto mu_ = *correctorComputer_->getMu();
        auto lambda_ = *correctorComputer_->getLambda();
        auto gamma = correctorComputer_->getGamma();
        auto basis = *correctorComputer_->getBasis();
        auto localView = basis.localView();

        auto localIndicatorFunction = (correctorComputer_->material_)->getLocalIndicatorFunction();

        auto matrixFieldAGVF  = Dune::Functions::makeGridViewFunction(matrixFieldFuncA, basis.gridView());
        auto matrixFieldA = localFunction(matrixFieldAGVF);
        auto matrixFieldBGVF  = Dune::Functions::makeGridViewFunction(matrixFieldFuncB, basis.gridView());
        auto matrixFieldB = localFunction(matrixFieldBGVF);

        for (const auto& e : elements(basis.gridView()))
        {
            localView.bind(e);
            matrixFieldA.bind(e);
            matrixFieldB.bind(e);
            localIndicatorFunction.bind(e);

            double elementEnergy = 0.0;

            auto geometry = e.geometry();
            const auto& localFiniteElement = localView.tree().child(0).finiteElement();

            int orderQR = 2*(dim*localFiniteElement.localBasis().order()-1);
            const auto& quad = Dune::QuadratureRules<double, dim>::rule(e.type(), orderQR);
            for (const auto& quadPoint : quad) 
            {
                const auto& quadPos = quadPoint.position();
                const double integrationElement = geometry.integrationElement(quadPos);

                double energyDensity= scalarProduct((correctorComputer_->material_)->applyElasticityTensor(matrixFieldA(quadPos),localIndicatorFunction(quadPos)),sym(matrixFieldB(quadPos)));

                elementEnergy += energyDensity * quadPoint.weight() * integrationElement;          
            }
            energy += elementEnergy;
        }
        return energy;
    }

}; // end class

#endif
