#ifndef DUNE_MICROSTRUCTURE_ENERGIES_DISCRETEKIRCHHOFFBENDINGENERGYPRESTRAINED_HH
#define DUNE_MICROSTRUCTURE_ENERGIES_DISCRETEKIRCHHOFFBENDINGENERGYPRESTRAINED_HH

#include <dune/common/fmatrix.hh>
#include <dune/geometry/quadraturerules.hh>

#include <dune/grid/io/file/vtk/subsamplingvtkwriter.hh>

#include <dune/gfe/assemblers/localenergy.hh>
#include <dune/gfe/bendingisometryhelper.hh>
#include <dune/gfe/tensor3.hh>

#include <dune/localfunctions/lagrange/lagrangesimplex.hh>


#include <dune/microstructure/microproblem.hh>

#include <cmath>
#include <string> 
#include <map>


  // template <class T>
  // constexpr std::string_view type_name()
  // {
  //     using namespace std;
  // #ifdef __clang__
  //     string_view p = __PRETTY_FUNCTION__;
  //     return string_view(p.data() + 34, p.size() - 34 - 1);
  // #elif defined(__GNUC__)
  //     string_view p = __PRETTY_FUNCTION__;
  // #  if __cplusplus < 201402
  //     return string_view(p.data() + 36, p.size() - 36 - 1);
  // #  else
  //     return string_view(p.data() + 49, p.find(';', 49) - 49);
  // #  endif
  // #elif defined(_MSC_VER)
  //     string_view p = __FUNCSIG__;
  //     return string_view(p.data() + 84, p.size() - 84 - 7);
  // #endif
  // }




/** 
 * \brief Assemble the discrete Kirchhoff bending energy for a single element. 
 * 
 *  The Kirchhoff bending energy consists of two parts: 
 * 
 *  1. The energy of  Qhom(II - Beff) where II is the second fundamental form of the surface parametrized by the deformation function 
 *     and Beff is a (given) effective prestrain tensor. 
 * 
 *     This contribution is split intro three parts which corresponds to the binomial formula (a+b)^2 = a^2 + 2ab + b^2
 *     each term is discretized separately.
 * 
 *  2. An integral over the scalar product of a forcing term and the discrete deformation 
 *  (i.e. the evaluation of localFunction_ ).
 */
namespace Dune::GFE
{
  template<class Basis, class LocalDiscreteKirchhoffFunction, class LocalForce, class TargetSpace>
  class DiscreteKirchhoffBendingEnergyPrestrained
    : public Dune::GFE::LocalEnergy<Basis,TargetSpace>
  {
    // grid types
    typedef typename Basis::GridView GridView;
    typedef typename GridView::ctype DT;
    typedef typename TargetSpace::ctype RT;
    typedef typename LocalDiscreteKirchhoffFunction::DerivativeType DerivativeType;
    typedef typename Dune::FieldMatrix<double,2,2> PrestrainType;
    typedef typename Dune::FieldMatrix<double,3,3> MatrixType;
    // Grid dimension
    constexpr static int gridDim = GridView::dimension;
    // P2-LocalFiniteElement used to represent the discrete Jacobian on the current element.
    typedef typename Dune::LagrangeSimplexLocalFiniteElement<DT, double, gridDim, 2> P2LocalFiniteElement;

    using Domain = typename Basis::GridView ::template Codim<0>::Geometry::GlobalCoordinate;

    using IndicatorType = std::function<int(const Domain&)>;
    using IndicatorGridViewFunctionType = GridViewFunction<int(const Domain&), GridView>;
    using LocalIndicatorFunctionType = LocalFunction<int(const Domain&), typename GridViewEntitySet<GridView, 0>::Element >;

    LocalDiscreteKirchhoffFunction& localFunction_;

    LocalForce& localForce_;

    P2LocalFiniteElement lagrangeLFE_;

    // mutable MatrixType Qhom_;
    // mutable PrestrainType Beff_;
    mutable std::vector<MatrixType> Qhom_;
    mutable std::vector<PrestrainType> Beff_;

    const Dune::ParameterTree parameterSet_;
    const Python::Module module_;


    const Python::Callable globalMicrostructureClass_;  //we use this class to create microstructure objects passed to microproblem.
    Python::Reference globalMicrostructure_;

    // Python::Callable LocalMicrostructureClass_;
    std::vector<Python::Callable> LocalMicrostructureClass_;
    // const Python::Callable microstructureClass_;  //we use this class to create microstructure objects passed to microproblem.
    // Dune::ParameterTree parameterSet_;
    // Python::Module module_;
    // Python::Callable microstructureClass_;  //we use this class to create microstructure objects passed to microproblem.

    // MicroProblem microProblem_; 

    // std::shared_ptr<Python::Callable> anotherMicroProblem_;  
    // std::shared_ptr<MicroProblem> microProblem_;  

    // std::vector<std::shared_ptr<MicroProblem>> microProblem_;  
    std::map<int,std::shared_ptr<MicroProblem> > microProblem_;  


    IndicatorType macroPhaseIndicator_;
    mutable LocalIndicatorFunctionType macroPhaseIndicatorLocalFunction_;

    int numberOfMacroPhases_; 


    std::vector<bool> constantMicrostructureVector_;


    /**
     * @brief This is just used to initialize the MicrostructureClass with some macroscopic point 
     *        in the case of a macroscopically varying microstructure.
     */
    // static constexpr Domain initialMacroPoint{0.0};
    // const Dune::FieldVector<double,gridDim> initialMacroPoint = FieldVector<double,gridDim>(0);






  public:
    DiscreteKirchhoffBendingEnergyPrestrained(LocalDiscreteKirchhoffFunction &localFunction,
                                   LocalForce &localForce,
                                   const Dune::ParameterTree& parameterSet,
                                   const Python::Module pyModule)
        : localFunction_(localFunction),
          localForce_(localForce),
          parameterSet_(parameterSet),
          module_(pyModule),
          globalMicrostructureClass_(setupGlobalMicrostructureClass())
          // microstructureClass_(setupMicrostructureClass())
          // 
          // microProblem_(setupMicroProblem())     // This is necessary since there is no default.constructor in for MicroProblem class.
    {
      // setupGlobalMicrostructureClass();
      setupEffectiveQuantities();
    }


    // const Python::Callable setupMicrostructureClass() const
    // { 
    //   std::cout << "Enter setupMicroStructureClass" <<  std::endl;
    //   Python::Callable microstructureClass;

    //   /**
    //    * @brief If we read the effective quantities from the Python module,
    //    *        we don't need the microstructureClass and use the default constructor of Python::Callable.
    //    *        Otherwise a 'class Microstructure' need to be present in the Python module.
    //    */
    //   if(!(parameterSet_.get<bool>("read_effectiveQuantities_from_Parset", 1)))
    //   {
    //     try {
    //       microstructureClass = module_.get("Microstructure");
    //     } catch (Dune::Exception &e) {
    //       std::cout << "There is no class 'Microstructure' in the Python module!" << std::endl;
    //       std::cerr << "Error during  DiscreteKirchhoffBendingEnergyPrestrained setup" << e << std::endl;
    //     }
    //   }
      
    //   return microstructureClass;
    // }

    


    
    const Python::Callable setupGlobalMicrostructureClass() const
    { 
      Python::Callable globalMicrostructureClass;

        try {
          globalMicrostructureClass = module_.get("GlobalMicrostructure");
        } catch (Dune::Exception &e) {
          std::cout << "There is no class 'GlobalMicrostructure' in the Python module!" << std::endl;
          std::cerr << "Error during  DiscreteKirchhoffBendingEnergyPrestrained setup" << e << std::endl;
        }

      // TEST
      // std::cout << "Trying to access inner class directly without creating an instance:" << std::endl;
      // Python::Callable TestClass = module_.get("LocalMicrostructure_1");

      return globalMicrostructureClass;
    };


    // const Python::Callable setupMicrostructureClass() const
    // { 
    //   // std::cout << "Enter setupMicroStructureClass" <<  std::endl;
    //   Python::Callable microstructureClass;

    //   /**
    //    * @brief If we read the effective quantities from the Python module,
    //    *        we don't need the microstructureClass and use the default constructor of Python::Callable.
    //    *        Otherwise a 'class Microstructure' need to be present in the Python module.
    //    */
    //   // if(!(parameterSet_.get<bool>("read_effectiveQuantities_from_Parset", 1)))
    //   // {
    //     try {
    //       microstructureClass = module_.get("Microstructure");
    //     } catch (Dune::Exception &e) {
    //       std::cout << "There is no class 'Microstructure' in the Python module!" << std::endl;
    //       std::cerr << "Error during  DiscreteKirchhoffBendingEnergyPrestrained setup" << e << std::endl;
    //     }
    //   // }
    //   return microstructureClass;
    // }


    // MicroProblem setupMicroProblem() const
    // {
    //     // if(parameterSet_.get<bool>("read_effectiveQuantities_from_Parset", 1))
    //     //     return MicroProblem();
    //     if(parameterSet_.get<bool>("macroscopically_varying_microstructure", 0))
    //         return MicroProblem(microstructureClass_(0),parameterSet_, module_);  //TODO Replace 0  with Domain input type. using Domain = typename GridView::template Codim<0>::Geometry::GlobalCoordinate;
    //     else 
    //         return MicroProblem(microstructureClass_(),parameterSet_, module_);   //pass a constant microstructure object
    // }


    void setupEffectiveQuantities()
    {
        /**
         * @brief If the microstructure is constant on a macroscopic level,
         *        we read the effective quantitites from the ParameterTree 
         *        and set them up once here.
         * 
         *        If no effective quanitities are provided, we solve the 
         *        micro problem once to obtain them.
         * 
         * 
         *        If the microscture is varying on a macroscopic scale,
         *        the microProblem is initialized with a default value
         *        and gets updated for each each macro point later on 
         *        in the quadrature loop.
         */





        //create instance of globalMicrostructureClass_
        // Python::Reference globalMicrostructure = globalMicrostructureClass_();
        globalMicrostructure_ = globalMicrostructureClass_();

         // = globalMicrostructure.get("macroPhases", 1)
        globalMicrostructure_.get("macroPhases").toC<int>(numberOfMacroPhases_);
        std::cout << "numberOfMacroPhases: " << numberOfMacroPhases_ << std::endl;



        // Get the macroPhaseIndicator function
        try {
          macroPhaseIndicator_ = Python::make_function<int>(Python::Callable(globalMicrostructure_.get("macroPhaseIndicator")));
          // auto macroPhaseIndicatorGVF = Dune::Functions::makeGridViewFunction(macroPhaseIndicator_, localFunction_.gridView());
          macroPhaseIndicatorLocalFunction_ = localFunction(Dune::Functions::makeGridViewFunction(macroPhaseIndicator_, localFunction_.gridView()));
        } catch (Dune::Exception &e) {
          std::cout << "Piecewise constant microstructure but no macroPhaseIndicator was provided. Assuming macroscopically constant microstructure" << std::endl;
          // Using constant True function
          macroPhaseIndicator_ = Python::make_function<int>(Python::Callable(Python::evaluate("lambda x: 1")));
        }
        //VTK-Write macroPhaseIndicator:
        if(parameterSet_.get<bool>("VTKwriteMacroPhaseIndicator", 0))
          VTKwriteMacroPhaseIndicator();
        


        Qhom_.resize(numberOfMacroPhases_);
        Beff_.resize(numberOfMacroPhases_);

        constantMicrostructureVector_.resize(numberOfMacroPhases_);
      

        for(size_t i=0; i<numberOfMacroPhases_; i++)
        {
            std::cout << "------------------- " + std::to_string(i+1) + "-th macro phase setup " + "------------------- " << std::endl;
            LocalMicrostructureClass_.push_back(globalMicrostructure_.get("LocalMicrostructure_"+std::to_string(i+1)));

            // microProblem_.push_back(std::make_shared<MicroProblem>(LocalMicrostructureClass_[i](),parameterSet_, module_));

            // also store these quantities in a vector instead?
            //---Create localMicrostructure object
            // auto LocalMicrostructure = microProblem_[i]->getMicrostructure();
            auto LocalMicrostructure = LocalMicrostructureClass_[i]();

            // std::string testString;
            // LocalMicrostructure.get("phase" + std::to_string(1) + "_type").toC<std::string>(testString);
            // std::cout << "Teststring:" << testString << std::endl;


            // bool .macroPhase2_readEffectiveQuantities
            bool constantMicrostructure = true;
            globalMicrostructure_.get("macroPhase" + std::to_string(i+1) + "_constantMicrostructure").toC<bool>(constantMicrostructure);

            constantMicrostructureVector_[i] = constantMicrostructure;

            if (constantMicrostructure)
            {
              std::cout << "[Constant Microstructure]" << std::endl;
                          // Check whether to read effective quantities for current phase or to compute them.
              bool readEffectiveQuantities = false; //default
              globalMicrostructure_.get("macroPhase" + std::to_string(i+1) + "_readEffectiveQuantities").toC<bool>(readEffectiveQuantities);

              std::cout << "readEffectiveQuantities:" << readEffectiveQuantities << std::endl;

              if(readEffectiveQuantities)
              { 
                std::cout << "(Read effective quantities for phase " + std::to_string(i+1) + ")" << std::endl;
                LocalMicrostructure.get("effectiveQuadraticForm").toC<Dune::FieldMatrix<double,3,3>>(Qhom_[i]);
                LocalMicrostructure.get("effectivePrestrain").toC<Dune::FieldMatrix<double,2,2>>(Beff_[i]);
              } else {
                std::cout << "(Solving the Micro-Problem once to obtain effective quantities ...)" << std::endl;

                microProblem_.insert({i,std::make_shared<MicroProblem>(LocalMicrostructureClass_[i](),parameterSet_, module_)});
                (microProblem_[i])->getEffectiveQuantities(Beff_[i],Qhom_[i]);
              }

              printmatrix(std::cout, Qhom_[i], "Setup effective Quadratic form (Qhom) for Phase " +std::to_string(i+1) + " as ", "--");
              printmatrix(std::cout, Beff_[i], "Setup effective prestrain (Beff) for Phase " +std::to_string(i+1) + " as ", "--");
            }
            else 
              std::cout << "[Macroscopically varying Microstructure]" << std::endl;

        } //end-macroPhaseLoop

        // TEST Call python inner-class
        // std::cout << "Try to access inner python-class: " << std::endl;
        // Python::Reference innerClass; 
        // globalMicrostructure_.get("lg").toC<Python::Reference>(innerClass); //conversion to C of Python::Reference not implemented!

        // Python::Callable LocalMicrostructureClass = globalMicrostructure_.get(Python::Callable("LocalMicrostructure"));
        // Python::Callable LocalMicrostructureClass = globalMicrostructure_.get("LocalMicrostructure_1");
        // LocalMicrostructureClass_[0] = globalMicrostructure_.get("LocalMicrostructure_1");

        // create instance of inner class: 
        // auto LocalMicrostructure = LocalMicrostructureClass();




        // int numberOfMacroPhases_ = parameterSet_.get<int>("numberOfMacroPhases_", 1);



        //  microProblem_ = std::make_shared<MicroProblem>(microstructureClass_(),parameterSet_, module_);
        // microProblem_ = std::make_shared<MicroProblem>(LocalMicrostructureClass_[0](),parameterSet_, module_);

        // auto LocalMicrostructure = microProblem_->getMicrostructure();


        // access members of inner class:
        // auto test = LocalMicrostructure.get("name");



        // std::string test;
        // LocalMicrostructure.get("name").toC<std::string>(test);
        // std::cout << "access member of inner class: " << test << std::endl;

        






        // if(parameterSet_.get<bool>("macroscopically_varying_microstructure", 0))
        // {
        //    std::cout << "MACROSCOPICALLY VARYING MICROSTRUCTURE is used! " << std::endl;
        //    /*
        //     *  This 'default' microstructureClass_() is just used for an initial setup.
        //    */
        //   //  microProblem_ = std::make_shared<MicroProblem>(microstructureClass_(initialMacroPoint),parameterSet_, module_);
        //   //  microProblem_ = std::make_shared<MicroProblem>(microstructureClass_(),parameterSet_, module_);
        // }
        // else{
        //   std::cout << "CONSTANT MICROSTRUCTURE is used with effective quantities..." << std::endl;


        //   // Get Number of MacroPhases

        //   // If possible read the effective quantities from ParameterTree
        //   if(parameterSet_.get<bool>("read_effectiveQuantities_from_Parset", 1))
        //   {



        //     std::cout << "READ EFFECTIVE QUANTITIES FROM PARSET..." << std::endl;
        //     // module_.get("effectiveQuadraticForm").toC<Dune::FieldMatrix<double,3,3>>(Qhom_[0]);
        //     // module_.get("effectivePrestrain").toC<Dune::FieldMatrix<double,2,2>>(Beff_[0]);


        //     std::cout << "Obtain effective quantitites from Microstructure-Class:" << std::endl;

       
        //     // (microProblem_->microstructure_).get("phase" + std::to_string(1) + "_type").toC<std::string>(testString);



        //     std::string testString;
        //     LocalMicrostructure.get("phase" + std::to_string(1) + "_type").toC<std::string>(testString);
        //     std::cout << "Teststring:" << testString << std::endl;


        //     LocalMicrostructure.get("effectiveQuadraticForm").toC<Dune::FieldMatrix<double,3,3>>(Qhom_[0]);
        //     LocalMicrostructure.get("effectivePrestrain").toC<Dune::FieldMatrix<double,2,2>>(Beff_[0]);

        //   }
        //   else{
        //     std::cout << "Solving the Micro/Cell-Problem (once) to obtain effective quantities ..." << std::endl;

        //     // microProblem_ = std::make_shared<MicroProblem>(microstructureClass_(),parameterSet_, module_);
        //     // microProblem_.updateMicrostructure(microstructureClass_()); // pass a constant microstructure object.
        //     microProblem_->getEffectiveQuantities(Beff_[0],Qhom_[0]);
        //   }

        //   printmatrix(std::cout, Qhom_[0], "Setup effective Quadratic form (Qhom) as  ", "--");
        //   printmatrix(std::cout, Beff_[0], "Setup effective prestrain (Beff) as  ", "--");
        // }
          
    }



    void VTKwriteMacroPhaseIndicator() const
    {
          std::cout << "VTK write macroPhaseIndicator" << std::endl;
          std::string resultPath = parameterSet_.get("resultPath", "../../outputs");
          // std::string baseName = parameterSet_.get("baseName", "CellProblem-result");
          std::string baseName = "MacroPhaseIndicator";
          int subsamplingRefinement = parameterSet_.get<int>("MacroPhaseSubsamplingRefinement", 0);

          Dune::SubsamplingVTKWriter<GridView> MacroPhaseVtkWriter(localFunction_.gridView(), Dune::refinementLevels(subsamplingRefinement));
          
          MacroPhaseVtkWriter.addCellData(
              macroPhaseIndicator_,
              Dune::VTK::FieldInfo("macroPhaseIndicator_", Dune::VTK::FieldInfo::Type::scalar, 1));

          MacroPhaseVtkWriter.write(resultPath + "/" + baseName );
          std::cout << "wrote data to file:" + resultPath + "/" + baseName  << std::endl;  

          return;
    };




    template<class Itype>
    void resetEffectiveQuadraticForm(const Dune::FieldMatrix<Itype,3,3>& A, const int macroPhase) const
    {
        std::cout << "reset quadratic form ..." << std::endl;
        convertFieldMatrix<RT>(A,Qhom_[macroPhase]);
    }

    template<class Itype>
    void resetEffectivePrestrain(const Dune::FieldMatrix<Itype,2,2>& B, const int macroPhase) const
    {
        std::cout << "reset effective prestrain ..." << std::endl;
        convertFieldMatrix<RT>(B,Beff_[macroPhase]);
    }


    /**
     * @brief Apply to effective quadratic form (Qhom).
     * 
     * @param Qhom effective elastic moduli
     * @param A coefficient vector of first input (coefficient vectors of R_sym^{2x2} matrix in orthonormal basis)
     * @param B coefficient vector of second input (coefficient vectors of R_sym^{2x2} matrix in orthonormal basis)
     * @return double
     */
    auto applyQhom(const MatrixType& Qhom,const Dune::FieldVector<RT,3>& A, const Dune::FieldVector<RT,3>& B) const
    {
      return Qhom[0][0]*A[0]*B[0] + Qhom[0][1]*A[1]*B[0] + Qhom[0][2]*A[2]*B[0] 
          +  Qhom[1][0]*A[0]*B[1] + Qhom[1][1]*A[1]*B[1] + Qhom[1][2]*A[2]*B[1] 
          +  Qhom[2][0]*A[0]*B[2] + Qhom[2][1]*A[1]*B[2] + Qhom[2][2]*A[2]*B[2];
    }
    


    /** \brief Assemble the energy for a single element */
    virtual RT energy (const typename Basis::LocalView& localView,         
               const  std::vector<TargetSpace>& localSolution) const
    {
      RT energy = 0;

  
      /**
       * @brief TEST reset effective quantities.
       * 
       */
      // Dune::FieldMatrix<double,2,2> A = {{2.0, 0.0}, {0.0, 2.0}};
      // Dune::FieldMatrix<double,3,3> Q = {{1.0, 2.0, 3.0}, {4.0,5.0,6.0}, {1.0,2.0,3.0}};

      // printmatrix(std::cout, Qhom_, "Qhom: ", "--");
      // printmatrix(std::cout, Beff_, "Beff_: ", "--");
      // resetEffectiveQuadraticForm(Q);
      // resetEffectivePrestrain(A);
      // printmatrix(std::cout, Qhom_, "Qhom: ", "--");
      // printmatrix(std::cout, Beff_, "Beff_: ", "--");


      // print some quantities for debugging purposes.
      bool PRINT_DEBUG = parameterSet_.get<bool>("print_debug", 0);


#if 1
      // TODO: More serious order selection
      // int quadOrder = 3;
      // int quadOrder = 2;  //previously
      // int quadOrder = 1;  //too low. bad results.
      // int quadOrder = 6;  // Same as used/explained by [Rumpf et al. - Two-scale Finite elements approximation of a homogenized model]


      int quadOrder = parameterSet_.get<int>("macroQuadOrder", 6);
#else
      if (localFiniteElement.localBasis().order()==1)
        quadOrder = 2;
      else
        quadOrder = (localFiniteElement.type().isSimplex()) ? (localFiniteElement.localBasis().order() - 1) * 2
                                                            : (localFiniteElement.localBasis().order() * gridDim - 1) * 2;
#endif
      
      const auto element = localView.element();
      auto geometry = element.geometry();
      auto gridView = localFunction_.gridView();
      const auto &indexSet = gridView.indexSet();

      // bind to current element.
      localFunction_.bind(element);
      localForce_.bind(element);

      macroPhaseIndicatorLocalFunction_.bind(element);



      /**
       * @brief We need to update the Coefficients of localFunction_
       *        on the current element.
       */
      localFunction_.updateLocalCoefficients(localSolution,element);



      if(PRINT_DEBUG)
      {
          std::cout << " -------------------------------------------- " << std::endl;
          std::cout << "ELEMENT NUMBER( indexSet.index(element)) ):" << indexSet.index(element) << std::endl;
          std::cout << "geometry.corner(0): " << geometry.corner(0) << std::endl;
          std::cout << "geometry.corner(1)): " << geometry.corner(1) << std::endl;
          std::cout << "geometry.corner(2): " << geometry.corner(2) << std::endl;
      }


      /**
       * @brief  Get Coefficients of the discrete Gradient 
       * 
       * The discrete Gradient is a special linear combination represented in a [P2]^2 - (Lagrange) space
       * The coefficients of this linear combination correspond to certain linear combinations of the Gradients of localfunction_ .
       * The coefficients are stored in the form [Basisfunctions x components x gridDim]
       * in a BlockVector<FieldMatrix<RT, 3, gridDim>> .
       */
      BlockVector<FieldMatrix<RT, 3, gridDim>> discreteJacobianCoefficients;
      discreteGradientCoefficients(discreteJacobianCoefficients,lagrangeLFE_,localFunction_,element);
    
    
    
  

      /**
       * @brief Compute The energy contribution of  Qhom(II - Beff).
       *        This is called 'harmonicEnergy' for historical reasons: 
       *        without prestrain and if Qhom=|.|^2 (squared Frobenius-Norm),
       *        This is just the harmonic energy of the deformation Gradient.
       * 
       * 
       *        Mixed version using the binomal formula (a-b)^2 = a^2 - 2ab + b^2 
       *        for | II - Beff |^2 and discretize every term individually. 
       */
      RT harmonicEnergy = 0;

      // Gauss-Legendre-Quadrature (default):
      const auto &quadRule = QuadratureRules<double, gridDim>::rule(lagrangeLFE_.type(), quadOrder);
      //TEST Gauß-Jacobi-Quadrature formula: 
      // const auto &quadRule = QuadratureRules<double, gridDim>::rule(lagrangeLFE_.type(), quadOrder, QuadratureType::GaussJacobi_1_0);


      // Trapezoidal-rule:
      // std::vector<Dune::QuadraturePoint<double,2>> quadRule = { {{0.0,0.0}, 1.0/6.0}, {{1.0,0.0}, 1.0/6.0}, {{0.0,1.0}, 1.0/6.0} };
      // Trapezoidal-rule + evaluation on edge centers (only for testing purposes):
      // std::vector<Dune::QuadraturePoint<double,2>> quadRule = { {{0.0,0.0}, 1.0/6.0}, {{0.5,0.0}, 1.0/6.0}, {{1.0,0.0}, 1.0/6.0}, {{0.0,0.5}, 1.0/6.0}, {{0.5,0.5}, 1.0/6.0}, {{0.0,1.0}, 1.0/6.0} };
      // int count = 0;


      // std::cout << "Number of quadrature points: " << quadRule.size() << std::endl;


      for (auto&& quadPoint : quadRule)
      {
        const auto& quadPos = quadPoint.position();
        const auto integrationElement = geometry.integrationElement(quadPos);

        // std::cout << "quadPos:" << quadPos << std::endl;

        // exit(0);

        // Determine macro phase and type 
        int currentPhase = macroPhaseIndicatorLocalFunction_(quadPos);

        // bool constantMicrostructure = true;
        // globalMicrostructure_.get("macroPhase" + std::to_string(currentPhase) + "_constantMicrostructure").toC<bool>(constantMicrostructure); // Put this into a vector

        bool constantMicrostructure = constantMicrostructureVector_[currentPhase-1];

        if(!constantMicrostructure)
        {
           std::cout << "Compute effective quantities at current quadrature point: " << element.geometry().global(quadPos) << std::endl;

          
          // microProblem_.insert({(currentPhase-1),std::make_shared<MicroProblem>(LocalMicrostructureClass_[(currentPhase-1)](),parameterSet_, module_)});
          // (microProblem_[(currentPhase-1)])->getEffectiveQuantities(Beff_[(currentPhase-1)],Qhom_[(currentPhase-1)]); //this doesnt work in const-method?!

          auto currentMicroProblem = std::make_shared<MicroProblem>(LocalMicrostructureClass_[(currentPhase-1)](),parameterSet_, module_);
          currentMicroProblem->getEffectiveQuantities(Beff_[(currentPhase-1)],Qhom_[(currentPhase-1)]);

          // auto microProblem = microProblem_.find(currentPhase);

          // ((*microProblem).second)->updateMicrostructure(LocalMicrostructureClass_[(currentPhase-1)](element.geometry().global(quadPos))); 
          // ((*microProblem).second)->getEffectiveQuantities(Beff_[(currentPhase-1)],Qhom_[(currentPhase-1)]);
        }


        // //------------------------------------------------------------------------------------
        // if(parameterSet_.get<bool>("macroscopically_varying_microstructure", 0))
        // {
        //   std::cout << "Compute effective quantities at current quadrature point." << element.geometry().global(quadPos) << std::endl;

        //   // std::cout << "type_name<decltype(quadPos)>():" << type_name<decltype(quadPos)>() << std::endl;
        //   // std::cout << "initialMacroPoint:" << initialMacroPoint << std::endl;

        //   // auto test = microstructureClass_(element.geometry().global(quadPos));

        //     //USAGE:

        //   //TODO 
        //   //create instance of globalMicrostructure_Class_
        //   // Python::Reference globalMicrostructure_ = globalMicrostructure_Class_();
        //   // //get LocalMicrostructureClass
        //   // Python::Callable LocalMicrostructureClass = globalMicrostructure_.get("LocalMicrostructure_1");

          

        //   auto microProblem = microProblem_.find(0);

        //   ((*microProblem).second)->updateMicrostructure(LocalMicrostructureClass_[0](element.geometry().global(quadPos))); 
        //   ((*microProblem).second)->getEffectiveQuantities(Beff_[0],Qhom_[0]);

        //   // (*microProblem_[0])->updateMicrostructure(LocalMicrostructureClass_[0](element.geometry().global(quadPos))); 
        //   // (*microProblem_[0])->getEffectiveQuantities(Beff_[0],Qhom_[0]);

          
        //   //easier to simply create new microProblem object ...


        //   // microProblem_->updateMicrostructure(LocalMicrostructureClass_[0](element.geometry().global(quadPos))); 
        //   // microProblem_->getEffectiveQuantities(Beff_[0],Qhom_[0]);
        //   // microProblem_->updateMicrostructure(microstructureClass_(element.geometry().global(quadPos))); 
        //   // microProblem_->getEffectiveQuantities(Beff_,Qhom_,count);

        //   // count++;

        

        //   /**
        //    * @brief  TODO .. 
        //    * 
        //    *        1.information from each quadrature point has to enter.
        //    *        2. Caching of Effective Quantities
        //    * 
        //    */

        //   // exit(0);
        // }




        //------------------------------------------------------------------------------------



        //Get values of the P2-Basis functions on current quadrature point
        std::vector<FieldVector<double,1>> basisValues;
        lagrangeLFE_.localBasis().evaluateFunction(quadPos, basisValues);

        const auto geometryJacobianInverse = geometry.jacobianInverse(quadPos);
        std::vector<FieldMatrix<double,1,gridDim> > gradients;
        lagrangeLFE_.localBasis().evaluateJacobian(quadPos, gradients);

        for (size_t i=0; i< gradients.size(); i++)
           gradients[i] = gradients[i] * geometryJacobianInverse;


        Tensor3<RT,3,gridDim,gridDim> discreteHessian(0); 
        FieldMatrix<RT, 3, gridDim> discreteGradient(0);

        for (int k=0; k<3; k++)
        for (int l=0; l<gridDim; l++)
        for (std::size_t i=0; i<lagrangeLFE_.size(); i++)
        {
            discreteGradient[k][l]    += discreteJacobianCoefficients[i][k][l]*basisValues[i];
            discreteHessian[k][l][0]  += discreteJacobianCoefficients[i][k][l]*gradients[i][0][0];
            discreteHessian[k][l][1]  += discreteJacobianCoefficients[i][k][l]*gradients[i][0][1];
        }
      
        if(PRINT_DEBUG)
        {
          /**
           * @brief print the (three) 2x2 slices of the discrete Hessian and check for symmetry
           */
          FieldMatrix<RT, gridDim, gridDim> discreteHessian_slice1(0);
          FieldMatrix<RT, gridDim, gridDim> discreteHessian_slice2(0);
          FieldMatrix<RT, gridDim, gridDim> discreteHessian_slice3(0);

          for (int i=0; i<gridDim; i++)
          for (int l=0; l<gridDim; l++)
          {
              discreteHessian_slice1[i][l] = discreteHessian[0][i][l];
              discreteHessian_slice2[i][l] = discreteHessian[1][i][l];
              discreteHessian_slice3[i][l] = discreteHessian[2][i][l];
          }
          printmatrix(std::cout, discreteHessian_slice1, "discreteHessian_slice1: ", "--");
          printmatrix(std::cout, discreteHessian_slice2, "discreteHessian_slice2: ", "--");
          printmatrix(std::cout, discreteHessian_slice3, "discreteHessian_slice3: ", "--");
          printmatrix(std::cout, sym(discreteHessian_slice1), "SYM discreteHessian_slice1: ", "--");
          printmatrix(std::cout, sym(discreteHessian_slice2), "SYM discreteHessian_slice2: ", "--");
          printmatrix(std::cout, sym(discreteHessian_slice3), "SYM discreteHessian_slice3: ", "--");
        }



        /**
         * @brief Compute the normal vector of the surface parametrized by the 
         *        discrete deformation. This is given by the 
         *        cross product of partial derivatives
         */       
        FieldVector<RT,3> partialX = {discreteGradient[0][0], discreteGradient[1][0], discreteGradient[2][0]};
        FieldVector<RT,3> partialY = {discreteGradient[0][1], discreteGradient[1][1], discreteGradient[2][1]};
        // Normalize if requested
        // auto normX = partialX.two_norm();
        // auto normY = partialY.two_norm();
        // partialX /= normX;
        // partialY /= normY;
        // surfaceNormal[0] = partialX[1]*partialY[2] - partialX[2]*partialY[1];
        // surfaceNormal[1] = partialX[2]*partialY[0] - partialX[0]*partialY[2];
        // surfaceNormal[2] = partialX[0]*partialY[1] - partialX[1]*partialY[0];
        FieldVector<RT,3> surfaceNormal = {partialX[1]*partialY[2] - partialX[2]*partialY[1],
                                           partialX[2]*partialY[0] - partialX[0]*partialY[2],
                                           partialX[0]*partialY[1] - partialX[1]*partialY[0]};



        /**
         * @brief Compute the second fundamental form of the surface parametrized by the
         *        discrete deformation.
         *        We have to multiply by (-1.0) to get the right sign/normal/orientation.
         */       
        FieldMatrix<RT,gridDim,gridDim> secondFF(0);

        for(size_t m=0; m<2; m++)
        for(size_t n=0; n<2; n++)
        for(size_t k=0; k<3; k++) 
          secondFF[m][n] += surfaceNormal[k]*discreteHessian[k][n][m];
        // To get the right sign:
        secondFF *= -1.0;


        /**
         * @brief Compute the first term: Qhom(II)
         */
        auto weight = quadPoint.weight() * element.geometry().integrationElement(quadPos);

        RT term1 = 0.0;


        std::vector<FieldVector<RT,3>> X(3);
        tensorToSymCoefficients(discreteHessian, X);

        // apply Qhom to each slice of the Hessian... 
        for(int k=0; k<3; k++)
          term1 += applyQhom(Qhom_[currentPhase-1],X[k], X[k]); 
      

        term1 *= weight;




        /**
         * @brief Compute the second (mixed) term
         */
        FieldVector<RT,3> secondFFCoefficients = matrixToSymCoefficients(secondFF);
        FieldVector<RT,3> BeffCoefficients = matrixToSymCoefficients(Beff_[currentPhase-1]);
        RT term2 = 2.0 * weight* applyQhom(Qhom_[currentPhase-1],BeffCoefficients,secondFFCoefficients) ;


        /**
         * @brief Compute the third term
         */
        RT term3 = weight * applyQhom(Qhom_[currentPhase-1],BeffCoefficients,BeffCoefficients);


        harmonicEnergy +=  term1 - term2 + term3;     
        // std::cout << "harmonicEnergy: " << harmonicEnergy << std::endl;
        

        /**
         * @brief Test: Discretize II-Beff directly.
         *              This time with higher quad orders.
         *              * Does not seem to produce good results..
         * 
         */
        // RT newTerm = weight* applyQhom(Qhom_[currentPhase-1],(secondFFCoefficients - BeffCoefficients),(secondFFCoefficients - BeffCoefficients));
        // harmonicEnergy += newTerm;

      }
      


      /**
       * @brief Compute contribution of the force Term
       * 
       * Integrate the scalar product of the force 'f' 
       * with the local deformation function 'localFunction_'
       * over the current element.
       */
      RT forceTermEnergy = 0;

      if (parameterSet_.get<bool>("assemble_force_term", 1))
      {
        for (auto&& quadPoint : quadRule)
        {
            auto deformationValue = localFunction_.evaluate(quadPoint.position());
            auto forceValue = localForce_(quadPoint.position());
            auto weight = quadPoint.weight() * element.geometry().integrationElement(quadPoint.position());
            forceTermEnergy += deformationValue.globalCoordinates() * forceValue * weight;
        }
      }

      /**
       * @brief Return total energy.
       */
      return harmonicEnergy - forceTermEnergy;
    }

  };

}  // namespace Dune::GFE
#endif