#ifndef DUNE_MICROSTRUCTURE_MATRIXOPERATIONS_HH
#define DUNE_MICROSTRUCTURE_MATRIXOPERATIONS_HH

namespace MatrixOperations {

	using MatrixRT = Dune::FieldMatrix< double, 3, 3>;
	using VectorRT = Dune::FieldVector< double, 3>;

	using std::sin;
	using std::cos;
	


	static MatrixRT sym (MatrixRT M) { // 1/2 (M^T + M)
    	MatrixRT ret(0);
    	for (int i=0; i<3; i++)
		{
			ret[i][i] = M[i][i];
    		for (int j=i+1; j<3; j++)
			{
    			ret[i][j] = 0.5*(M[i][j] + M[j][i]);
				ret[j][i] = ret[i][j];
			}
		}
    	return ret;
	}


	static VectorRT crossProduct (VectorRT v, VectorRT w) { // v otimes w
    	return {v[1]*w[2] - v[2]*w[1], -1*(v[0]*w[2] - v[2]*w[0]), v[0]*w[1] - v[1]*w[0]};
	}

	static MatrixRT rankoneTensorproduct (VectorRT v, VectorRT w) { // v otimes w
    	return 
    	{{v[0]*w[0], v[1]*w[0], v[2]*w[0]}, 
	     {v[0]*w[1], v[1]*w[1], v[2]*w[1]}, 
	     {v[0]*w[2], v[1]*w[2], v[2]*w[2]}};
	}

	static MatrixRT nematicLiquidCrystal (double p, VectorRT n){ //B = 1/6*p*Id + 1/2*p*(n otimes n)
		MatrixRT B(0);
	    for (int i=0;i<3;i++)
	        B[i][i]=p/6.0;
	    MatrixRT n_ot_n = rankoneTensorproduct(n,n);
	    n_ot_n*=p/2.0;
	    B += n_ot_n;
	    return B;
	}

	static MatrixRT biotStrainApprox (VectorRT U, VectorRT k, VectorRT e_cs){ //E_h = (U +  k x e_cs, 0, 0)
	    VectorRT k_x_ecs = crossProduct(k, e_cs);
	    VectorRT U_plus_k_x_ecs = U + k_x_ecs;
	    VectorRT e_1 = {1, 0, 0};
	    return rankoneTensorproduct(U_plus_k_x_ecs, e_1);
	}

	
    static MatrixRT crossSectionDirectionScaling(double w, MatrixRT M){
		return {{M[0][0], M[0][1], w*M[0][2]},
                {M[1][0], M[1][1], w*M[1][2]},
                {M[2][0], M[2][1], w*M[2][2]}
               };
	}

	static double trace (MatrixRT M){ 
	    return M[0][0]+ M[1][1] + M[2][2];
	}

	static double scalarProduct (MatrixRT M1, MatrixRT M2){ 
		double sum = 0.0;
		for (int i=0; i<3; i++)
		    for (int j=0; j<3; j++)
		    	sum += M1[i][j] * M2[i][j];
	    return sum;
	}


	/**
	 * @brief Determines rotation matrix based on an axis and an angle
	 * 
	 * @param axis 
	 * @param angle 
	 * @return MatrixRT 
	 */
	static MatrixRT rotationMatrix(int axis, double angle){

		switch (axis)
		{
			case 0:
			{
				return {{1.0,          0,            0   },
						{  0, cos(angle), -1.0*sin(angle)},
						{  0, sin(angle),      cos(angle)}};
			}
			case 1:
			{
				return {{     cos(angle),   0, sin(angle)},
						{              0, 1.0,      0    },
						{-1.0*sin(angle),   0, cos(angle)}};
			}
			case 2:
			{
				return {{cos(angle), -1.0*sin(angle),    0},
						{sin(angle),      cos(angle),    0},
						{         0,               0,  1.0}};
			}
			default:
				DUNE_THROW(Dune::Exception, " axis not feasible. rotationMatrix is only implemented for 3x3-matrices. Choose between 0: x-axis, 1: y-axis, 2: z-axis");
		}
	}

	/**
	 * @brief 6x6 matrix that transforms the strain tensor. This matrix is used to 
	 * 		  transform the compliance matrix (given in Voigt notation) into another frame.
	 * 		  see 'https://en.wikipedia.org/wiki/Orthotropic_material#Condition_for_material_symmetry_2' 
	 * 		  for details.
	 * 
	 * @param axis 
	 * @param angle 
	 * @return MatrixRT 
	 */
	static Dune::FieldMatrix<double,6,6> rotationMatrixCompliance(int axis, double angle){
		
		MatrixRT R = rotationMatrix(axis,angle);

        return {{    R[0][0]*R[0][0],     R[0][1]*R[0][1],     R[0][2]*R[0][2],                 R[0][1]*R[0][2],                 R[0][0]*R[0][2],                 R[0][0]*R[0][1]},
                {    R[1][0]*R[1][0],     R[1][1]*R[1][1],     R[1][2]*R[1][2],                 R[1][1]*R[1][2],                 R[1][0]*R[1][2],                 R[1][0]*R[1][1]},
                {    R[2][0]*R[2][0],     R[2][1]*R[2][1],     R[2][2]*R[2][2],                 R[2][1]*R[2][2],                 R[2][0]*R[2][2],                 R[2][0]*R[2][1]},
                {2.0*R[1][0]*R[2][0], 2.0*R[1][1]*R[2][1], 2.0*R[1][2]*R[2][2], R[1][1]*R[2][2]+R[1][2]*R[2][1], R[1][0]*R[2][2]+R[1][2]*R[2][0], R[1][0]*R[2][1]+R[1][1]*R[2][0]},
                {2.0*R[0][0]*R[2][0], 2.0*R[0][1]*R[2][1], 2.0*R[0][2]*R[2][2], R[0][1]*R[2][2]+R[0][2]*R[2][1], R[0][0]*R[2][2]+R[0][2]*R[2][0], R[0][0]*R[2][1]+R[0][1]*R[2][0]},
                {2.0*R[0][0]*R[1][0], 2.0*R[0][1]*R[1][1], 2.0*R[0][2]*R[1][2], R[0][1]*R[1][2]+R[0][2]*R[1][1], R[0][0]*R[1][2]+R[0][2]*R[1][0], R[0][0]*R[1][1]+R[0][1]*R[1][0]}
               };
	}


	/**
	 * @brief  Scale last three columns of stiffness matrix by a factor of 2.
	 * 		   Inserting this facotr in the stiffness matrix allows to use the 
	 * 		   same Matrix-to-Vector mapping for both the stress and the strain.
	 * 
	 * @param S  Stiffness matrix
	 */
	static void scaleStiffnessMatrix(Dune::FieldMatrix<double,6,6>& S){
      for(size_t i = 0; i<6; i++) 
      for(size_t j = 3; j<6; j++)
      {
        S[i][j] = 2.0*S[i][j];
      }
	}

    static double linearizedStVenantKirchhoffDensity(double mu, double lambda, MatrixRT E1, MatrixRT E2)  // CHANGED
    {  
        auto t1 = 2.0 * mu * sym(E1) + MatrixRT(Dune::ScaledIdentityMatrix<double,3>(lambda * trace(sym(E1))));
        auto tmp1 = scalarProduct(t1,sym(E2));
        return tmp1; 
        
	}



    // --- Generalization: Define Quadratic QuadraticForm
    static double QuadraticForm(const double mu, const double lambda, const MatrixRT M){
        
        auto tmp1 = sym(M);
        double tmp2 = tmp1.frobenius_norm();
//         double tmp2 = norm(M);                                //TEST 
        return lambda*std::pow(trace(M),2) + 2.0*mu*pow( tmp2 ,2);
//         return lambda*std::pow(trace(M),2) + 2*mu*pow( norm( sym(M) ),2);
    }

    

	static double generalizedDensity(const double mu, const double lambda, MatrixRT F, MatrixRT G){
     /// Write this whole File as a Class that uses lambda,mu as members ? 
        
     // Define L via Polarization-Identity from QuadratifForm
     // <LF,G> := (1/2)*(Q(F+G) - Q(F) - Q(G) ) 
        return (1.0/2.0)*(QuadraticForm(mu,lambda,F+G) - QuadraticForm(mu,lambda,F) - QuadraticForm(mu,lambda,G) );
    }

	static MatrixRT matrixSqrt(MatrixRT M){
		std::cout << "matrixSqrt not implemented!!!" << std::endl;//implement this
		return M;
	}

	static double lameMu(double E, double nu){
		return 0.5 * 1.0/(1.0 + nu) * E;
	}

	static double lameLambda(double E, double nu){
		return nu/(1.0-2.0*nu) * 1.0/(1.0+nu) * E;
	}

	static bool isInRotatedPlane(double phi, double x1, double x2){
		return cos(phi)*x1 + sin(phi)*x2 > 0;
	}


	extern "C" 
	{

		MatrixRT new_sym(MatrixRT M)
		{
			return sym(M);
		}

	}


	/**
	 * @brief This is used to convert FieldMatrices to 'adouble'
	 *        since adouble-type can not be read from ParSet.
	 *        is there a better way?
	 * 
	 * @tparam Rtype 
	 * @tparam IMatrixType 
	 * @param A 
	 * @param B 
	*/
	template<class Rtype>
	static void convertFieldMatrix(auto& A,
								auto& B)
	{
	for(size_t i=0; i<B.N(); i++)
	for(size_t j=0; j<B.M(); j++)
	B[i][j] = (Rtype)A[i][j];

	return;
	}



}

#endif
