#ifndef DUNE_MICROSTRUCTURE_MICROPROBLEM_HH
#define DUNE_MICROSTRUCTURE_MICROPROBLEM_HH


#include <fstream>
#include <iostream>

#include <dune/common/parametertree.hh>
#include <dune/common/parametertreeparser.hh>
#include <dune/common/float_cmp.hh>
#include <dune/common/math.hh>


#include <dune/fufem/dunepython.hh>

#include <dune/grid/uggrid.hh>
#include <dune/grid/yaspgrid.hh>
#include <dune/grid/io/file/vtk/subsamplingvtkwriter.hh>

#include <dune/istl/matrix.hh>
#include <dune/istl/bcrsmatrix.hh>
#include <dune/istl/multitypeblockmatrix.hh>
#include <dune/istl/multitypeblockvector.hh>
#include <dune/istl/matrixindexset.hh>
#include <dune/istl/solvers.hh>
#include <dune/istl/spqr.hh>
#include <dune/istl/preconditioners.hh>
#include <dune/istl/io.hh>
#include <dune/istl/eigenvalue/test/matrixinfo.hh> // TEST: compute condition Number 

#include <dune/functions/functionspacebases/interpolate.hh>
#include <dune/functions/backends/istlvectorbackend.hh>
#include <dune/functions/functionspacebases/powerbasis.hh>
#include <dune/functions/functionspacebases/compositebasis.hh>
#include <dune/functions/functionspacebases/lagrangebasis.hh>
#include <dune/functions/functionspacebases/periodicbasis.hh>
#include <dune/functions/functionspacebases/subspacebasis.hh>
#include <dune/functions/functionspacebases/boundarydofs.hh>
#include <dune/functions/gridfunctions/discreteglobalbasisfunction.hh>
#include <dune/functions/gridfunctions/gridviewfunction.hh>

#include <dune/microstructure/matrix_operations.hh>
#include <dune/microstructure/CorrectorComputer.hh>    
#include <dune/microstructure/EffectiveQuantitiesComputer.hh>  
#include <dune/microstructure/prestrainedMaterial.hh>  

#include <dune/solvers/solvers/umfpacksolver.hh>   

// #include <iomanip>   // needed when working with relative paths e.g. from python-scripts

// using namespace Dune;
// using namespace MatrixOperations;


  // method to print types of objects:
//   template <class T>
//   constexpr std::string_view type_name()
//   {
//       using namespace std;
//   #ifdef __clang__
//       string_view p = __PRETTY_FUNCTION__;
//       return string_view(p.data() + 34, p.size() - 34 - 1);
//   #elif defined(__GNUC__)
//       string_view p = __PRETTY_FUNCTION__;
//   #  if __cplusplus < 201402
//       return string_view(p.data() + 36, p.size() - 36 - 1);
//   #  else
//       return string_view(p.data() + 49, p.find(';', 49) - 49);
//   #  endif
//   #elif defined(_MSC_VER)
//       string_view p = __FUNCSIG__;
//       return string_view(p.data() + 84, p.size() - 84 - 7);
//   #endif
//   }


//--------------------------------------------------------
class MicroProblem
{
    private:
        static const int dim = 3;
        using CellGridType = Dune::YaspGrid<dim, Dune::EquidistantOffsetCoordinates<double, dim> >;
        using GridView = typename CellGridType::LeafGridView;
        // typedef typename    Dune::Functions::DefaultGlobalBasis<Dune::Functions::PowerPreBasis<Dune::Functions::BasisFactory::FlatLexicographic, \
        //             Dune::Functions::Experimental::TransformedIndexPreBasis<Dune::Functions::LagrangePreBasis<Dune::GridView< \
        //             Dune::DefaultLeafGridViewTraits<const Dune::YaspGrid<3, Dune::EquidistantOffsetCoordinates<double, dim> > > >, 1, double>, \
        //             Dune::Functions::BasisFactory::Experimental::Impl::PeriodicIndexingTransformation>, dim> >  BasisType; 


        using BasisType = typename Dune::Functions::DefaultGlobalBasis<Dune::Functions::PowerPreBasis<Dune::Functions::BasisFactory::FlatLexicographic, \
                    Dune::Functions::Experimental::TransformedIndexPreBasis<Dune::Functions::LagrangePreBasis<Dune::GridView< \
                    Dune::DefaultLeafGridViewTraits<const Dune::YaspGrid<3, Dune::EquidistantOffsetCoordinates<double, dim> > > >, 1, double>, \
                    Dune::Functions::BasisFactory::Experimental::Impl::PeriodicIndexingTransformation>, dim> >;

        using MaterialType = prestrainedMaterial<GridView>;


        Python::Reference microstructure_;
        const Dune::ParameterTree parameterSet_;
        const Python::Module module_;
        const CellGridType grid_;
        const GridView gridView_; 
        const BasisType basis_;
        // Dune::ParameterTree parameterSet_;
        // Python::Module module_;
        // CellGridType grid_;
        // GridView gridView_; 
        // BasisType basis_;


        // std::fstream log_; // deprecated: only used to define CorrectorComputer object. 


        // MaterialType material_;
        std::shared_ptr<MaterialType> material_;

        // CorrectorComputer<BasisType, MaterialType> correctorComputer_;
        std::shared_ptr<CorrectorComputer<BasisType, MaterialType> > correctorComputer_; 
        EffectiveQuantitiesComputer<BasisType,MaterialType> effectiveQuantitiesComputer_;
        


    public:
        

        //Constructor
        MicroProblem(const Python::Reference microstructure,
                     const Dune::ParameterTree& parameterSet,
                     const Python::Module pyModule)
            : microstructure_(microstructure), 
            parameterSet_(parameterSet),
            module_(pyModule),
            grid_(createGrid()),
            gridView_(grid_.leafGridView()),
            basis_(createPeriodicBasis()),
            // material_(setupMaterial(gridView_))
            // material_(gridView_,microstructure_,parameterSet_,module_),
            material_(std::make_shared<MaterialType>(gridView_,microstructure_,parameterSet_,module_)),
            correctorComputer_(std::make_shared<CorrectorComputer<BasisType, MaterialType>>(basis_, material_, parameterSet_)),
            // correctorComputer_(basis_, material_, parameterSet_),
            effectiveQuantitiesComputer_(correctorComputer_)
            // effectiveQuantitiesComputer_(correctorComputer_,material_)  // Remove material dependency 
        {};
        // //Constructor
        // MicroProblem(const Dune::ParameterTree& parameterSet,
        //              Python::Module pyModule)
        //     : parameterSet_(parameterSet),
        //     module_(pyModule),
        //     grid_(createGrid()),
        //     gridView_(grid_.leafGridView()),
        //     basis_(createPeriodicBasis())
        // {
        //     initialSetup();
        // };


        // Create a default constructor like this?
        // MicroProblem() = default;


        // MaterialType setupMaterial(const GridView& gridView) const 
        // {
        //     return MaterialType(gridView,microstructure_,parameterSet_,module_);
        // }



        /**
         * @brief Infrastructure for handling periodicity.
         *         Check whether two points are equal on R/Z x R/Z x R
         */
        static constexpr auto equivalent = [](const Dune::FieldVector<double,3>& x, const Dune::FieldVector<double,3>& y)
        {
            return ( (Dune::FloatCmp::eq(x[0],y[0]) or Dune::FloatCmp::eq(x[0]+1,y[0]) or Dune::FloatCmp::eq(x[0]-1,y[0]))
                    and (Dune::FloatCmp::eq(x[1],y[1]) or Dune::FloatCmp::eq(x[1]+1,y[1]) or Dune::FloatCmp::eq(x[1]-1,y[1]))
                    and (Dune::FloatCmp::eq(x[2],y[2]))
                );
        };


        /**
         * @brief Generate the grid.
         *        Corrector Problem Domain (-1/2,1/2)^3.
         */
        const CellGridType createGrid() const
        {
            Dune::FieldVector<double,dim> lower({-1.0/2.0, -1.0/2.0, -1.0/2.0});
            Dune::FieldVector<double,dim> upper({1.0/2.0, 1.0/2.0, 1.0/2.0});
            int microGridLevel = parameterSet_.get<int>("microGridLevel", 0);
            std::array<int, dim> nElements = {(int)std::pow(2,microGridLevel) ,(int)std::pow(2,microGridLevel) ,(int)std::pow(2,microGridLevel)};
            // std::cout << "Number of Grid-Elements in each direction: " << nElements << std::endl;
            std::cout << "Number of Micro Grid-Elements in each direction: " << (int)std::pow(2,microGridLevel) << std::endl;

            return CellGridType(lower,upper,nElements);
        }

        const BasisType createPeriodicBasis() const
        {
            // using namespace Functions::BasisFactory;
            Dune::Functions::BasisFactory::Experimental::PeriodicIndexSet periodicIndices;

            //--- Get PeriodicIndices for periodicBasis (Don't do the following in real life: It has quadratic run-time in the number of vertices.)
            for (const auto& v1 : vertices(gridView_))
            for (const auto& v2 : vertices(gridView_))
                if (equivalent(v1.geometry().corner(0), v2.geometry().corner(0)))
                {
                    periodicIndices.unifyIndexPair({gridView_.indexSet().index(v1)}, {gridView_.indexSet().index(v2)});
                }

            return makeBasis(gridView_,
                            power<dim>(                                  
                            Dune::Functions::BasisFactory::Experimental::periodic(lagrange<1>(), periodicIndices),
                            flatLexicographic()
                            ));   
        }


        void updateMicrostructure(Python::Reference microstructure) 
        {
            microstructure_ = microstructure;

            // material_.updateMicrostructure(microstructure_);
            material_->updateMicrostructure(microstructure_);
            correctorComputer_->updateMaterial(material_);
            effectiveQuantitiesComputer_.updateCorrectorComputer(correctorComputer_);
        }


        Python::Reference getMicrostructure()
        {
            return microstructure_;
        }




        // template<class Btype,class Qtype>
        // void getEffectiveQuantities(Btype& B, Qtype& Q, int count = 0)
        template<class Btype,class Qtype>
        void getEffectiveQuantities(Btype& B, Qtype& Q)
        {
            Dune::Timer microProblemTimer;
            // std::cout << "getting effective quantities.. " << std::endl;

            // --- Get scale ratio 
            // double gamma = parameterSet_.get<double>("gamma",1.0); 
            double gamma;
            microstructure_.get("gamma").toC<double>(gamma);
            if (parameterSet_.get<bool>("printMicroOutput ", false))
              std::cout << "Microstructure scale ratio (gamma) set to : " << gamma << std::endl;


            // /**
            //  * @brief Create Log (Remove later) 
            //  * 
            //  */
            // std::string baseName = parameterSet_.get("baseName", "CellProblem-result");
            // std::string outputPath = parameterSet_.get("resultPath", "../../outputs") ;
            // //--- setup Log-File
            // std::fstream log;
            // log.open(outputPath + "/" + baseName + "_log.txt" ,std::ios::out);

            // std::cout << "log done" << std::endl;
          
            // //  Create prestrained material object
            // auto material = prestrainedMaterial(basis_.gridView(),microstructure_,parameterSet_,module_);


            // std::cout << "type_name<decltype(material_)>():" << type_name<decltype(material_)>() << std::endl;
            // std::cout << "type_name<decltype(material)>():" << type_name<decltype(material)>() << std::endl;





            // // std::cout << "TEST-VTK write.." << std::endl;
            // material.writeVTKMaterialFunctions(parameterSet_.get<int>("microGridLevel", 0));
            // std::cout << "Part1 works." << std::endl;


            if (parameterSet_.get<bool>("write_materialFunctions", false))
               material_->writeVTKMaterialFunctions(parameterSet_.get<int>("microGridLevel", 0));

            // exit(0);

            // std::cout << "Material setup done" << std::endl;

            // // Compute Correctors
            // auto correctorComputer = CorrectorComputer(basis_, material, log, parameterSet_);
            // #if 0

            // std::cout << "Starting Corrector assembly" << std::endl;
            correctorComputer_->assemble();
            // std::cout << "Starting Corrector solve..." << std::endl;
            correctorComputer_->solve();


            // std::cout << "Corrector done" << std::endl;
            // //--- Compute effective quantities
            // auto effectiveQuantitiesComputer = EffectiveQuantitiesComputer(correctorComputer,material);
            effectiveQuantitiesComputer_.computeEffectiveQuantities();


            // std::cout << "effective Quantities  done" << std::endl;

            // //--- Get effective quantities
            auto Beffvec = effectiveQuantitiesComputer_.getBeff();
            // // convert CoefficientVector to matrix
            auto Beff = CoefficientsToSymMatrix(Beffvec);
            Q = effectiveQuantitiesComputer_.getQeff();
            B = CoefficientsToSymMatrix(Beffvec);


            if (parameterSet_.get<bool>("printMicroOutput ", false))
            {
                printmatrix(std::cout, Q, "Matrix Qeff", "--");
                printmatrix(std::cout, B, "Beff", "--");
            }

            std::cout << "Micro-Problem took " << microProblemTimer.elapsed() << " seconds." << std::endl;

            // #endif

            // exit(0);
        }

};


#endif 