#ifndef DUNE_MICROSTRUCTURE_PRESTRAINEDMATERIAL_HH
#define DUNE_MICROSTRUCTURE_PRESTRAINEDMATERIAL_HH

#include <dune/common/parametertree.hh>
#include <dune/functions/gridfunctions/gridviewfunction.hh>
#include <dune/functions/gridfunctions/gridviewentityset.hh>
#include <dune/fufem/dunepython.hh>
#include <dune/grid/uggrid.hh>
#include <dune/grid/io/file/vtk/subsamplingvtkwriter.hh>
#include <dune/grid/yaspgrid.hh>
#include <dune/microstructure/matrix_operations.hh>
#include <dune/microstructure/voigthelper.hh>

// //TEST dune-vtk
// // #include <dune/vtk/function.hh>
// #include <dune/vtk/vtkwriter.hh>
// // #include <dune/vtk/vtkwriterbase.hh>
// #include <dune/vtk/writers/unstructuredgridwriter.hh>
// #include <dune/vtk/datacollectors/continuousdatacollector.hh>
// #include <dune/vtk/datacollectors/discontinuousdatacollector.hh>
// #include <dune/vtk/datacollectors/quadraticdatacollector.hh>
// #include <dune/vtk/datacollectors/lagrangedatacollector.hh>

using namespace MatrixOperations;
using namespace Dune::Functions;
using namespace Dune::Functions::BasisFactory;
using std::pow;
using std::abs;
using std::sqrt;
using std::sin;
using std::cos;
using std::shared_ptr;
using std::make_shared;


/**
 * @brief The 'prestrainedMaterial' class descripes the LOCAL microstructure.
 *        It stores all the necessary information to define the local Cell-problem.
 *        This includes:
 * 
 *        Indicatorfunction, #Phases, Prestrain, elasticity/compliance tensors ,scale-ratio 'gamma'
 *        
 */

//-------------------------------------------------------- 
template <class GridView>         
class prestrainedMaterial
{

public:
  static const int dimworld = 3;  //GridView::dimensionworld;
  static const int dim = 3;       //Domain::dimension;
  using Domain = typename GridView::template Codim<0>::Geometry::GlobalCoordinate;
  using LocalDomain = typename GridView::template Codim<0>::Geometry::LocalCoordinate;
  using Element = typename GridViewEntitySet<GridView, 0>::Element;
  using ScalarRT = Dune::FieldVector< double, 1>;
  using VectorRT = Dune::FieldVector< double, dimworld>;
  using MatrixRT = Dune::FieldMatrix< double, dimworld, dimworld>;
  using FuncScalar = std::function< double(const Domain&) >;
  using Func2int = std::function< int(const Domain&) >;
  using Func2Tensor = std::function< MatrixRT(const Domain&) >;
  using Func2TensorParam = std::function< MatrixRT(const MatrixRT& ,const Domain&) >;
  using Func2TensorPhase = std::function< MatrixRT(const MatrixRT& ,const int&) >;
  using MatrixFunc  = std::function< MatrixRT(const MatrixRT&) >;
  using MatrixPhase = std::function< MatrixRT(const int&) >;

  using IndicatorGridViewFunctionType = GridViewFunction<int(const Domain&), GridView>;
  using LocalIndicatorFunctionType = LocalFunction<int(const Domain&), typename GridViewEntitySet<GridView, 0>::Element >;


  // const GridView& gridView_;
  const GridView& gridView_;
  const Dune::ParameterTree parameterSet_;

  // Elasticity tensors (in Voigt notation)
  std::vector<VoigtMatrix<double,3> > L_;

  std::vector<Func2Tensor> prestrain_;

  const Python::Module module_;  // actually deprecated now?



  Python::Reference microstructure_;


  Func2int indicatorFunction_;
  // IndicatorFunctionType indicatorFunction_;
  // GridViewFunction<int(const Domain&), GridView> indicatorFunctionGVF_;
  // LocalFunction<int(const Domain&), typename GridViewEntitySet<GridView, 0>::Element >  localIndicatorFunction_;
  IndicatorGridViewFunctionType indicatorFunctionGVF_;
  LocalIndicatorFunctionType localIndicatorFunction_;

  double gamma_; 


public: 
    prestrainedMaterial(const GridView& gridView,
                        Python::Reference microstructure,
                        const Dune::ParameterTree& parameterSet,
                        const Python::Module pyModule)
                       : gridView_(gridView), 
                         microstructure_(microstructure),
                         parameterSet_(parameterSet),
                         module_(pyModule) 
    {
       setup();
    } 
    ~prestrainedMaterial(){std::cout << "prestrainedMaterial object destroyed!" << std::endl;};

    // prestrainedMaterial() = default;
    
    //copy constructor 
    prestrainedMaterial(const prestrainedMaterial& that) = default;

    // Copy assignment operator
    // prestrainedMaterial& operator=(prestrainedMaterial const& that)
    // {
    //   gridView_ = that.gridView_;
    //   parameterSet_ = that.parameterSet_;
    //   L_ = that.L_;
    //   prestrain_ = that.prestrain_;
    //   module_ = that.module_;
    //   microstructure_ = that.microstructure_;

    //   return *this;
    // }

  
  void updateMicrostructure(const Python::Reference microstructure) 
  {
      // std::cout << "updateMicrostrucrue of prestrainedMaterial" << std::endl;
      microstructure_ = microstructure;
  }


  void updateGamma(const double gamma) 
  {
      gamma_ = gamma;
  }


/**
 * @brief TODO: Code allows a spatially varying prestrain. However most of the time we only 
 *              use a constant prestrain. 
 *              - Add switch (constant/non-constant) prestrain?
 * 
 * @param phase           material phase
 * @return Func2Tensor 
 */
Func2Tensor setupPrestrainPhase(const int phase){

  // auto prestrain = Python::make_function<MatrixRT>(module_.get("prestrain_phase" + std::to_string(phase)));
  auto prestrain = Python::make_function<MatrixRT>(Python::Callable(microstructure_.get("prestrain_phase" + std::to_string(phase))));

  //TODO...

  int axis = 0;
  double angle = 0;
  try 
  {
    // module_.get("phase" + std::to_string(phase) + "_axis").toC<int>(axis);
    // module_.get("phase" + std::to_string(phase) + "_angle").toC<double>(angle);
    axis = parameterSet_.get<int>("phase" + std::to_string(phase) + "_axis", 0);
    angle = parameterSet_.get<double>("phase" + std::to_string(phase) + "_angle", 0);
  }
  catch(Dune::Exception&)
  {
    //default frame is used.
  }
  /**
   * @brief (optional) Tramsform prestrain to the correct frame.
   */
  if(abs(angle) > 1e-12){
    auto R = rotationMatrix(axis,angle);
    Func2Tensor f = [prestrain,R] (const Domain& x) { return (prestrain(x).leftmultiply(R)).rightmultiply(R.transposed()); };

    return f;
  }
  else 
    return prestrain;
}



// Dune::FieldMatrix<double,6,6> setupPhase(const std::string phaseType, Python::Module module, int phase)
Dune::FieldMatrix<double,6,6> setupPhase(const int phase)
{
  std::string phaseType;

  // phaseType = parameterSet_.get<std::string>("phase" + std::to_string(phase) + "_type");
  microstructure_.get("phase" + std::to_string(phase) + "_type").toC<std::string>(phaseType);

  std::cout << "Setup phase "<< phase << " as " << phaseType << " material." << std::endl;

  if(phaseType == "isotropic")
  {
      Dune::FieldVector<double,2> materialParameters(0);
      // module_.get("materialParameters_phase" + std::to_string(phase)).toC<Dune::FieldVector<double,2>>(materialParameters);
      microstructure_.get("materialParameters_phase" + std::to_string(phase)).toC<Dune::FieldVector<double,2>>(materialParameters);
      std::cout << "Lame-Parameters (mu,lambda): (" << materialParameters[0] << "," << materialParameters[1] << ") " << std::endl;
      return isotropic(materialParameters[0],materialParameters[1]);
  }


  /**
   * @brief For other material classes. The stiffness-/compliance matrix depends on the 
   *        coordinate frame.
   * 
   *        (optional) rotate material coordinate frame by an 'angle' around 'axis'
   */
  int axis = 0;
  double angle = 0;

  try 
  {
    // axis = parameterSet_.get<int>("phase" + std::to_string(phase) + "_axis", 0);
    // angle = parameterSet_.get<double>("phase" + std::to_string(phase) + "_angle", 0);
    microstructure_.get("phase" + std::to_string(phase) + "_axis").toC<int>(axis);
    microstructure_.get("phase" + std::to_string(phase) + "_angle").toC<double>(angle);
  }
  catch(Dune::Exception& e)
  {
    std::cout << "(Could not read rotation axis & angle for material phase) " << phase << ": - Canonical Frame (default) is used." << std::endl;
  }


  if(phaseType == "orthotropic")
  {
      Dune::FieldVector<double,9> materialParameters(0);
      // module_.get("materialParameters_phase" + std::to_string(phase)).toC<Dune::FieldVector<double,9>>(materialParameters);
      microstructure_.get("materialParameters_phase" + std::to_string(phase)).toC<Dune::FieldVector<double,9>>(materialParameters);
      return orthotropic(axis,angle,materialParameters);
  }
  if(phaseType == "transversely_isotropic")
  {
      Dune::FieldVector<double,5> materialParameters(0);
      // module_.get("materialParameters_phase" + std::to_string(phase)).toC<Dune::FieldVector<double,5>>(materialParameters);
      microstructure_.get("materialParameters_phase" + std::to_string(phase)).toC<Dune::FieldVector<double,5>>(materialParameters);
      return transversely_isotropic(axis,angle,materialParameters);
  }
  if(phaseType == "general_anisotropic")
  {
      Dune::FieldMatrix<double,6,6> materialParameters(0); //compliance matric
      // module_.get("materialParameters_phase" + std::to_string(phase)).toC<Dune::FieldMatrix<double,6,6>>(materialParameters);
      microstructure_.get("materialParameters_phase" + std::to_string(phase)).toC<Dune::FieldMatrix<double,6,6>>(materialParameters);
      return general_anisotropic(axis,angle,materialParameters);
  }
  else
    DUNE_THROW(Dune::Exception, " Unknown material class for phase " + std::to_string(phase));
}

  /**
   * @brief Setup method that determines the correct stiffness-matrix (constitutive law)
   *        for each material phase separately. 
   * 
   */
  void setup()
  {

    //--- Get gamma (scale-ratio).
    microstructure_.get("gamma").toC<double>(gamma_);
    std::cout << "microstructure gamma: " << gamma_ << std::endl;

    //--- Get indicator function.
    indicatorFunction_ = Python::make_function<int>(Python::Callable(microstructure_.get("indicatorFunction")));
    // indicatorFunction_ = Python::make_function<int>(module_.get("indicatorFunction"));
    indicatorFunctionGVF_ = Dune::Functions::makeGridViewFunction(indicatorFunction_, gridView_);
    localIndicatorFunction_ = localFunction(indicatorFunctionGVF_);

    int phases;
    // phases = parameterSet_.get<int>("Phases"); 
    microstructure_.get("phases").toC<int>(phases);
    std::cout << "---- Starting material setup... (Number of Phases: "<< phases << ") "  << std::endl;

    L_.resize(phases);
    prestrain_.resize(phases);

    for (int i=0; i<phases; i++)
    {
      L_[i] = setupPhase(i+1);
      prestrain_[i] = setupPrestrainPhase(i+1);
    }

    std::cout << "Material setup done." << std::endl;
  }

  ////////////////////////////////////////////////////////
  //      MATERIAL CLASSES DEFINITIONS:
  ////////////////////////////////////////////////////////

  /** \brief Isotropic stiffness matrix in Voigt notation but scaled by a factor of 2 in the last three columns. */
  Dune::FieldMatrix<double,6,6> isotropic(const double& mu, const double& lambda)
  {
        return {{lambda+2.0*mu, lambda       , lambda       , 0.0   , 0.0   , 0.0    },
                {lambda       , lambda+2.0*mu, lambda       , 0.0   , 0.0   , 0.0    },
                {lambda       , lambda       , lambda+2.0*mu, 0.0   , 0.0   , 0.0    },
                {  0.0        ,  0.0         , 0.0          , 2.0*mu, 0.0   , 0.0    },
                {  0.0        ,  0.0         , 0.0          , 0.0   , 2.0*mu, 0.0    },
                {  0.0        ,  0.0         , 0.0          , 0.0   , 0.0   , 2.0*mu }
              };
  }
  //-------------------------------------------------------------------------------


  /**
   * @brief Definition of orthotropic elasticity/stiffness tensor (in Voigt notation).
   *        The convention in literature is to provide the elastic constants in form of the 'compliance matrix'
   *        In order to compute stresses from strain, we have to invert the 'compliance matrix' to obtain the 'stiffness matrix'
   *        (see https://en.wikipedia.org/wiki/Orthotropic_material for more details)
   * 
   * @param axis    rotate material frame around axis
   * @param angle   rotation angle 
   * @param p       elastic constants: (E1,E2,E3,G12,G23,G31,nu12,nu13,nu23)
   * @return        Stiffness matrix
   */
  Dune::FieldMatrix<double,6,6> orthotropic(const int& axis,
                                            const double& angle,
                                            const Dune::FieldVector<double,9>& p
                                            ) 
  {
      auto E_1 = p[0];
      auto E_2 = p[1];
      auto E_3 = p[2];
      auto G_12 = p[3];
      auto G_23 = p[4];
      auto G_31 = p[5];
      auto nu_12 = p[6];
      auto nu_13 = p[7];
      auto nu_23 = p[8];
      //other three poisson ratios are not independent 
      auto nu_21 = (p[6]/p[0])*p[1];
      auto nu_31 = (p[7]/p[0])*p[2];
      auto nu_32 = (p[8]/p[1])*p[2]; 

      std::string axis_string;
      switch (axis)
      {
        case 0:
             axis_string = "x - axis";
             break;
        case 1: 
             axis_string = "y - axis";
             break;
        case 2: 
             axis_string = "z - axis";
             break;
        default: break;
      }

      // Compliance matrix
      Dune::FieldMatrix<double,6,6> Compliance = {{1.0/E_1      , -nu_21/E_2   , -nu_31/E_3   , 0.0     , 0.0     , 0.0     },
                                                  {-nu_12/E_1   , 1.0/E_2      , -nu_32/E_3   , 0.0     , 0.0     , 0.0     },
                                                  {-nu_13/E_1   , -nu_23/E_2   , 1.0/E_3      , 0.0     , 0.0     , 0.0     },
                                                  {  0.0        ,  0.0         , 0.0          , 1.0/G_23, 0.0     , 0.0     },
                                                  {  0.0        ,  0.0         , 0.0          , 0.0     , 1.0/G_31, 0.0     },
                                                  {  0.0        ,  0.0         , 0.0          , 0.0     , 0.0     , 1.0/G_12}
                                                  };


      /**
       * @brief (optional) Tramsform compliance matrix to the correct frame.
       */
      if(abs(angle) > 1e-12){
        std::cout << "material frame rotated around axis: " << axis_string << " by an angle of: " << angle << std::endl;
        Dune::FieldMatrix<double,6,6> C_rot = rotationMatrixCompliance(axis,angle);
        (Compliance.leftmultiply(C_rot)).rightmultiply(C_rot.transposed());
      }

      printmatrix(std::cout, Compliance, "Compliance matrix used:", "--");

      // invert Compliance matrix to obtain elasticity/stiffness matrix
      Compliance.invert();
      // printmatrix(std::cout, Compliance, "Stiffness Tensor", "--");
      scaleStiffnessMatrix(Compliance);
      // printmatrix(std::cout, Compliance, "Stiffness Tensor scaled", "--");
      // printmatrix(std::cout, Compliance, "Stiffness Tensor", "--");
      return Compliance;
  }
  //-------------------------------------------------------------------------------

  /**
   * @brief Definition of transversely isotropic elasticity tensor (in Voigt notation)
   *        (see https://en.wikipedia.org/wiki/Poisson%27s_ratio)
   * @param axis    rotate material frame around axis
   * @param angle   rotation angle 
   * @param p       elastic constants: {E1,E2,G12,nu12,nu23}
   * @return        Stiffness matrix
   */
  Dune::FieldMatrix<double,6,6> transversely_isotropic(const int& axis,
                                          const double& angle,
                                          const Dune::FieldVector<double,5>& p
                                          ) 
  {
    // (see https://en.wikipedia.org/wiki/Poisson%27s_ratio)
      auto E_1 = p[0];
      auto E_2 = p[1];
      auto E_3 = E_2;
      auto nu_12 = p[3];
      auto nu_13 = nu_12;
      auto nu_21 = (nu_12/E_1)*E_2;
      auto nu_31 = nu_21;
      auto nu_23 = p[4];
      auto nu_32 = nu_23;
      auto G_12 = p[2];
      auto G_23 = E_2/(2.0*(1+nu_23));
      auto G_31 = G_23;
      //other three poisson ratios are not independent 

      std::string axis_string;
      switch (axis)
      {
        case 0:
             axis_string = "x - axis";
             break;
        case 1: 
             axis_string = "y - axis";
             break;
        case 2: 
             axis_string = "z - axis";
             break;
        default: break;
      }
      
      //---compliance matrix
      Dune::FieldMatrix<double,6,6> Compliance = {{1.0/E_1      , -nu_21/E_2   , -nu_31/E_3   , 0.0     , 0.0     , 0.0   },
                                                  {-nu_12/E_1   , 1.0/E_2      , -nu_32/E_3   , 0.0     , 0.0     , 0.0   },
                                                  {-nu_13/E_1   , -nu_23/E_2   , 1.0/E_3      , 0.0     , 0.0     , 0.0   },
                                                  {  0.0        ,  0.0         , 0.0          , 1.0/G_23, 0.0     , 0.0   },
                                                  {  0.0        ,  0.0         , 0.0          , 0.0     , 1.0/G_31, 0.0   },
                                                  {  0.0        ,  0.0         , 0.0          , 0.0     , 0.0     , 1.0/G_12}
                                                  };

      /**
       * @brief (optional) Tramsform compliance matrix to the correct frame.
       */
      if(abs(angle) > 1e-12){
        // std::cout << "material frame rotated around axis: " << axis << " by an angle of: " << angle << std::endl;
        std::cout << "material frame rotated around axis: " << axis_string << " by an angle of: " << angle << std::endl;
        Dune::FieldMatrix<double,6,6> C_rot = rotationMatrixCompliance(axis,angle);
        (Compliance.leftmultiply(C_rot)).rightmultiply(C_rot.transposed());
      }

      printmatrix(std::cout, Compliance, "Compliance matrix used:", "--");

      // invert Compliance matrix to obtain elasticity/stiffness matrix
      Compliance.invert();
      // printmatrix(std::cout, Compliance, "Stiffness Tensor", "--");
      
      scaleStiffnessMatrix(Compliance);

      return Compliance;
  }
  //-------------------------------------------------------------------------------


  /**
   * @brief Definition of general anisotropic elasticity tensor (in Voigt notation)
   *        
   * @param axis        rotate material frame around axis
   * @param angle       rotation angle 
   * @param Compliance  Compliance matrix
   * @return            Stiffness matrix
   */  
  Dune::FieldMatrix<double,6,6> general_anisotropic(const int& axis,
                                                    const double& angle,
                                                    Dune::FieldMatrix<double,6,6>& Compliance
                                                    ) 
  {
      std::string axis_string;
      switch (axis)
      {
        case 0:
             axis_string = "x - axis";
             break;
        case 1: 
             axis_string = "y - axis";
             break;
        case 2: 
             axis_string = "z - axis";
             break;
        default: break;
      }

      /**
       * @brief (optional) Tramsform compliance matrix to the correct frame.
       */
      if(abs(angle) > 1e-12){
        // std::cout << "material frame rotated around axis: " << axis << " by an angle of: " << angle << std::endl;
        std::cout << "material frame rotated around axis: " << axis_string << " by an angle of: " << angle << std::endl;
        Dune::FieldMatrix<double,6,6> C_rot = rotationMatrixCompliance(axis,angle);
        (Compliance.leftmultiply(C_rot)).rightmultiply(C_rot.transposed());
      }
      printmatrix(std::cout, Compliance, "Compliance matrix used:", "--");


      // invert Compliance matrix to obtain elasticity/stiffness matrix
      Compliance.invert();
      // printmatrix(std::cout, Compliance, "Stiffness Tensor", "--");
      
      scaleStiffnessMatrix(Compliance);

      return Compliance;
  }
  //-------------------------------------------------------------------------------


  VoigtVector<double,3> applyElasticityTensor(const VoigtVector<double,3>& G, const int& phase)  const
  {
    VoigtVector<double,3> out(0);

    L_[phase-1].mv(G,out);

    return out;
  }


  MatrixRT prestrain(const int& phase,const Domain& x) const  
  {
    return prestrain_[phase-1](x);
  }


  int IndicatorFunction(const Domain& x) const
  {
    return indicatorFunction_(x);
  }


    ///////////////////////////////
    // Getter
    ///////////////////////////////
    Dune::ParameterTree getParameterSet() const {return parameterSet_;}
    std::vector<VoigtMatrix<double,3>> getElasticityTensor() const {return L_;}



    MatrixPhase getPrestrainFunction() const {return prestrain_;} 

    auto getLocalIndicatorFunction() const {return localIndicatorFunction_;}   //get as localFunction
    auto getIndicatorFunctionGVF() const {return indicatorFunctionGVF_;}   //get as GridViewFunction
    auto getIndicatorFunction() const {return indicatorFunction_;}

    auto getGamma() const {return gamma_;}


  ///////////////////////////////
  // VTK - Writer
  ///////////////////////////////

  /**
   * @brief Write material indicator function as a grid function to VTK.
   *        The Writer uses piecewise constant (P0) interpolation
   * 
   * @param level  GridLevel
   */
  // void writeVTKMaterialFunctions(const int level, double vtkIndex = 1) const
  void writeVTKMaterialFunctions(const int level) const
  {
        // std::string outputPath = parameterSet_.get("outputPath", "../../outputs");
        std::string resultPath = parameterSet_.get("resultPath", "../../outputs");
        std::string baseName = parameterSet_.get("baseName", "CellProblem-result");

        int subsamplingRefinement = parameterSet_.get<int>("MaterialSubsamplingRefinement", 2);

        Dune::SubsamplingVTKWriter<GridView> MaterialVtkWriter(gridView_, Dune::refinementLevels(subsamplingRefinement));
        
        MaterialVtkWriter.addCellData(
            indicatorFunction_,
            Dune::VTK::FieldInfo("materialIndicatorFunction_", Dune::VTK::FieldInfo::Type::scalar, 1));

        MaterialVtkWriter.write(resultPath + "/" + baseName + "_MaterialFunction-level"+ std::to_string(level) );
        std::cout << "wrote data to file:" + resultPath + "/" + baseName + "_MaterialFunction-level"+ std::to_string(level) << std::endl;  

        
        //-- Version using the Dune::VTKSequenceWriter
        // std::shared_ptr<Dune::SubsamplingVTKWriter<GridView> > MaterialVtkWriter = std::make_shared<Dune::SubsamplingVTKWriter<GridView>>(gridView_, Dune::refinementLevels(subsamplingRefinement));

        // MaterialVtkWriter->addCellData(indicatorFunction_,
        //                                Dune::VTK::FieldInfo("materialIndicatorFunction_", Dune::VTK::FieldInfo::Type::scalar, 1));


        // Dune::VTKSequenceWriter<GridView> SequenceMaterialVTK(MaterialVtkWriter, resultPath + "/" + baseName + "_MaterialFunction-level"+ std::to_string(level));
        // SequenceMaterialVTK.write(vtkIndex);
        // std::cout << "wrote data to file:" + resultPath + "/" + baseName + "_MaterialFunction-level"+ std::to_string(level) + "-" + std::to_string(vtkIndex) << std::endl;  

        //--Version using dune-vtk 
        // Dune::Vtk::DiscontinuousDataCollector<GridView> discontinuousDataCollector(gridView_);
        // Dune::Vtk::YaspDataCollector<GridView, 4> yaspDataCollector(gridView_);
        // Dune::Vtk::LagrangeDataCollector<GridView, 4> lagrangeDataCollector(gridView_);
        // Dune::Vtk::UnstructuredGridWriter writer(lagrangeDataCollector, Dune::Vtk::FormatTypes::ASCII, Dune::Vtk::DataTypes::FLOAT32);
        // // Dune::Vtk::UnstructuredGridWriter writer(discontinuousDataCollector, Dune::Vtk::FormatTypes::ASCII, Dune::Vtk::DataTypes::FLOAT32);

        // auto f2 = Dune::Functions::makeAnalyticGridViewFunction(indicatorFunction_ ,gridView_);
        // // writer.addPointData(f2 , "IndicatorFunction");
        // writer.addCellData(f2 , "IndicatorFunction");
        // writer.write(resultPath + "/" + baseName + "_MaterialFunction-level"+ std::to_string(level) + "_DuneVTK");
	      return;
  };

};

#endif