import math
import numpy as np

class ParameterSet(dict):
    def __init__(self, *args, **kwargs):
        super(ParameterSet, self).__init__(*args, **kwargs)
        self.__dict__ = self

parameterSet = ParameterSet()

""""
    Experiment: Taken from 
                [Rumof,Simon,Smoch]:
                TWO-SCALE FINITE ELEMENT APPROXIMATION OF A HOMOGENIZED PLATE MODEL
                * Microstructure with periodic diagonal aligned stripe pattern ("Twisted valley") 
"""

#############################################
#  Paths
#############################################
parameterSet.resultPath = '/home/klaus/Desktop/Dune_bendIso/dune-microstructure/outputs_twisted-valley'
# parameterSet.outputPath = '/home/klaus/Desktop/Dune_bendIso/dune-microstructure/outputs_twisted-valley'
parameterSet.baseName= 'twisted-valley'

#############################################
#  Grid parameters
#############################################
nX=8
nY=8

parameterSet.structuredGrid = 'simplex'
parameterSet.lower = '0 0'
parameterSet.upper = '1 1'
parameterSet.elements = str(nX)+' '+  str(nY)

parameterSet.macroGridLevel = 2   #good results only for refinementLevel >=4

#############################################
#  Options
#############################################
parameterSet.measure_analyticalError = False
parameterSet.measure_isometryError = False
parameterSet.vtkWrite_analyticalSurfaceNormal = False
parameterSet.vtkwrite_analyticalSolution = False


parameterSet.conforming_DiscreteJacobian = 0
#############################################
#  Solver parameters
#############################################
# Tolerance of the multigrid solver
parameterSet.tolerance = 1e-12
# Maximum number of multigrid iterations
parameterSet.maxProximalNewtonSteps = 200
# Initial regularization
parameterSet.initialRegularization = 200
# Measure convergence
parameterSet.instrumented = 0

############################
#   Problem specifications
############################
# Dimension of the domain (only used for numerical-test python files)
parameterSet.dim = 2

#############################################
#  VTK/Output
#############################################
# Write discrete solution as .vtk-File
parameterSet.writeVTK = 1
parameterSet.vtkwrite_analyticalSolution = 0
parameterSet.vtkWrite_analyticalSurfaceNormal = 0

# The grid can be refined several times for a higher resolution in the VTK-file.
parameterSet.subsamplingRefinement = 0

# Write Dof-Vector to .txt-file
parameterSet.writeDOFvector = 0

#############################################
#  Dirichlet boundary indicator
#############################################
def dirichlet_indicator(x) :
    if( (x[0] <= 0.01) or (x[0] >= 0.99)):
        return True
    else:
        return False



#############################################
#  MICROSTRUCTURE
############################################
parameterSet.printMicroOutput = False
parameterSet.VTKwriteMacroPhaseIndicator = True
parameterSet.MacroPhaseSubsamplingRefinement = 3

class GlobalMicrostructure:
    """ Class that represents the global microstructure.
    
        The global microstructure can be defined individually 
        for different (finitely many) macroscopic phases. 
        Each macroscopic phase corresponds to a 'LocalMicrostructure_'
        sub-class that defines the necessary data for the local 
        micro-problem. 

        Currently, there are three possibilities for the LocalMicrostructure:
            (1) Read the effective quantities (Qhom,Beff) from this parset.
            (Constant local microstructure)
            (2) Solve the micro-problem once to obtain the effective quantities. 
            (Constant local microstructure)
            (3) Solve for micro-problem for each macroscopic quadrature point.
            (Macroscopocally varying local microstructure)) 
    """
    def __init__(self,macroPoint=[0,0]):
        self.macroPhases = 1

        # define first local microstructure options.
        self.macroPhase1_constantMicrostructure = True
        self.macroPhase1_readEffectiveQuantities = False;



    def macroPhaseIndicator(self,y): #y :macroscopic point
            """ Indicatorfunction that determines the domains 
                i.e. macro-phases of different local microstructures.
            """
            return 1;
                

        
    # Represents the local microstructure in Phase 1
    class LocalMicrostructure_1:  
        def __init__(self,macroPoint=[0,0]):
            self.macroPoint = macroPoint
            # self.macroPoint = macroPoint
            self.gamma = 0.1    #in the future this might change depending on macroPoint.
            self.phases = 2     #in the future this might change depending on macroPoint.
            #--- Define different material phases:
            #- PHASE 1
            self.phase1_type="isotropic"
            self.materialParameters_phase1 = [5.0/3.0, 5.0/2.0]

            self.materialRatio = 1.0/50   
            #- PHASE 2
            self.phase2_type="isotropic"
            self.materialParameters_phase2 = [(5.0/3.0)*self.materialRatio, (5.0/2.0)*self.materialRatio]    


            # self.effectivePrestrain= np.array([[0.75, 0.0],
            #                                    [0.0, 1.0]]);

            # self.effectiveQuadraticForm = np.array([[3.0, 0.0, 0.0],
            #                                         [0.0, 1.0, 0.0],
            #                                         [0.0, 0.0, 0.8]]);

        def one_norm(self,v,w):
            return np.abs(v) + np.abs(w)


        #--- Indicator function for material phases
        def indicatorFunction(self,y):
            #shift domain
            x = [y[0]+0.5,y[1]+0.5]

            indicator = (( (self.one_norm(x[0]-1,x[1])) < (3/4)) & ((self.one_norm(x[0]-1,x[1])) > (1/4)))\
                    | (( (self.one_norm(x[0]-1,x[1])) < (7/4)) & ((self.one_norm(x[0]-1,x[1])) > (5/4)))\
                    
            # indicator = (( (self.one_norm(x[1]-1,x[0])) < (3/4)) & ((self.one_norm(x[1]-1,x[0])) > (1/4)))\
            # | (( (self.one_norm(x[1]-1,x[0])) < (7/4)) & ((self.one_norm(x[1]-1,x[0])) > (5/4)))\
                
            if(indicator):
                return 1    #Phase1   
            else :
                return 2    #Phase2
            
        #--- Define prestrain function for each phase (also works with non-constant values)
        def prestrain_phase1(self,x):
            return [[0, 0, 0], [0,0,0], [0,0,0]]

        def prestrain_phase2(self,x):
            return [[0, 0, 0], [0,0,0], [0,0,0]]
        


# #############################################
# #  Microstructure
# ############################################
# parameterSet.prestrainFlag = True
# parameterSet.macroscopically_varying_microstructure = False
# parameterSet.read_effectiveQuantities_from_Parset = False # Otherwise the Micro/Cell-Problem is solved once to obtain the quantities.
# # parameterSet.macroscopically_varying_microstructure = True
# # parameterSet.read_effectiveQuantities_from_Parset = False # Otherwise the Micro/Cell-Problem is solved once to obtain the quantities.


# parameterSet.printMicroOutput = True





# class Microstructure:
#     def __init__(self,macroPoint=[0,0]):  #default value for initialization
#         self.macroPoint = macroPoint
#         # self.macroPoint = macroPoint
#         self.gamma = 0.1    #in the future this might change depending on macroPoint.
#         self.phases = 2     #in the future this might change depending on macroPoint.
#         #--- Define different material phases:
#         #- PHASE 1
#         self.phase1_type="isotropic"
#         self.materialParameters_phase1 = [5.0/3.0, 5.0/2.0]

#         self.materialRatio = 1.0/50   
#         #- PHASE 2
#         self.phase2_type="isotropic"
#         self.materialParameters_phase2 = [(5.0/3.0)*self.materialRatio, (5.0/2.0)*self.materialRatio]    

#     def one_norm(self,v,w):
#         return np.abs(v) + np.abs(w)


#     #--- Indicator function for material phases
#     def indicatorFunction(self,y):
#         #shift domain
#         x = [y[0]+0.5,y[1]+0.5]

#         indicator = (( (self.one_norm(x[0]-1,x[1])) < (3/4)) & ((self.one_norm(x[0]-1,x[1])) > (1/4)))\
#                 | (( (self.one_norm(x[0]-1,x[1])) < (7/4)) & ((self.one_norm(x[0]-1,x[1])) > (5/4)))\
                
#         # indicator = (( (self.one_norm(x[1]-1,x[0])) < (3/4)) & ((self.one_norm(x[1]-1,x[0])) > (1/4)))\
#         # | (( (self.one_norm(x[1]-1,x[0])) < (7/4)) & ((self.one_norm(x[1]-1,x[0])) > (5/4)))\
            
#         if(indicator):
#             return 1    #Phase1   
#         else :
#             return 2    #Phase2
        
#     #--- Define prestrain function for each phase (also works with non-constant values)
#     def prestrain_phase1(self,x):
#         return [[0, 0, 0], [0,0,0], [0,0,0]]

#     def prestrain_phase2(self,x):
#         return [[0, 0, 0], [0,0,0], [0,0,0]]




# effectivePrestrain= np.array([[0.75, 0.0],
#                               [0.0, 1.0]]);

# effectiveQuadraticForm = np.array([[3.0, 0.0, 0.0],
#                                    [0.0, 1.0, 0.0],
#                                    [0.0, 0.0, 0.8]]);



#############################################
#  Initial iterate function
#############################################
#RUMPF EXAMPLE
def f(x):
    a = (3.0/16.0)
    if(x[0] <= 0.01):
        return [x[0]+a, x[1], 0]
    elif(x[0] >= 0.99):
        return [x[0] - a, x[1], 0]
    else:
        return [x[0], x[1], 0]



def df(x):
    return ((1,0),
            (0,1),
            (0,0))





fdf = (f, df)
#############################################
#  Force
############################################
parameterSet.assemble_force_term = True

def force(x):
    return [0, 0, -1e-4]


#############################################
#  Grid parameters
#############################################
parameterSet.microGridLevel = 4

#############################################
#  Assembly options
#############################################
parameterSet.set_IntegralZero = 1            #(default = false)
parameterSet.set_oneBasisFunction_Zero = 1   #(default = false)
#parameterSet.arbitraryLocalIndex = 7            #(default = 0)
#parameterSet.arbitraryElementNumber = 3         #(default = 0)

#############################################
#  Solver Options, Type: #1: CG - SOLVER , #2: GMRES - SOLVER, #3: QR - SOLVER, #4: UMFPACK - SOLVER (default)
#############################################
parameterSet.Solvertype = 4        # recommended to use iterative solver (e.g GMRES) for finer grid-levels
parameterSet.Solver_verbosity = 0  #(default = 2)  degree of information for solver output


#############################################
#  Write/Output options      #(default=false)
#############################################
# --- (Optional output) write Material / prestrain / Corrector functions to .vtk-Files:
parameterSet.write_materialFunctions = 1   # VTK indicator function for material/prestrain definition
#parameterSet.write_prestrainFunctions = 1  # VTK norm of B (currently not implemented)
parameterSet.MaterialSubsamplingRefinement = 1

# --- (Additional debug output)
parameterSet.print_debug = 0  #(default=false)

parameterSet.print_corrector_matrices = 0

# --- Write Correctos to VTK-File:  
parameterSet.writeCorrectorsVTK = 1

# --- Use caching of element matrices:  
parameterSet.cacheElementMatrices = 1



# --- check orthogonality (75) from paper: 
parameterSet.write_checkOrthogonality = 0
