import math

class ParameterSet(dict):
    def __init__(self, *args, **kwargs):
        super(ParameterSet, self).__init__(*args, **kwargs)
        self.__dict__ = self

parameterSet = ParameterSet()


""""
    Experiment: Parametrized laminate composite with isotopic material and vanishing Poisson ratio.
    
                extensively used in 
                    [Böhnlein,Neukamm,Padilla-Garza,Sander - A homogenized bending theory for prestrained plates]
                    https://link.springer.com/article/10.1007/s00332-022-09869-8

    theta: width of center material
    theta_rho: ratio between the prestrain of the fibres.
    theta_mu: ratio between lema 

"""

#############################################
#  Paths
#############################################
parameterSet.resultPath = '/home/klaus/Desktop/Dune_bendIso/dune-microstructure/outputs_parametrized_laminate'
parameterSet.baseName= 'parametrized_laminate'   #(needed for Output-Filename)

#############################################
#  Material Setup
#############################################


#Parameters used for analytical solutions:
# theta = 0.5
# mu_1 = 1.0;
# theta_mu = 10.0
# mu_2 = theta_mu * mu_1;
# rho_1 = 1.0;
# theta_rho = 10.0
# rho_2 = rho_1*theta_rho;

    


class Microstructure:
    def __init__(self):
        # self.macroPoint = macroPoint
        self.gamma = 1.0    #in the future this might change depending on macroPoint.
        self.phases = 4     #in the future this might change depending on macroPoint.
        #--- Define different material phases:

        # ########### Options for material phases: #################################
        # #     1. "isotropic"     2. "orthotropic"      3. "transversely_isotropic"   4. "general_anisotropic"
        # #########################################################################
        # ## Notation - Parameter input :
        # # isotropic (Lame parameters) : [mu , lambda]
        # #         orthotropic         : [E1,E2,E3,G12,G23,G31,nu12,nu13,nu23]   # see https://en.wikipedia.org/wiki/Poisson%27s_ratio with x=1,y=2,z=3
        # # transversely_isotropic      : [E1,E2,G12,nu12,nu23]
        # # general_anisotropic         : full compliance matrix C
        # ######################################################################

        # theta_mu = 10.0
        theta_mu = 2.0
        #- PHASE 1
        self.phase1_type="isotropic"
        self.materialParameters_phase1 = [theta_mu*1.0, 0]   
        #- PHASE 2
        self.phase2_type="isotropic"
        self.materialParameters_phase2 = [1.0, 0]   
        #- PHASE 3
        self.phase3_type="isotropic"
        self.materialParameters_phase3 = [theta_mu*1.0, 0]
        #- PHASE 4
        self.phase4_type="isotropic"
        self.materialParameters_phase4 = [1.0, 0]



    #--- Indicator function for material phases
    def indicatorFunction(self,x):
        theta = 0.5
        if (abs(x[0]) < (theta/2) and x[2] < 0 ):
            return 1    #Phase1
        elif (abs(x[0]) > (theta/2) and x[2] > 0 ):
            return 2    #Phase2
        elif (abs(x[0]) < (theta/2) and x[2] > 0 ):
            return 3    #Phase3
        else :
            return 4    #Phase4
        
        

    #--- Define prestrain function for each phase (also works with non-constant values)
    def prestrain_phase1(self,x):
        theta_rho = 10.0
        return [[theta_rho*1.0, 0, 0], [0,theta_rho*1.0,0], [0,0,theta_rho*1.0]]

    def prestrain_phase2(self,x):
        return [[1, 0, 0], [0,1,0], [0,0,1]]

    def prestrain_phase3(self,x):
        return [[0, 0, 0], [0,0,0], [0,0,0]]

    def prestrain_phase4(self,x):
        return [[0, 0, 0], [0,0,0], [0,0,0]]


#############################################
#  Grid parameters
#############################################
parameterSet.microGridLevel = 3

#############################################
#  Assembly options
#############################################
parameterSet.set_IntegralZero = 1            #(default = false)
parameterSet.set_oneBasisFunction_Zero = 1   #(default = false)
#parameterSet.arbitraryLocalIndex = 7            #(default = 0)
#parameterSet.arbitraryElementNumber = 3         #(default = 0)

#############################################
#  Solver Options, Type: #1: CG - SOLVER , #2: GMRES - SOLVER, #3: QR - SOLVER, #4: UMFPACK - SOLVER (default)
#############################################
parameterSet.Solvertype = 4        # recommended to use iterative solver (e.g GMRES) for finer grid-levels
parameterSet.Solver_verbosity = 0  #(default = 2)  degree of information for solver output


#############################################
#  Write/Output options      #(default=false)
#############################################
# --- (Optional output) write Material / prestrain / Corrector functions to .vtk-Files:
parameterSet.write_materialFunctions = 1   # VTK indicator function for material/prestrain definition
#parameterSet.write_prestrainFunctions = 1  # VTK norm of B (currently not implemented)
parameterSet.MaterialSubsamplingRefinement= 2


parameterSet.write_EffectiveQuantitiesToTxt= True

# --- (Additional debug output)
parameterSet.print_debug = 0  #(default=false)

parameterSet.print_corrector_matrices = 0

# --- Write Correctos to VTK-File:  
parameterSet.writeCorrectorsVTK = 1

# --- Use caching of element matrices:  
parameterSet.cacheElementMatrices = 1

# --- check orthogonality (75) from paper: 
parameterSet.write_checkOrthogonality = 0

# --- write effective quantities (Qhom.Beff) to .txt-files
parameterSet.write_EffectiveQuantitiesToTxt = True