#include <config.h>
#include <array>
#include <vector>
#include <fstream>
#include <fenv.h>

#include <iostream>
#include <dune/common/indices.hh>
#include <dune/common/bitsetvector.hh>
#include <dune/common/parametertree.hh>
#include <dune/common/parametertreeparser.hh>
#include <dune/common/float_cmp.hh>
#include <dune/common/math.hh>
#include <dune/common/fvector.hh>
#include <dune/common/fmatrix.hh> 

#include <dune/geometry/quadraturerules.hh>

#include <dune/grid/uggrid.hh>
#include <dune/grid/yaspgrid.hh>
#include <dune/grid/io/file/vtk/subsamplingvtkwriter.hh>

#include <dune/istl/matrix.hh>
#include <dune/istl/bcrsmatrix.hh>
#include <dune/istl/multitypeblockmatrix.hh>
#include <dune/istl/multitypeblockvector.hh>
#include <dune/istl/matrixindexset.hh>
#include <dune/istl/solvers.hh>
#include <dune/istl/spqr.hh>
#include <dune/istl/preconditioners.hh>
#include <dune/istl/io.hh>
#include <dune/istl/eigenvalue/test/matrixinfo.hh> // TEST: compute condition Number 

#include <dune/functions/functionspacebases/interpolate.hh>
#include <dune/functions/backends/istlvectorbackend.hh>
#include <dune/functions/functionspacebases/powerbasis.hh>
#include <dune/functions/functionspacebases/compositebasis.hh>
#include <dune/functions/functionspacebases/lagrangebasis.hh>
#include <dune/functions/functionspacebases/periodicbasis.hh>
#include <dune/functions/functionspacebases/subspacebasis.hh>
#include <dune/functions/functionspacebases/boundarydofs.hh>
#include <dune/functions/gridfunctions/discreteglobalbasisfunction.hh>
#include <dune/functions/gridfunctions/gridviewfunction.hh>

#include <dune/fufem/dunepython.hh>

#include <dune/microstructure/matrix_operations.hh>
#include <dune/microstructure/CorrectorComputer.hh>    
#include <dune/microstructure/EffectiveQuantitiesComputer.hh>  
#include <dune/microstructure/prestrainedMaterial.hh>  

#include <dune/solvers/solvers/umfpacksolver.hh>   //TEST 

#include <any>
#include <variant>
#include <string>
#include <iomanip>   // needed when working with relative paths e.g. from python-scripts

using namespace Dune;
using namespace MatrixOperations;


using GridView = Dune::YaspGrid<3, Dune::EquidistantOffsetCoordinates<double, 3>>::LeafGridView;

//////////////////////////////////////////////////////////////////////
// Helper functions for Table-Output
//////////////////////////////////////////////////////////////////////
/*! Center-aligns string within a field of width w. Pads with blank spaces
    to enforce alignment. */
std::string center(const std::string s, const int w) {
    std::stringstream ss, spaces;
    int padding = w - s.size();                 // count excess room to pad
    for(int i=0; i<padding/2; ++i)
        spaces << " ";
    ss << spaces.str() << s << spaces.str();    // format with padding
    if(padding>0 && padding%2!=0)               // if odd #, add 1 space
        ss << " ";
    return ss.str();
}

/* Convert double to string with specified number of places after the decimal
   and left padding. */
template<class type>
std::string prd(const type x, const int decDigits, const int width) {
    std::stringstream ss;
//     ss << std::fixed << std::right;
    ss << std::scientific << std::right;                     // Use scientific Output!
    ss.fill(' ');        // fill space around displayed #
    ss.width(width);     // set  width around displayed #
    ss.precision(decDigits); // set # places after decimal
    ss << x;
    return ss.str();
}

/**
 * @brief Infrastructure for handling periodicity.
 *         Check whether two points are equal on R/Z x R/Z x R
 */
auto equivalent = [](const FieldVector<double,3>& x, const FieldVector<double,3>& y)
                {
                    return ( (FloatCmp::eq(x[0],y[0]) or FloatCmp::eq(x[0]+1,y[0]) or FloatCmp::eq(x[0]-1,y[0]))
                            and (FloatCmp::eq(x[1],y[1]) or FloatCmp::eq(x[1]+1,y[1]) or FloatCmp::eq(x[1]-1,y[1]))
                            and (FloatCmp::eq(x[2],y[2]))
                        );
                };



////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
int main(int argc, char *argv[])
{
    feenableexcept(FE_INVALID);
    MPIHelper::instance(argc, argv);
    
    Dune::Timer globalTimer;

    if (argc < 3)
        DUNE_THROW(Exception, "Usage: ./Cell-Problem <python path> <python module without extension>");

    // Start Python interpreter
    Python::start();
    auto pyMain = Python::main();
    pyMain.runStream()
        << std::endl << "import math"
        << std::endl << "import sys"
        << std::endl << "sys.path.append('" << argv[1] << "')"  << std::endl;
    auto pyModule = pyMain.import(argv[2]);

    ParameterTree parameterSet;
    pyModule.get("parameterSet").toC(parameterSet);
    // read possible further parameters from the command line
    ParameterTreeParser::readOptions(argc, argv, parameterSet);
    // Print all parameters, to make them appear in the log file
    std::cout << "Input parameters:" << std::endl;
    parameterSet.report();

  //--- Output setter
  std::string baseName = parameterSet.get("baseName", "CellProblem-result");
  std::string resultPath = parameterSet.get("resultPath", "../../outputs") ;
  //--- setup Log-File
  std::fstream log;
  log.open(resultPath + "/" + baseName + "_log.txt" ,std::ios::out);

  constexpr int dim = 3;

  // Debug/Print Options 
  bool print_debug = parameterSet.get<bool>("print_debug", false);
  // VTK-write options
  // bool write_prestrainFunctions  = parameterSet.get<bool>("write_prestrainFunctions", false);  //Not implemented yet.

  /**
  * @brief Generate the grid.
  *        Corrector Problem Domain (-1/2,1/2)^3.
  */
//   FieldVector<double,dim> lower({-1.0/2.0, -1.0/2.0, -1.0/2.0});
//   FieldVector<double,dim> upper({1.0/2.0, 1.0/2.0, 1.0/2.0});

//   std::array<int,2> numLevels = parameterSet.get<std::array<int,2>>("numLevels", {3,3});
//   int levelCounter = 0;

Dune::FieldVector<double,dim> lower({-1.0/2.0, -1.0/2.0, -1.0/2.0});
Dune::FieldVector<double,dim> upper({1.0/2.0, 1.0/2.0, 1.0/2.0});
int microGridLevel = parameterSet.get<int>("microGridLevel", 1);
std::array<int, dim> nElements = {(int)std::pow(2,microGridLevel) ,(int)std::pow(2,microGridLevel) ,(int)std::pow(2,microGridLevel)};
// std::cout << "Number of Grid-Elements in each direction: " << nElements << std::endl;
std::cout << "Number of Grid-Elements in each direction: " << (int)std::pow(2,microGridLevel) << std::endl;
   
  
  ///////////////////////////////////
  // Create Data Storage
  ///////////////////////////////////
  //--- Storage:: #1 level #2 L2SymError #3 L2SymErrorOrder #4  L2Norm(sym) #5 L2Norm(sym-analytic) #6 L2Norm(phi_1)
    //std::vector<std::variant<std::string, size_t , double>> Storage_Error;
  //--- Storage:: | level | q1 | q2 | q3 | q12 | q13 | q23 | b1 | b2 | b3 |           
//   std::vector<std::variant<std::string, size_t , double>> Storage_Quantities;         

  //--- GridLevel-Loop:
//   for(size_t level = numLevels[0] ; level <= numLevels[1]; level++)     
//   {
//     std::cout << " ----------------------------------" << std::endl;
//     std::cout << "GridLevel: " << level << std::endl;
//     std::cout << " ----------------------------------" << std::endl;

//     // Storage_Error.push_back(level);
//     Storage_Quantities.push_back(level);
//     std::array<int, dim> nElements = {(int)std::pow(2,level) ,(int)std::pow(2,level) ,(int)std::pow(2,level)};
//     std::cout << "Number of Grid-Elements in each direction: " << nElements << std::endl;
//     log << "Number of Grid-Elements in each direction: " << nElements << std::endl;

    using CellGridType = YaspGrid<dim, EquidistantOffsetCoordinates<double, dim> >;
    CellGridType grid_CE(lower,upper,nElements);
    using GridView = CellGridType::LeafGridView;
    const GridView gridView_CE = grid_CE.leafGridView();
    if(print_debug)
       std::cout << "Host grid has " << gridView_CE.size(dim) << " vertices." << std::endl;

    //--- Choose a finite element space for Cell Problem
    using namespace Functions::BasisFactory;
    Functions::BasisFactory::Experimental::PeriodicIndexSet periodicIndices;

    //--- Get PeriodicIndices for periodicBasis (Don't do the following in real life: It has quadratic run-time in the number of vertices.)
    for (const auto& v1 : vertices(gridView_CE))
        for (const auto& v2 : vertices(gridView_CE))
            if (equivalent(v1.geometry().corner(0), v2.geometry().corner(0)))
            {
                periodicIndices.unifyIndexPair({gridView_CE.indexSet().index(v1)}, {gridView_CE.indexSet().index(v2)});
            }

    Dune::Timer basisSetupTimer;
    //--- Setup first order periodic Lagrange-Basis
    auto Basis_CE = makeBasis(
        gridView_CE,
        power<dim>(                                  
        Functions::BasisFactory::Experimental::periodic(lagrange<1>(), periodicIndices),
        flatLexicographic()
        //blockedInterleaved()   // Not Implemented
        ));     
    std::cout << "Basis setup took " << basisSetupTimer.elapsed() << " seconds " << std::endl;

    if(print_debug)
       std::cout << "power<periodic> basis has " << Basis_CE.dimension() << " degrees of freedom" << std::endl;



    // -------------------------------------------------------
    // This needs to be declared as Callable otherwise Class(param) wont work. if _call__(self,x) is not implemented
    Python::Callable MicrostructureClass  = pyModule.get("Microstructure");

    // Create instance of class
    // This needs to be a 'Python::Reference' and not 'Python::Callable' since Callable only works if __call__(self,x) is implemented in the python module!!!
    // Python::Reference microstructure = MicrostructureClass(2);

    //Setup for a constant microstructure.
    Python::Reference microstructure = MicrostructureClass();

    ///////////////////////////////////
    //  Create prestrained material object
    ///////////////////////////////////
    Dune::Timer materialSetupTimer;

    using MaterialType = prestrainedMaterial<GridView>;

    // auto material = prestrainedMaterial(gridView_CE,parameterSet,pyModule);
    // auto material = prestrainedMaterial(gridView_CE,microstructure,parameterSet,pyModule);
    std::shared_ptr<MaterialType> material = std::make_shared<MaterialType>(gridView_CE,microstructure,parameterSet,pyModule);
    std::cout << "Material setup took " << materialSetupTimer.elapsed() << " seconds " << std::endl;

    // --- Get scale ratio 
    // double gamma = parameterSet.get<double>("gamma",1.0); 
    std::cout << "scale ratio (gamma) set to : " << material->gamma_ << std::endl;

    //--- Compute Correctors
    // auto correctorComputer = CorrectorComputer(Basis_CE, material, log, parameterSet);
    // auto correctorComputer = CorrectorComputer(Basis_CE, materialPointer, log, parameterSet);
    std::shared_ptr<CorrectorComputer<decltype(Basis_CE),MaterialType> > correctorComputer = std::make_shared<CorrectorComputer<decltype(Basis_CE),MaterialType>>(Basis_CE, material, parameterSet);
    correctorComputer->assemble();
    correctorComputer->solve();

    //--- Check Correctors (options):
    if(parameterSet.get<bool>("write_L2Error", false))
         correctorComputer->computeNorms();
    if(parameterSet.get<bool>("write_VTK", false))
         correctorComputer->writeCorrectorsVTK(microGridLevel);
    //--- Additional Test: check orthogonality (75) from paper:
    if(parameterSet.get<bool>("write_checkOrthogonality", false))
        correctorComputer->check_Orthogonality();
    //--- Check symmetry of stiffness matrix
    if(print_debug)
        correctorComputer->checkSymmetry();

    //--- Compute effective quantities
    // auto effectiveQuantitiesComputer = EffectiveQuantitiesComputer(correctorComputer,material);
    auto effectiveQuantitiesComputer = EffectiveQuantitiesComputer(correctorComputer);
    effectiveQuantitiesComputer.computeEffectiveQuantities();

    //--- Write material indicator function to VTK
    if (parameterSet.get<bool>("write_materialFunctions", false))
        material->writeVTKMaterialFunctions(microGridLevel);

    //--- Get effective quantities
    auto Qeff = effectiveQuantitiesComputer.getQeff();
    auto Beff = effectiveQuantitiesComputer.getBeff();
    printmatrix(std::cout, Qeff, "Matrix Qeff", "--");
    printvector(std::cout, Beff, "Beff", "--");


    //--- Write effective quantities to matlab folder (for symbolic minimization)
    if(parameterSet.get<bool>("write_EffectiveQuantitiesToTxt", true))
        effectiveQuantitiesComputer.writeEffectiveQuantitiesToTxt(resultPath);




    std::cout << "Total time elapsed: " << globalTimer.elapsed() << std::endl;
}
