#include <filesystem>
#include <config.h>
#include <signal.h>
#include <memory>

#include <fenv.h>
#include <array>

#include <math.h>

// Includes for the ADOL-C automatic differentiation library
// Need to come before (almost) all others.
#include <adolc/adouble.h>
#include <dune/fufem/utilities/adolcnamespaceinjections.hh>

#include <dune/common/typetraits.hh>

#include <dune/common/bitsetvector.hh>
#include <dune/common/parametertree.hh>
#include <dune/common/parametertreeparser.hh>

#include <dune/common/timer.hh>

#include <dune/grid/uggrid.hh>
#include <dune/grid/utility/structuredgridfactory.hh>

#include <dune/grid/io/file/gmshreader.hh>
#include <dune/grid/io/file/vtk.hh>

#include <dune/functions/gridfunctions/discreteglobalbasisfunction.hh>
#include <dune/functions/functionspacebases/lagrangebasis.hh>
#include <dune/functions/functionspacebases/powerbasis.hh>
#include <dune/functions/functionspacebases/interpolate.hh>
#include <dune/functions/functionspacebases/reducedcubichermitetrianglebasis.hh>

#include <dune/fufem/boundarypatch.hh>
#include <dune/fufem/functiontools/boundarydofs.hh>
#include <dune/fufem/dunepython.hh>
#include <dune/fufem/discretizationerror.hh>

#include <dune/solvers/solvers/iterativesolver.hh>
#include <dune/solvers/norms/energynorm.hh>

#include <dune/gfe/spaces/productmanifold.hh>
#include <dune/gfe/spaces/realtuple.hh>
#include <dune/gfe/spaces/rotation.hh>
#include <dune/gfe/bendingisometryhelper.hh>
#include <dune/gfe/localdiscretekirchhoffbendingisometry.hh>
#include <dune/gfe/assemblers/localgeodesicfeadolcstiffness.hh>
// #include <dune/gfe/assemblers/harmonicenergy.hh>
#include <dune/gfe/assemblers/geodesicfeassembler.hh>
// #include <dune/gfe/riemannianpnsolver.hh>
#include <dune/gfe/embeddedglobalgfefunction.hh>


#include <dune/gmsh4/gmsh4reader.hh>
#include <dune/gmsh4/gridcreators/lagrangegridcreator.hh>

const int dim = 2;

using namespace Dune;


/**
 * @brief Test the interpolation in deformation space.
 *        A linear isometry can be represented exactly in that space.
 *        compare energy values / L2-error / H1-error 
 *        for die difference | I_h[u] - u | where u is a given analytic linear isometry
 */


int main(int argc, char *argv[])
{
  MPIHelper::instance(argc, argv);
  Dune::Timer globalTimer;

  std::cout << "Current path is " << std::filesystem::current_path() << '\n';
  std::filesystem::path file_path = (__FILE__);
  std::cout<< "File path: " << file_path<<std::endl;
  std::cout << "dir_path: " << file_path.parent_path() << std::endl;
  std::string dir_path  = file_path.parent_path();

  // Start Python interpreter
  Python::start();
  auto pyMain = Python::main();
  pyMain.runStream()
      << std::endl << "import math"
      << std::endl << "import sys"
      << std::endl << "sys.path.append('" << dir_path << "')"  << std::endl;
  auto pyModule = pyMain.import("hermitebasis-evaluation-test");

  // parse data file
  ParameterTree parameterSet;
  pyModule.get("parameterSet").toC(parameterSet);

  // read possible further parameters from the command line
  ParameterTreeParser::readOptions(argc, argv, parameterSet);

  // Print all parameters, to make them appear in the log file
  std::cout << "Executable: bending-isometries, with parameters:" << std::endl;
  parameterSet.report();

  /////////////////////////////////////////
  //   Create the grid
  /////////////////////////////////////////
  using GridType = UGGrid<dim>;
  std::shared_ptr<GridType> grid;
//   FieldVector<double,dim> lower(0), upper(1);
//   std::array<unsigned int,dim> elementsArray;

  // create unstructured grid.
  std::cout << "Read GMSH grid." << std::endl;
  GridFactory<GridType> factory;
  Gmsh4::LagrangeGridCreator creator{factory};
  Gmsh4Reader reader{creator};
  reader.read(dir_path + "/" + std::string("UnstructuredGrid.msh"));
  grid = factory.createGrid();


  const int numLevels = parameterSet.get<int>("numLevels");
  grid->globalRefine(numLevels-1);

  using GridView = typename GridType::LeafGridView;
  GridView gridView = grid->leafGridView();

  ///////////////////////////////////////////////////////
  //  Set up the function space and the energy
  ///////////////////////////////////////////////////////

  // General coefficient vector of a Discrete Kirchhoff deformation function
  using VectorSpaceCoefficients = BlockVector<FieldVector<double,3>>;

  using namespace Functions::BasisFactory;
  auto deformationBasis = makeBasis(gridView,
                                    power<3>(reducedCubicHermiteTriangle(),
                                             blockedInterleaved()));


  ///////////////////////////////////////////
  //   Get initial Iterate
  ///////////////////////////////////////////
  auto pythonInitialIterate = Python::makeDifferentiableFunction<FieldVector<double,3>(FieldVector<double,2>)>(pyModule.get("f"), pyModule.get("df"));

  VectorSpaceCoefficients x(deformationBasis.size());
  interpolate(deformationBasis, x, pythonInitialIterate);   // Interpolation on a python function


  std::cout << "Number of Elements in the grid:" << gridView.size(0)<< std::endl;
  std::cout << "Number of Nodes in the grid:" << gridView.size(dim)<< std::endl;
  std::cout << "deformationBasis.size():" << deformationBasis.size() << std::endl;
  std::cout << "deformationBasis.dimension():" << deformationBasis.dimension() << std::endl;


  ////////////////////////////////
  //   Output result
  ////////////////////////////////
  std::string baseNameDefault = "bending-isometries-";
  std::string baseName = parameterSet.get("baseName", baseNameDefault);
  std::string resultFileName =  parameterSet.get("resultPath", "")
                                + "/" + baseName
                                + "_level" + std::to_string(parameterSet.get<int>("numLevels"));

  if (parameterSet.get<bool>("conforming_DiscreteJacobian", 1))
    resultFileName = resultFileName + "_C";
  else 
    resultFileName = resultFileName + "_NC";

  /**
    * @brief Compute the displacement from the deformation.
    *        interpolate the identity and substract from the coefficient vector.
    */
  VectorSpaceCoefficients identity(deformationBasis.size());
  IdentityGridEmbedding<double> identityGridEmbedding;
  interpolate(deformationBasis, identity, identityGridEmbedding);
  // Compute the displacement
  auto displacement = x;
  displacement -= identity;

  auto deformationFunction = Functions::makeDiscreteGlobalBasisFunction<FieldVector<double,3>>(deformationBasis, x);
  auto displacementFunction = Functions::makeDiscreteGlobalBasisFunction<FieldVector<double,3>>(deformationBasis, displacement);
  auto deformationFunctionLocal = localFunction(deformationFunction);
  auto deformationFunctionLocalDerivative = derivative(deformationFunctionLocal);


  if (parameterSet.get<bool>("writeVTK", 1))
  {
    std::cout << "write VTK..." << std::endl;


    /**
     * @brief  We need to subsample, because VTK cannot natively display real third-order functions
     */
    int subsamplingRefinement = parameterSet.get<int>("subsamplingRefinement", 2);
    SubsamplingVTKWriter<GridView> vtkWriter(gridView, Dune::refinementLevels(subsamplingRefinement));



    /**
     * @brief Interpolate and VTK-write analytical solution "u" from ParamterFile.
     */
    if (parameterSet.get<bool>("vtkwrite_analyticalSolution", 0))
    {
      auto pythonAnalyticalSolution = Python::makeDifferentiableFunction<FieldVector<double,3>(FieldVector<double,2>)>(pyModule.get("displacement"), pyModule.get("displacementGradient"));
      vtkWriter.addVertexData((pythonAnalyticalSolution), VTK::FieldInfo("displacement_analytical", VTK::FieldInfo::Type::vector, 3));
    }
    vtkWriter.addVertexData(displacementFunction, VTK::FieldInfo("displacement", VTK::FieldInfo::Type::vector, 3));
    vtkWriter.write(resultFileName);
  }




  /**
   * @brief Compute the L2-Error and H1-SemiError between the discrete deformation u_h and the analytical deformation u 
   * 
   */
    auto referenceSolution =  Python::makeDifferentiableFunction<FieldVector<double,3>(FieldVector<double,2>)>(pyModule.get("deformation"), pyModule.get("deformationGradient"));
    auto referenceDerivative = derivative(referenceSolution);


    // QuadratureRule for the integral of the L^2 error
    QuadratureRuleKey quadKey(dim,6);

    // The error to be computed
    double l2ErrorSquared = 0;
    double h1ErrorSquared = 0;

    for (auto&& element : elements(gridView))
    {

      deformationFunctionLocal.bind(element);
      deformationFunctionLocalDerivative.bind(element);

      // Get quadrature formula
      quadKey.setGeometryType(element.type());
      const auto& quad = QuadratureRuleCache<double, dim>::rule(quadKey);

      for (auto quadPoint : quad)
      {
        auto quadPos = quadPoint.position();
        const auto integrationElement = element.geometry().integrationElement(quadPos);
        const auto weight = quadPoint.weight();

        auto refValue = referenceSolution(element.geometry().global(quadPos));
        auto numValue = deformationFunctionLocal(quadPos);

        auto numDir = deformationFunctionLocalDerivative(quadPos);
        auto refDir = referenceDerivative(element.geometry().global(quadPos));
        auto diff = numDir - refDir;

        // // print evaluation difference
        // std::cout << "Evaluation difference at quadPoint:(" << element.geometry().global(quadPos) << ") is : "<< numValue-refValue << std::endl;
        std::cout << "Evaluation difference (of derivative) at quadPoint:(" << element.geometry().global(quadPos) << ") is : "<< diff << std::endl;
        // print discrete function evaluation
        // std::cout << "Evaluation of discreteGlobalBasisFunction at quadPoint:(" << element.geometry().global(quadPos) << ") is : "<< numValue << std::endl;
        std::cout << "Evaluation of discreteGlobalBasisFunction-derivative at quadPoint:(" << element.geometry().global(quadPos) << ") is : "<< numDir << std::endl;


        // integrate error
        l2ErrorSquared += (numValue - refValue)* (numValue - refValue) * weight * integrationElement;
        h1ErrorSquared += diff.frobenius_norm2() * weight * integrationElement;
      }
    }
    std::cout << "elements: " << gridView.size(0)
              << "      "
              << "L^2 error: " << std::sqrt(l2ErrorSquared)
              << "      ";
    std::cout << "h^1 error: " << std::sqrt(h1ErrorSquared) << std::endl;

   
    //Check Quantities
    if( std::sqrt(l2ErrorSquared) > 1e-12)
    {
        std::cerr << std::setprecision(9);
        std::cerr << "L2-error is to large! " << std::endl;
        return 1;
    }
    if( std::sqrt(h1ErrorSquared) > 1e-12)
    {
        std::cerr << std::setprecision(9);
        std::cerr << "H1-error is to large! " << std::endl;
        return 1;
    }
  std::cout << "Total time elapsed: " << globalTimer.elapsed() << std::endl;
  std::cout << "Test: hermitebasis-evaluation-test passed." << std::endl;
  return 0;
  }