#include <config.h>
#include <array>
#include <vector>
#include <fstream>

#include <iostream>
#include <dune/common/indices.hh>
#include <dune/common/bitsetvector.hh>
#include <dune/common/parametertree.hh>
#include <dune/common/parametertreeparser.hh>
#include <dune/common/float_cmp.hh>
#include <dune/common/math.hh>

#include <dune/geometry/quadraturerules.hh>

#include <dune/grid/uggrid.hh>
#include <dune/grid/yaspgrid.hh>
#include <dune/grid/io/file/vtk/subsamplingvtkwriter.hh>

#include <dune/istl/matrix.hh>
#include <dune/istl/bcrsmatrix.hh>
#include <dune/istl/multitypeblockmatrix.hh>
#include <dune/istl/multitypeblockvector.hh>
#include <dune/istl/matrixindexset.hh>
#include <dune/istl/solvers.hh>
#include <dune/istl/spqr.hh>
#include <dune/istl/preconditioners.hh>
#include <dune/istl/io.hh>

#include <dune/functions/functionspacebases/interpolate.hh>
#include <dune/functions/backends/istlvectorbackend.hh>
#include <dune/functions/functionspacebases/powerbasis.hh>
#include <dune/functions/functionspacebases/compositebasis.hh>
#include <dune/functions/functionspacebases/lagrangebasis.hh>
#include <dune/functions/functionspacebases/periodicbasis.hh>
#include <dune/functions/functionspacebases/subspacebasis.hh>
#include <dune/functions/functionspacebases/boundarydofs.hh>
#include <dune/functions/gridfunctions/discreteglobalbasisfunction.hh>
#include <dune/functions/gridfunctions/gridviewfunction.hh>

#include <dune/common/fvector.hh>
#include <dune/common/fmatrix.hh>

#include <dune/microstructure/matrix_operations.hh>
#include <dune/microstructure/CorrectorComputer.hh>
#include <dune/microstructure/EffectiveQuantitiesComputer.hh>
#include <dune/microstructure/prestrainedMaterial.hh>

#include <dune/solvers/solvers/umfpacksolver.hh>
#include <dune/istl/eigenvalue/test/matrixinfo.hh>

#include <dune/fufem/dunepython.hh>

#include <any>
#include <variant>
#include <string>
#include <iomanip>
#include <filesystem>

using namespace Dune;
using namespace MatrixOperations;


/**
 * @brief  Compute effective quantities for a trilayer setup with 3 orthotropic phases:
 *         Once with a canonical orthonormal material frame and once with a material frame 
 *         that is rotated pi(180 degree) around different axes.
 *         Aim of the test is to check the correctness of the material frame transformation.   
 */


//////////////////////////////////////////////////
//   Infrastructure for handling periodicity
//////////////////////////////////////////////////
// Check whether two points are equal on R/Z x R/Z x R
auto equivalent = [](const FieldVector<double,3>& x, const FieldVector<double,3>& y)
                  {
                    return ( (FloatCmp::eq(x[0],y[0]) or FloatCmp::eq(x[0]+1,y[0]) or FloatCmp::eq(x[0]-1,y[0]))
                             and (FloatCmp::eq(x[1],y[1]) or FloatCmp::eq(x[1]+1,y[1]) or FloatCmp::eq(x[1]-1,y[1]))
                             and (FloatCmp::eq(x[2],y[2]))
                             );
                  };


int main(int argc, char *argv[])
{
  MPIHelper::instance(argc, argv);

  Dune::Timer globalTimer;
  //--- setup Log-File 
  std::fstream log;

  std::cout << "Current path is " << std::filesystem::current_path() << '\n';
  std::filesystem::path file_path = (__FILE__);
  std::cout<< "File path: " << file_path<<std::endl;
  std::cout << "dir_path: " << file_path.parent_path() << std::endl;
  std::string dir_path  = file_path.parent_path();

  // ParameterTree parameterSet;
  // if (argc < 2)
  //   ParameterTreeParser::readINITree(dir_path + "/orthotropicrotation_1.parset", parameterSet);
  // else
  // {
  //   ParameterTreeParser::readINITree(argv[1], parameterSet);
  //   ParameterTreeParser::readOptions(argc, argv, parameterSet);
  // }

  // //--- Start Python interpreter
  // Python::start();
  // Python::Reference main = Python::import("__main__");
  // Python::run("import math");
  // Python::runStream()
  //   << std::endl << "import sys"
  //   << std::endl << "sys.path.append('" << dir_path  << "')"
  //   << std::endl;


   // Start Python interpreter
    Python::start();
    auto pyMain = Python::main();
    pyMain.runStream()
        << std::endl << "import math"
        << std::endl << "import sys"
        << std::endl << "sys.path.append('" << dir_path << "')"  << std::endl;
    auto pyModule = pyMain.import("orthotropicrotation_1");

    ParameterTree parameterSet;
    pyModule.get("parameterSet").toC(parameterSet);
    // read possible further parameters from the command line
    ParameterTreeParser::readOptions(argc, argv, parameterSet);
    // Print all parameters, to make them appear in the log file
    std::cout << "Input parameters:" << std::endl;
    parameterSet.report();



  constexpr int dim = 3;
  constexpr int dimWorld = 3;
  ///////////////////////////////////
  // Generate the grid
  ///////////////////////////////////
  // --- Corrector Problem Domain (-1/2,1/2)^3:
  FieldVector<double,dim> lower({-1.0/2.0, -1.0/2.0, -1.0/2.0});
  FieldVector<double,dim> upper({1.0/2.0, 1.0/2.0, 1.0/2.0});

  int gridLevel = 3;

  std::array<int, dim> nElements = {(int)std::pow(2,gridLevel) ,(int)std::pow(2,gridLevel) ,(int)std::pow(2,gridLevel)};
  std::cout << "Number of Grid-Elements in each direction: " << nElements << std::endl;

  using CellGridType = YaspGrid<dim, EquidistantOffsetCoordinates<double, dim> >;
  CellGridType grid_CE(lower,upper,nElements);
  using GridView = CellGridType::LeafGridView;
  const GridView gridView_CE = grid_CE.leafGridView();

  //--- Choose a finite element space for Cell Problem
  using namespace Functions::BasisFactory;
  Functions::BasisFactory::Experimental::PeriodicIndexSet periodicIndices;

  //--- get PeriodicIndices for periodicBasis (Don't do the following in real life: It has quadratic run-time in the number of vertices.)
  for (const auto& v1 : vertices(gridView_CE))
    for (const auto& v2 : vertices(gridView_CE))
      if (equivalent(v1.geometry().corner(0), v2.geometry().corner(0)))
      {
        periodicIndices.unifyIndexPair({gridView_CE.indexSet().index(v1)}, {gridView_CE.indexSet().index(v2)});
      }

  //--- setup first order periodic Lagrange-Basis
  auto Basis_CE = makeBasis(
    gridView_CE,
    power<dim>(
      Functions::BasisFactory::Experimental::periodic(lagrange<1>(), periodicIndices),
      flatLexicographic()
      ));


  // Create Microstructure object.
  Python::Callable MicrostructureClass  = pyModule.get("Microstructure");
  Python::Reference microstructure = MicrostructureClass(); //Setup a constant microstructure

  ///////////////////////////////////
  //  Create prestrained material object
  ///////////////////////////////////
  using MaterialType = prestrainedMaterial<GridView>;
  std::shared_ptr<MaterialType> material = std::make_shared<MaterialType>(gridView_CE,microstructure,parameterSet,pyModule);
  // auto material_ = prestrainedMaterial(gridView_CE,microstructure,parameterSet,pyModule);

  // --- Get scale ratio
  // double gamma = parameterSet.get<double>("gamma",1.0);

  //------------------------------------------------------------------------------------------------
  //--- Compute Correctors
  // auto correctorComputer = CorrectorComputer(Basis_CE, material_, log, parameterSet);
  // correctorComputer.assemble();
  // correctorComputer.solve();
  std::shared_ptr<CorrectorComputer<decltype(Basis_CE),MaterialType> > correctorComputer = std::make_shared<CorrectorComputer<decltype(Basis_CE),MaterialType>>(Basis_CE, material, parameterSet);
  correctorComputer->assemble();
  correctorComputer->solve();

  //--- Compute effective quantities
  auto effectiveQuantitiesComputer = EffectiveQuantitiesComputer(correctorComputer);
  effectiveQuantitiesComputer.computeEffectiveQuantities();

  //--- get effective quantities
  auto Qeff = effectiveQuantitiesComputer.getQeff();
  auto Beff = effectiveQuantitiesComputer.getBeff();
//   printmatrix(std::cout, Qeff, "Matrix Qeff", "--");
//   printvector(std::cout, Beff, "Beff", "--");


  /**
   * @brief Compute "rotated" effective quantitites
   *
   */
//  ParameterTreeParser::readINITree("../test/orthotropicrotation_2.parset", parameterSet);
//  ParameterTreeParser::readINITree(dir_path + "/orthotropicrotation_2.parset", parameterSet);
  pyModule = pyMain.import("orthotropicrotation_2");
  pyModule.get("parameterSet").toC(parameterSet);


  Python::Callable MicrostructureClass_2  = pyModule.get("Microstructure");
  Python::Reference microstructure_2 = MicrostructureClass_2(); //Setup a constant microstructure




  //-- Alternative: 
  material->updateMicrostructure(microstructure_2);
  correctorComputer->updateMaterial(material);
  effectiveQuantitiesComputer.updateCorrectorComputer(correctorComputer);

  correctorComputer->assemble();
  correctorComputer->solve();
  effectiveQuantitiesComputer.computeEffectiveQuantities();
  auto Qeff_2 = effectiveQuantitiesComputer.getQeff();
  auto Beff_2 = effectiveQuantitiesComputer.getBeff();
  // correctorComputer->updateMaterial(material_2);
  // effectiveQuantitiesComputer.updateCorrectorComputer(correctorComputer);



  /**
   * @brief Old way:
   * 
   */
  // auto material_2 = prestrainedMaterial(gridView_CE,microstructure_2,parameterSet,pyModule);
  // std::shared_ptr<MaterialType> material_2 = std::make_shared<MaterialType>(gridView_CE,microstructure_2,parameterSet,pyModule);


  //--- Compute Correctors
  // auto correctorComputer_2 = CorrectorComputer(Basis_CE, material_2, log, parameterSet);
  // correctorComputer_2.assemble();
  // correctorComputer_2.solve();


  // std::shared_ptr<CorrectorComputer<decltype(Basis_CE),MaterialType> > correctorComputer_2 = std::make_shared<CorrectorComputer<decltype(Basis_CE),MaterialType>>(Basis_CE, material_2, parameterSet);
  // correctorComputer_2->assemble();
  // correctorComputer_2->solve();

  // //--- Compute effective quantities
  // auto effectiveQuantitiesComputer_2 = EffectiveQuantitiesComputer(correctorComputer_2);
  // effectiveQuantitiesComputer_2.computeEffectiveQuantities();

  // //--- get effective quantities
  // auto Qeff_2 = effectiveQuantitiesComputer_2.getQeff();
  // auto Beff_2 = effectiveQuantitiesComputer_2.getBeff();

  /**
   * @brief  Compare results.
   *
   */
  if( std::abs(Qeff[0][0] - Qeff_2[0][0]) > 1e-6)
  {
    std::cerr << std::setprecision(9);
    std::cerr << "Computed Qeff[0][0] is " << Qeff[0][0] << " but with rotated material frame we get : " << Qeff_2[0][0] << " !" << std::endl;
    return 1;
  }
  if( std::abs(Qeff[1][0] - Qeff_2[1][0]) > 1e-6)
  {
    std::cerr << std::setprecision(9);
    std::cerr << "Computed Qeff[1][0] is " << Qeff[1][0] << " but with rotated material frame we get : " << Qeff_2[1][0] << " !" << std::endl;
    return 1;
  }
  if( std::abs(Qeff[2][0] - Qeff_2[2][0]) > 1e-6)
  {
    std::cerr << std::setprecision(9);
    std::cerr << "Computed Qeff[2][0] is " << Qeff[2][0] << " but with rotated material frame we get : " << Qeff_2[2][0] << " !" << std::endl;
    return 1;
  }
  if( std::abs(Qeff[0][1] - Qeff_2[0][1]) > 1e-6)
  {
    std::cerr << std::setprecision(9);
    std::cerr << "Computed Qeff[0][1] is " << Qeff[0][1] << " but with rotated material frame we get : " << Qeff_2[0][1] << " !" << std::endl;
    return 1;
  }
  if( std::abs(Qeff[1][1] - Qeff_2[1][1]) > 1e-6)
  {
    std::cerr << std::setprecision(9);
    std::cerr << "Computed Qeff[1][1] is " << Qeff[1][1] << " but with rotated material frame we get : " << Qeff_2[1][1] << " !" << std::endl;
    return 1;
  }
  if( std::abs(Qeff[2][1] - Qeff_2[2][1]) > 1e-6)
  {
    std::cerr << std::setprecision(9);
    std::cerr << "Computed Qeff[2][1] is " << Qeff[2][1] << " but with rotated material frame we get : " << Qeff_2[2][1] << " !" << std::endl;
    return 1;
  }
  if( std::abs(Qeff[0][2] - Qeff_2[0][2]) > 1e-6)
  {
    std::cerr << std::setprecision(9);
    std::cerr << "Computed Qeff[0][2] is " << Qeff[0][2] << " but with rotated material frame we get : " << Qeff_2[0][2] << " !" << std::endl;
    return 1;
  }
  if( std::abs(Qeff[1][2] - Qeff_2[1][2]) > 1e-6)
  {
    std::cerr << std::setprecision(9);
    std::cerr << "Computed Qeff[1][2] is " << Qeff[1][2] << " but with rotated material frame we get : " << Qeff_2[1][2] << " !" << std::endl;
    return 1;
  }
  if( std::abs(Qeff[2][2] - Qeff_2[2][2]) > 1e-6)
  {
    std::cerr << std::setprecision(9);
    std::cerr << "Computed Qeff[2][2] is " << Qeff[2][2] << " but with rotated material frame we get : " << Qeff_2[2][2] << " !" << std::endl;
    return 1;
  }


  if( std::abs(Beff[0] - Beff_2[0]) > 1e-6)
  {
    std::cerr << std::setprecision(9);
    std::cerr << "Computed Beff[0] is " << Beff[0] << " but with rotated material frame we get : " << Beff_2[0] << " !" << std::endl;
    return 1;
  }
  if( std::abs(Beff[1] - Beff_2[1]) > 1e-6)
  {
    std::cerr << std::setprecision(9);
    std::cerr << "Computed Beff[1] is " << Beff[1] << " but with rotated material frame we get : " << Beff_2[1] << " !" << std::endl;
    return 1;
  }
  if( std::abs(Beff[2] - Beff_2[2]) > 1e-6)
  {
    std::cerr << std::setprecision(9);
    std::cerr << "Computed Beff[2] is " << Beff[2] << " but with rotated material frame we get : " << Beff_2[2] << " !" << std::endl;
    return 1;
  }

  std::cout << "Total time elapsed: " << globalTimer.elapsed() << std::endl;
  std::cout << "Test: orthotropicrotationtest passed." << std::endl;
}
