#include <config.h>
#include <signal.h>
#include <memory>

#include <fenv.h>
#include <array>

#include <math.h>

// Includes for the ADOL-C automatic differentiation library
// Need to come before (almost) all others.
#include <adolc/adouble.h>
#include <dune/fufem/utilities/adolcnamespaceinjections.hh>

#include <dune/common/typetraits.hh>

#include <dune/common/bitsetvector.hh>
#include <dune/common/parametertree.hh>
#include <dune/common/parametertreeparser.hh>

#include <dune/grid/uggrid.hh>
#include <dune/grid/utility/structuredgridfactory.hh>

#include <dune/grid/io/file/gmshreader.hh>
#include <dune/grid/io/file/vtk.hh>

#include <dune/functions/gridfunctions/discreteglobalbasisfunction.hh>
#include <dune/functions/functionspacebases/lagrangebasis.hh>
#include <dune/functions/functionspacebases/powerbasis.hh>
#include <dune/functions/functionspacebases/interpolate.hh>
#include <dune/functions/functionspacebases/reducedcubichermitetrianglebasis.hh>

#include <dune/fufem/boundarypatch.hh>
#include <dune/fufem/functiontools/boundarydofs.hh>
#include <dune/fufem/dunepython.hh>
#include <dune/fufem/discretizationerror.hh>

#include <dune/solvers/solvers/iterativesolver.hh>
#include <dune/solvers/norms/energynorm.hh>

#include <dune/gfe/spaces/productmanifold.hh>
#include <dune/gfe/spaces/realtuple.hh>
#include <dune/gfe/spaces/rotation.hh>
#include <dune/gfe/localdiscretekirchhoffbendingisometry.hh>
#include <dune/gfe/assemblers/localgeodesicfeadolcstiffness.hh>
#include <dune/gfe/assemblers/harmonicenergy.hh>
#include <dune/gfe/assemblers/geodesicfeassembler.hh>
#include <dune/gfe/bendingisometryhelper.hh>
#include <dune/gfe/riemannianpnsolver.hh>
#include <dune/gfe/embeddedglobalgfefunction.hh>
#include <dune/gfe/energies/discretekirchhoffbendingenergy.hh>
#include <dune/gfe/energies/discretekirchhoffbendingenergyconforming.hh>
#include <dune/gfe/energies/discretekirchhoffbendingenergyzienkiewicz.hh>
#include <dune/gfe/energies/discretekirchhoffbendingenergyzienkiewiczprojected.hh>
// #include <dune/gfe/energies/discretekirchhoffbendingenergyprestrained.hh>

#include <dune/microstructure/energies/discretekirchhoffbendingenergyprestrained.hh>

#include <dune/gfe/spaces/stiefelmanifold.hh>
// #include <dune/gfe/spaces/stiefelmatrix.hh>

#include <dune/gmsh4/gmsh4reader.hh>
#include <dune/gmsh4/gridcreators/lagrangegridcreator.hh>


#include <filesystem>

const int dim = 2;

using namespace Dune;


int main(int argc, char *argv[])
{

  /**
   * @brief We use this to catch a 'Floating point exception' caused by the 'innerSolver' 
   *        in RiemannianProximalNewton
   */
  //   std::shared_ptr<void(int)> handler(
  //       signal(SIGFPE, [](int signum) {throw std::logic_error("FPE"); }),
  //       [](__sighandler_t f) { signal(SIGFPE, f); });



  // feenableexcept(FE_INVALID);

  std::vector<double> Energy_threshold       = {1.0,   0.5, 2.0  };
  std::vector<double> L2_threshold           = {0.1,  0.05, 0.25 };
  std::vector<double> H1_threshold           = {0.15,  0.1, 0.5  };
  std::vector<double> Isometry_threshold     = {0.05, 0.05, 0.05 };


  for(size_t experimentNumber = 1; experimentNumber<4; experimentNumber++)
  {

      MPIHelper::instance(argc, argv);
      Dune::Timer globalTimer;

      // if (argc < 3)
      //   DUNE_THROW(Exception, "Usage: ./bending-isometries <python path> <python module without extension>");

      // Start Python interpreter
      // Python::start();
      // auto pyMain = Python::main();
      // pyMain.runStream()
      //     << std::endl << "import math"
      //     << std::endl << "import sys"
      //     << std::endl << "sys.path.append('" << argv[1] << "')"  << std::endl;
      // auto pyModule = pyMain.import(argv[2]);


      std::cout << "Current path is " << std::filesystem::current_path() << '\n';
      std::filesystem::path file_path = (__FILE__);
      std::cout<< "File path: " << file_path<<std::endl;
      std::cout << "dir_path: " << file_path.parent_path() << std::endl;
      std::string dir_path  = file_path.parent_path();


      std::string experiment = "properties-shape-test-" + std::to_string(experimentNumber);

      // Start Python interpreter
      Python::start();
      auto pyMain = Python::main();
      pyMain.runStream()
          << std::endl << "import math"
          << std::endl << "import sys"
          << std::endl << "sys.path.append('" << dir_path << "')"  << std::endl;
      // auto pyModule = pyMain.import("properties-shape-test-2");
      auto pyModule = pyMain.import(experiment);

      // parse data file
      ParameterTree parameterSet;
      pyModule.get("parameterSet").toC(parameterSet);

      // read possible further parameters from the command line
      ParameterTreeParser::readOptions(argc, argv, parameterSet);

      // Print all parameters, to make them appear in the log file
      std::cout << "Executable: bending-isometries, with parameters:" << std::endl;
      parameterSet.report();

      /////////////////////////////////////////
      //   Create the grid
      /////////////////////////////////////////
      using GridType = UGGrid<dim>;

      std::shared_ptr<GridType> grid;
      FieldVector<double,dim> lower(0), upper(1);
      std::array<unsigned int,dim> elementsArray;

      std::string structuredGridType = parameterSet["structuredGrid"];
      if (structuredGridType != "false" )
      {
        lower = parameterSet.get<FieldVector<double,dim> >("lower");
        upper = parameterSet.get<FieldVector<double,dim> >("upper");
        elementsArray = parameterSet.get<std::array<unsigned int,dim> >("elements");

        if (structuredGridType == "simplex")
          grid = StructuredGridFactory<GridType>::createSimplexGrid(lower, upper, elementsArray);
        else if (structuredGridType == "cube")
          grid = StructuredGridFactory<GridType>::createCubeGrid(lower, upper, elementsArray);
        else
          DUNE_THROW(Exception, "Unknown structured grid type '" << structuredGridType << "' found!");
      } else {
        std::cout << "Read GMSH grid." << std::endl;
        std::string gridPath                = parameterSet.get<std::string>("gridPath");
        std::string gridFile                = parameterSet.get<std::string>("gridFile");
        GridFactory<GridType> factory;
        Gmsh4::LagrangeGridCreator creator{factory};
        Gmsh4Reader reader{creator};
        reader.read(gridPath + "/" + gridFile);
        grid = factory.createGrid();
      }

      const int macroGridLevel = parameterSet.get<int>("macroGridLevel");
      grid->globalRefine(macroGridLevel-1);

      using GridView = typename GridType::LeafGridView;
      GridView gridView = grid->leafGridView();

      ///////////////////////////////////////////////////////
      //  Set up the function spaces
      ///////////////////////////////////////////////////////

      // General coefficient vector of a Discrete Kirchhoff deformation function
      using VectorSpaceCoefficients = BlockVector<FieldVector<double,3>>;

      /**
       * @brief Coefficient vector of a Discrete Kirchhoff deformation function that is constrained to be an isometry.
       *        we need both 'double' and 'adouble' versions.
       */
      using Coefficient = GFE::ProductManifold<RealTuple<double,3>, Rotation<double,3> >;
      using IsometryCoefficients = std::vector<Coefficient>;
      using ACoefficient = typename Coefficient::template rebind<adouble>::other;
      using AIsometryCoefficients = std::vector<ACoefficient>;

      using namespace Functions::BasisFactory;
      auto deformationBasis = makeBasis(gridView,
                                        power<3>(reducedCubicHermiteTriangle(),
                                        blockedInterleaved()));

      using DeformationBasis = decltype(deformationBasis);

      /**
       * @brief The next basis is used to assign (nonlinear) degrees of freedom to the grid vertices.
       *        The actual basis function values are never used.
       */
      using CoefficientBasis = Functions::LagrangeBasis<GridView, 1>;
      CoefficientBasis coefficientBasis(gridView);

      // A basis for the tangent space (used to set DirichletNodes)
      auto tangentBasis = makeBasis(gridView,
                                    power<Coefficient::TangentVector::dimension>(
                                    lagrange<1>(),
                                    blockedInterleaved()));

      // Print some information on the grid and degrees of freedom.                        
      std::cout << "Number of Elements in the grid: " << gridView.size(0)<< std::endl;
      std::cout << "Number of Nodes in the grid: "    << gridView.size(dim)<< std::endl;
      std::cout << "deformationBasis.size(): "        << deformationBasis.size() << std::endl;
      std::cout << "deformationBasis.dimension(): "   << deformationBasis.dimension() << std::endl;
      std::cout << "Degrees of Freedom: "             << deformationBasis.dimension() << std::endl;


      ///////////////////////////////////////////
      //   Read Dirichlet values   (NEW VERSION)
      ///////////////////////////////////////////
      BitSetVector<1> dirichletVertices(gridView.size(dim), false);
      const typename GridView::IndexSet &indexSet = gridView.indexSet();


      BitSetVector<Coefficient::TangentVector::dimension> dirichletNodes(tangentBasis.size(), false); //tangentBasis.size()=coefficientBasis.size()

      /**
       * @brief Make Python function that computes which vertices are on the Dirichlet boundary,
       *        based on the vertex positions.
       */
      auto dirichletIndicatorFunction = Python::make_function<bool>(pyModule.get("dirichlet_indicator"));



      /**
       * @brief If we want to clamp DOFs inside the domain, we connot use 'BoundaryPatch' 
       *        and 'constructBoundaryDofs'. This is a workaround for now.
       *        
       *
       */
      for (auto &&vertex : vertices(gridView))
      {
        dirichletVertices[indexSet.index(vertex)] = dirichletIndicatorFunction(vertex.geometry().corner(0));

          if(dirichletIndicatorFunction(vertex.geometry().corner(0)))
          { 
            dirichletNodes[indexSet.index(vertex)] = true;
          }
      
      }
        
      // std::cout << "tangentBasis.size():" << tangentBasis.size() << std::endl;
      // std::cout << "coefficientBasis.size():" << coefficientBasis.size() << std::endl;

      // // deprecated: 
      // // BoundaryPatch<GridView> dirichletBoundary(gridView, dirichletVertices);
      // // constructBoundaryDofs(dirichletBoundary, tangentBasis, dirichletNodes);

      //     // TEST: print dirichletNodes 
      // std::cout << "print dirichletVertices:" << std::endl;
      // std::cout <<  dirichletVertices << std::endl;

      // std::cout << "print dirichletNodes:" << std::endl;
      // std::cout <<  dirichletNodes << std::endl;



      // ///////////////////////////////////////////
      // //   Read Dirichlet values  (OLD VERSION)
      // ///////////////////////////////////////////
      // BitSetVector<1> dirichletVertices(gridView.size(dim), false);
      // const typename GridView::IndexSet &indexSet = gridView.indexSet();

      // /**
      //  * @brief Make Python function that computes which vertices are on the Dirichlet boundary,
      //  *        based on the vertex positions.
      //  */
      // auto dirichletIndicatorFunction = Python::make_function<bool>(pyModule.get("dirichlet_indicator"));

      // for (auto &&vertex : vertices(gridView))
      // {
      //   dirichletVertices[indexSet.index(vertex)] = dirichletIndicatorFunction(vertex.geometry().corner(0));
      //   // if(dirichletIndicatorFunction(vertex.geometry().corner(0)))
      //   // {
      //   //   std::cout << "Dirichlet Vertex with coordinates:" << vertex.geometry().corner(0) << std::endl;
      //   // }
      // }

      // BoundaryPatch<GridView> dirichletBoundary(gridView, dirichletVertices);
      // BitSetVector<Coefficient::TangentVector::dimension> dirichletNodes(tangentBasis.size(), false);
      // constructBoundaryDofs(dirichletBoundary, tangentBasis, dirichletNodes);

      ///////////////////////////////////////////
      //   Get initial Iterate
      ///////////////////////////////////////////
      auto pythonInitialIterate = Python::makeDifferentiableFunction<FieldVector<double,3>(FieldVector<double,2>)>(pyModule.get("f"), pyModule.get("df"));

      VectorSpaceCoefficients x(deformationBasis.size());
      interpolate(deformationBasis, x, pythonInitialIterate); 






      /**
       * @brief We need to setup LocalDiscreteKirchhoffBendingIsometry with a coefficient 
       * vector of ctype 'adouble' while the solver gets a coefficient vector 
       * of ctype 'double'.
       */
      IsometryCoefficients isometryCoefficients(coefficientBasis.size());
      AIsometryCoefficients isometryCoefficients_adouble(coefficientBasis.size());


      /**
       * @brief Copy the current iterate into a data type that encapsulates the isometry constraint
       *        i. e. convert coefficient data structure from 'VectorSpaceCoefficients' to 'IsometryCoefficients'
       */
      vectorToIsometryCoefficientMap(deformationBasis,coefficientBasis,x,isometryCoefficients);
      vectorToIsometryCoefficientMap(deformationBasis,coefficientBasis,x,isometryCoefficients_adouble);

      /**
       * @brief TEST: conversion of coefficient vectors (back & forth)
       */
      coefficientConversionTest<VectorSpaceCoefficients, DeformationBasis, CoefficientBasis, IsometryCoefficients>(x, deformationBasis, coefficientBasis);


        

      using LocalDKFunction = GFE::LocalDiscreteKirchhoffBendingIsometry<DeformationBasis, CoefficientBasis, AIsometryCoefficients>;
      LocalDKFunction localDKFunction(deformationBasis, coefficientBasis, isometryCoefficients_adouble);


      /**
       * @brief TEST: check isometry condition on the nodes of the grid for initial deformation.
       */
        // auto initialDeformationFunction = Functions::makeDiscreteGlobalBasisFunction<FieldVector<double,3>>(deformationBasis, x);
        // nodewiseIsometryTest(initialDeformationFunction);
      nodewiseIsometryTest(localDKFunction, gridView);


      /**
       * @brief Read force term.
       */
      auto pythonForce = Python::make_function<FieldVector<double,3>>(pyModule.get("force"));
      auto forceGVF  = Dune::Functions::makeGridViewFunction(pythonForce, gridView);
      auto localForce = localFunction(forceGVF);


      /**
       * @brief Read effective prestrain Tensor
       */
      // Dune::FieldVector<adouble,3> Beffvec = parameterSet.get<Dune::FieldVector<adouble,3>>("effectivePrestrain", {0.0, 0.0, 0.0});
      // Dune::FieldMatrix<adouble,2,2> effectivePrestrain = {{(adouble)Beffvec[0], (adouble)Beffvec[2]}, {(adouble)Beffvec[2],(adouble)Beffvec[1]} };
      // printmatrix(std::cout, effectivePrestrain, "effective prestrain (Beff): ", "--");


      /**
       * @brief Get effective quadratic form Qhom
       * 
       *  input-vector: [q_1,q_2,q_3,q_12,q_13,q_23]
       *  is assembled into a matrix where the off-diagonal entries are divided by 2 (compare with definition in the paper)
       *            (   q_1    , 0.5*q_12  ,  0.5*q_13 )
       *     Q =    ( 0.5*q_12 ,   q_2     ,  0.5*q_23 )
       *            ( 0.5*q_13 , 0.5*q_23  ,    q_3    )
       */
      // Dune::FieldVector<adouble,6> Qhomvec = parameterSet.get<Dune::FieldVector<adouble,6>>("effectiveQuadraticForm", {1.0, 1.0, 1.0, 0.0, 0.0, 0.0});
      // Dune::FieldMatrix<adouble,3,3> Qhom = {{(adouble)Qhomvec[0], (adouble)0.5*Qhomvec[3], (adouble)0.5*Qhomvec[4]}, 
      //                                       {(adouble)0.5*Qhomvec[3], (adouble)Qhomvec[1], (adouble)0.5*Qhomvec[5]},
      //                                       {(adouble)0.5*Qhomvec[4], (adouble)0.5*Qhomvec[5], (adouble)Qhomvec[2]}};
      // printmatrix(std::cout, Qhom, "effective quadratic form (Qhom): ", "--");

      ////////////////////////////////////////////////////////////////////////
      //   Create an assembler for the Discrete Kirchhoff Energy Functional  (using ADOL-C)
      ////////////////////////////////////////////////////////////////////////

      /**
       * @brief Setup nonconforming energy
       */
      // auto localEnergy_nonconforming = std::make_shared<GFE::DiscreteKirchhoffBendingEnergy<CoefficientBasis, LocalDKFunction, decltype(localForce), ACoefficient>>(localDKFunction, localForce);
      /**
       * @brief Setup conforming energy (WIP)
       */
      // auto localEnergy_conforming = std::make_shared<GFE::DiscreteKirchhoffBendingEnergyZienkiewicz<CoefficientBasis, LocalDKFunction, decltype(localForce), ACoefficient>>(localDKFunction, localForce);
      // auto localEnergy_conforming = std::make_shared<GFE::DiscreteKirchhoffBendingEnergyZienkiewiczProjected<CoefficientBasis, LocalDKFunction, decltype(localForce), ACoefficient>>(localDKFunction, localForce);
      /**
       * @brief Setup energy featuring prestrain
       */
      //  auto localEnergy_prestrain = std::make_shared<GFE::DiscreteKirchhoffBendingEnergyPrestrained<CoefficientBasis, LocalDKFunction, decltype(localForce), ACoefficient>>(localDKFunction, localForce, Qhom, effectivePrestrain,  parameterSet);
      auto localEnergy_prestrain = std::make_shared<GFE::DiscreteKirchhoffBendingEnergyPrestrained<CoefficientBasis, LocalDKFunction, decltype(localForce), ACoefficient>>(localDKFunction, localForce, parameterSet, pyModule);

      // LocalGeodesicFEADOLCStiffness<CoefficientBasis, Coefficient> localGFEADOLCStiffness_conforming(localEnergy_conforming.get());
      // LocalGeodesicFEADOLCStiffness<CoefficientBasis, Coefficient> localGFEADOLCStiffness_nonconforming(localEnergy_nonconforming.get());
      LocalGeodesicFEADOLCStiffness<CoefficientBasis, Coefficient> localGFEADOLCStiffness_prestrain(localEnergy_prestrain.get());
      // GeodesicFEAssembler<CoefficientBasis, Coefficient> assembler_conforming(coefficientBasis, localGFEADOLCStiffness_conforming);
      // GeodesicFEAssembler<CoefficientBasis, Coefficient> assembler_nonconforming(coefficientBasis, localGFEADOLCStiffness_nonconforming);
      GeodesicFEAssembler<CoefficientBasis, Coefficient> assembler_prestrain(coefficientBasis, localGFEADOLCStiffness_prestrain);







      /**
       * @brief Create a solver: 
       *        - Riemannian Newton with Hessian modification 
       *        - Riemannian Trust-region 
       */
      RiemannianProximalNewtonSolver<CoefficientBasis, Coefficient> RNHMsolver;
      RiemannianTrustRegionSolver<CoefficientBasis, Coefficient> RTRsolver;


      std::string Solver_name = parameterSet.get<std::string>("Solver", "RNHM");
      double numerical_energy; //final discrete energy

      if(Solver_name == "RNHM")
      {

        RNHMsolver.setup(*grid,
              &assembler_prestrain,
              isometryCoefficients,
              dirichletNodes,
              parameterSet);

        RNHMsolver.solve();
        isometryCoefficients = RNHMsolver.getSol();
        numerical_energy = RNHMsolver.getStatistics().finalEnergy;
      } else if (Solver_name =="RiemannianTR")
      {
        std::cout << "Using Riemannian Trust-region method for energy minimization." << std::endl;
        RTRsolver.setup(*grid,
                &assembler_prestrain,
                isometryCoefficients,
                dirichletNodes,
                parameterSet);
        RTRsolver.solve();
        isometryCoefficients = RTRsolver.getSol();
        numerical_energy = RTRsolver.getStatistics().finalEnergy;
      } else 
          DUNE_THROW(Dune::Exception, "Unknown Solver type for bending isometries.");



      // Convert coefficient data structure from 'IsometryCoefficients' to 'VectorSpaceCoefficients'
      VectorSpaceCoefficients x_out(deformationBasis.size());
      isometryToVectorCoefficientMap(deformationBasis,coefficientBasis,x_out,isometryCoefficients);

      ////////////////////////////////
      //   Output result
      ////////////////////////////////
      std::string baseNameDefault = "bending-isometries-";
      std::string baseName = parameterSet.get("baseName", baseNameDefault);
      std::string resultFileName =  parameterSet.get("resultPath", "")
                                    + "/" + baseName
                                    + "_level" + std::to_string(parameterSet.get<int>("macroGridLevel"));

      if (parameterSet.get<bool>("conforming_DiscreteJacobian", 1))
        resultFileName = resultFileName + "_C";
      else 
        resultFileName = resultFileName + "_NC";


      // Create a deformation function but this time with double-types.
      using LocalDKFunctionD = GFE::LocalDiscreteKirchhoffBendingIsometry<DeformationBasis, CoefficientBasis, IsometryCoefficients>;
      LocalDKFunctionD localDKFunctionDouble(deformationBasis, coefficientBasis, isometryCoefficients);





  if (parameterSet.get<bool>("writeVTK", 1))
  {
    std::cout << "write VTK..." << std::endl;

    /**
     * @brief Compute the displacement from the deformation.
     *        interpolate the identity and substract from the coefficient vector.
     */
    VectorSpaceCoefficients identity(deformationBasis.size());
    IdentityGridEmbedding<double> identityGridEmbedding;
    interpolate(deformationBasis, identity, identityGridEmbedding);

    // auto identity_tmp = identity;

    // Compute the displacement
    auto displacement = x_out;
    displacement -= identity;

    // std::cout << "displacement.size():" << displacement.size() << std::endl;
    auto deformationFunction = Functions::makeDiscreteGlobalBasisFunction<FieldVector<double,3>>(deformationBasis, x_out);
    auto displacementFunction = Functions::makeDiscreteGlobalBasisFunction<FieldVector<double,3>>(deformationBasis, displacement);

    /**
     * @brief  We need to subsample, because VTK cannot natively display real third-order functions
     */
    int subsamplingRefinement = parameterSet.get<int>("subsamplingRefinement", 2);
    SubsamplingVTKWriter<GridView> vtkWriter(gridView, Dune::refinementLevels(subsamplingRefinement));
    // SubsamplingVTKWriter<GridView> vtkWriter(gridView, Dune::refinementLevels(2));



    /**
     * @brief Basis used to represent normal vector fields
     * 
     */
    auto normalBasis = makeBasis(gridView,
                      power<3>(
                      lagrange<1>(),
                      flatLexicographic()));
    // auto normalBasis = makeBasis(gridView,
    //                               power<3>(
    //                               lagrange<1>(),
    //                               blockedInterleaved()));


    // TEST Compute 


    // auto normalLambda = [deformationFunction](const FieldVector<double,2>& x)
    // {
    //   // deformationfunction.derivative() ... //needs binding?!
    

    // }

    


    /**
     * @brief Interpolate and VTK-write analytical solution "u" from ParamterFile.
     */
    if (parameterSet.get<bool>("vtkwrite_analyticalSolution", 0))
    {
      // auto pythonAnalyticalSolution = Python::makeDifferentiableFunction<FieldVector<double,3>(FieldVector<double,2>)>(pyModule.get("u"), pyModule.get("du"));
      auto pythonAnalyticalSolution = Python::makeDifferentiableFunction<FieldVector<double,3>(FieldVector<double,2>)>(pyModule.get("displacement"), pyModule.get("displacementGradient"));


      // deprecated: interpolate ...
      // VectorSpaceCoefficients y(deformationBasis.size());
      // interpolate(deformationBasis, y, pythonAnalyticalSolution);
      // // Compute the displacement
      // auto displacementAnalytical = y;
      // // displacementAnalytical -= identity_tmp;
      // displacementAnalytical -= identity;


      // auto pythonIdentity =   Python::make_function<FieldVector<double,3>>(IdentityGridEmbedding<double>::operator())


      // auto displacementFunctionAnalytical = Functions::makeDiscreteGlobalBasisFunction<FieldVector<double,3>>(deformationBasis, displacementAnalytical);
      // auto deformationFunctionAnalytical = Functions::makeDiscreteGlobalBasisFunction<FieldVector<double,3>>(deformationBasis, y);
      // vtkWriter.addVertexData(displacementFunctionAnalytical, VTK::FieldInfo("displacement_analytical", VTK::FieldInfo::Type::vector, 3));
      // vtkWriter.addVertexData(deformationFunctionAnalytical, VTK::FieldInfo("deformation_analytical", VTK::FieldInfo::Type::vector, 3));
    
      vtkWriter.addVertexData((pythonAnalyticalSolution), VTK::FieldInfo("displacement_analytical", VTK::FieldInfo::Type::vector, 3));
      /**
       * @brief Get the normal vector field of the surface parametrized 
       *        by the analytical solution.
       *        - We represent the normal vector in a first order Lagrange-Power basis ('normalBasis').
       */
      if (parameterSet.get<bool>("vtkWrite_analyticalSurfaceNormal", 0))
      {
      // Get the surface normal function.
      auto pythonSurfaceNormal = Python::make_function<FieldVector<double,3>>(pyModule.get("surfaceNormal"));


      // deprecated: interpolate ...
      // std::vector<FieldVector<double,3>> normalVectorCoefficients(normalBasis.size());
      // Dune::Functions::interpolate(normalBasis, normalVectorCoefficients, pythonSurfaceNormal); 
      // auto surfaceNormalAnalytical = Functions::makeDiscreteGlobalBasisFunction<FieldVector<double,3>>(normalBasis, normalVectorCoefficients);
      // vtkWriter.addVertexData(surfaceNormalAnalytical, VTK::FieldInfo("surfaceNormal_analytical", VTK::FieldInfo::Type::vector, 3));

      vtkWriter.addVertexData(pythonSurfaceNormal, VTK::FieldInfo("surfaceNormal_analytical", VTK::FieldInfo::Type::vector, 3));
      }
    }
    //-------------------------------------------------------------------------------------    TODO: OUTSOURCE
    /**
     * @brief TEST: Compute the discrete normal vector of the surface parametrized 
     *              by the discrete solution.
     */
     auto surfaceNormalDiscreteCoefficients = computeDiscreteSurfaceNormal(localDKFunctionDouble, normalBasis);
    //-------------------------------------------------------------------------------------

    // Create DiscreteGlobalBasisFunctions.
    auto surfaceNormalDiscrete   = Functions::makeDiscreteGlobalBasisFunction<FieldVector<double,3>>(normalBasis, surfaceNormalDiscreteCoefficients);
    vtkWriter.addVertexData(displacementFunction, VTK::FieldInfo("displacement", VTK::FieldInfo::Type::vector, 3));
    vtkWriter.addVertexData(surfaceNormalDiscrete , VTK::FieldInfo("surfaceNormal_discrete", VTK::FieldInfo::Type::vector, 3));
    vtkWriter.write(resultFileName);
  }


        /**
         * @brief Measure errors.   
         *        1.  Compute the Isometry-Error
         *        2.  Compute the L2-Error and H1-SemiError as well as the energy difference 
         *            between the discrete deformation u_h and the analytical deformation u .
         */
        if (parameterSet.get<bool>("measure_isometryError", 0))
        {
            auto isometry_errors = isometryError(localDKFunctionDouble, gridView);
            if(isometry_errors[1] > Isometry_threshold[experimentNumber-1])  // Isometry-L2-error
            {
                std::cerr << std::setprecision(9);
                std::cerr << "Isometry-L2-error is to large! " << std::endl;
                std::cerr << "Isometry-L2-error is to large: "<< isometry_errors[1] << " but threshold is " << Isometry_threshold[experimentNumber-1] << std::endl;
                return 1;
            }

        }
            
        if (parameterSet.get<bool>("measure_analyticalError", 0))
        {
            auto discretization_errors = measureAnalyticalError(localDKFunctionDouble,gridView,pyModule);


            /**
             * @brief Check Quantities
             * 
             */
            // L2 - Discretization error
            if(discretization_errors[0] > L2_threshold[experimentNumber-1])
            {
                std::cerr << std::setprecision(9);
                std::cerr << "L2-error is to large: "<<discretization_errors[0] << " but threshold is " << L2_threshold[experimentNumber-1] << std::endl;
                return 1;
            }
            // H1(semi) - Discretization error
            if(discretization_errors[1] > H1_threshold[experimentNumber-1])
            {
                std::cerr << std::setprecision(9);
                std::cerr << "H1-error is to large: "<< discretization_errors[1] << " but threshold is " << H1_threshold[experimentNumber-1] << std::endl;
                return 1;
            }
        }
            


        /**
         * @brief Write Energy values.
         * 
         */
        // auto numerical_energy = solver.getStatistics().finalEnergy;
        std::cout << "(Final) Energy of discrete solution: " << numerical_energy<< std::endl;
        if (parameterSet.get<bool>("compare_EnergyValues", 0))
        {
          auto analytical_energy = parameterSet.get<double>("analytical_energy");
          auto energy_difference = analytical_energy - numerical_energy;
          std::cout << "Analytical energy: " << analytical_energy<< std::endl;  
          std::cout << "Energy difference: " << energy_difference << std::endl;

          if(energy_difference > Energy_threshold[experimentNumber-1] )
          {
            std::cerr << std::setprecision(9);
            std::cerr << "Energy difference is to large:" << energy_difference  << " but threshold is " << Energy_threshold[experimentNumber-1] << std::endl;
            return 1;
          }
        }

      // // Write the corresponding coefficient vector: verbatim in binary, to be completely lossless
      //   char iFilename[200];
      //   sprintf(iFilename, (resultFileName + ".data").c_str());

      //   FILE* fpIterate = fopen(iFilename, "wb");
      //   if (!fpIterate)
      //       DUNE_THROW(SolverError, "Couldn't open file " << iFilename << " for writing");

      //   for (size_t j=0; j<isometryCoefficients.size(); j++)
      //       fwrite(&isometryCoefficients[j], sizeof(Coefficient), 1, fpIterate);

      //   fclose(fpIterate);


      #if 0
        // Write the corresponding coefficient vector: verbatim in binary, to be completely lossless
        typedef BlockVector<typename Coefficient::CoordinateType> EmbeddedVectorType;
        EmbeddedVectorType xEmbedded(isometryCoefficients.size());
        for (size_t i = 0; i<isometryCoefficients.size(); i++)
        {
          xEmbedded[i] = isometryCoefficients[i].globalCoordinates();
          std::cout << "isometryCoefficients[i].globalCoordinates():" << isometryCoefficients[i].globalCoordinates() << std::endl;

        }
        std::ofstream outFile(resultFileName + ".data", std::ios_base::binary);
        MatrixVector::Generic::writeBinary(outFile, xEmbedded);
        outFile.close();
        ///////////////////////////////////////////
        //  (Option) write Solution(DOF)-vector and vertex coordinates to .txt-File
        ///////////////////////////////////////////
        if (parameterSet.get<bool>("writeDOFvector", 0))
          writeDOFVector(gridView,targetDim, x, "DOFVectorR" + std::to_string(dim) + "_R" + std::to_string(targetDim) + ".txt");
      #endif



      std::cout << "TEST - " << experiment << "passed." <<  std::endl;
      std::cout << "Total time elapsed: " << globalTimer.elapsed() << std::endl;
  }

  std::cout << "Properties-shape-test passed." <<  std::endl;
  return 0;
  }