#include <config.h>
#include <array>
#include <vector>
#include <fstream>

#include <iostream>
#include <dune/common/indices.hh>
#include <dune/common/bitsetvector.hh>
#include <dune/common/parametertree.hh>
#include <dune/common/parametertreeparser.hh>
#include <dune/common/float_cmp.hh>
#include <dune/common/math.hh>

#include <dune/geometry/quadraturerules.hh>

#include <dune/grid/uggrid.hh>
#include <dune/grid/yaspgrid.hh>
#include <dune/grid/io/file/vtk/subsamplingvtkwriter.hh>

#include <dune/istl/matrix.hh>
#include <dune/istl/bcrsmatrix.hh>
#include <dune/istl/multitypeblockmatrix.hh>
#include <dune/istl/multitypeblockvector.hh>
#include <dune/istl/matrixindexset.hh>
#include <dune/istl/solvers.hh>
#include <dune/istl/spqr.hh>
#include <dune/istl/preconditioners.hh>
#include <dune/istl/io.hh>

#include <dune/functions/functionspacebases/interpolate.hh>
#include <dune/functions/backends/istlvectorbackend.hh>
#include <dune/functions/functionspacebases/powerbasis.hh>
#include <dune/functions/functionspacebases/compositebasis.hh>
#include <dune/functions/functionspacebases/lagrangebasis.hh>
#include <dune/functions/functionspacebases/periodicbasis.hh>
#include <dune/functions/functionspacebases/subspacebasis.hh>
#include <dune/functions/functionspacebases/boundarydofs.hh>
#include <dune/functions/gridfunctions/discreteglobalbasisfunction.hh>
#include <dune/functions/gridfunctions/gridviewfunction.hh>

#include <dune/common/fvector.hh>
#include <dune/common/fmatrix.hh>

#include <dune/microstructure/matrix_operations.hh>
#include <dune/microstructure/CorrectorComputer.hh>
#include <dune/microstructure/EffectiveQuantitiesComputer.hh>
#include <dune/microstructure/prestrainedMaterial.hh>

#include <dune/solvers/solvers/umfpacksolver.hh>
#include <dune/istl/eigenvalue/test/matrixinfo.hh>

#include <dune/fufem/dunepython.hh>

#include <any>
#include <variant>
#include <string>
#include <iomanip>
#include <filesystem>

using namespace Dune;
using namespace MatrixOperations;


/**
 * @brief Objective: Read a microstructure configuration (Python class) from Parameterset/Python-module
 *                   using the different material class definitions.
 */


//////////////////////////////////////////////////
//   Infrastructure for handling periodicity
//////////////////////////////////////////////////
// Check whether two points are equal on R/Z x R/Z x R
auto equivalent = [](const FieldVector<double,3>& x, const FieldVector<double,3>& y)
                  {
                    return ( (FloatCmp::eq(x[0],y[0]) or FloatCmp::eq(x[0]+1,y[0]) or FloatCmp::eq(x[0]-1,y[0]))
                             and (FloatCmp::eq(x[1],y[1]) or FloatCmp::eq(x[1]+1,y[1]) or FloatCmp::eq(x[1]-1,y[1]))
                             and (FloatCmp::eq(x[2],y[2]))
                             );
                  };


int main(int argc, char *argv[])
{
  MPIHelper::instance(argc, argv);

  Dune::Timer globalTimer;
  //--- setup Log-File 
  std::fstream log;

  std::cout << "Current path is " << std::filesystem::current_path() << '\n';
  std::filesystem::path file_path = (__FILE__);
  std::cout<< "File path: " << file_path<<std::endl;
  std::cout << "dir_path: " << file_path.parent_path() << std::endl;
  std::string dir_path  = file_path.parent_path();

  // if (argc < 3)
  //     DUNE_THROW(Exception, "Usage: ./Cell-Problem <python path> <python module without extension>");

  // Start Python interpreter
  // Python::start();
  // auto pyMain = Python::main();
  // pyMain.runStream()
  //     << std::endl << "import math"
  //     << std::endl << "import sys"
  //     << std::endl << "sys.path.append('" << argv[1] << "')"  << std::endl;
  // auto pyModule = pyMain.import(argv[2]);


  // Relative path/Test version:
  Python::start();
  auto pyMain = Python::main();
  pyMain.runStream()
      << std::endl << "import math"
      << std::endl << "import sys"
      << std::endl << "sys.path.append('" << dir_path << "')"  << std::endl;
  auto pyModule = pyMain.import("readmicrostructuretest");

  ParameterTree parameterSet;
  pyModule.get("parameterSet").toC(parameterSet);
  // read possible further parameters from the command line
  ParameterTreeParser::readOptions(argc, argv, parameterSet);
  // Print all parameters, to make them appear in the log file
  std::cout << "Input parameters:" << std::endl;
  parameterSet.report();



  constexpr int dim = 3;
  constexpr int dimWorld = 3;
  ///////////////////////////////////
  // Generate the grid
  ///////////////////////////////////
  // --- Corrector Problem Domain (-1/2,1/2)^3:
  FieldVector<double,dim> lower({-1.0/2.0, -1.0/2.0, -1.0/2.0});
  FieldVector<double,dim> upper({1.0/2.0, 1.0/2.0, 1.0/2.0});

  int gridLevel = 3;

  std::array<int, dim> nElements = {(int)std::pow(2,gridLevel) ,(int)std::pow(2,gridLevel) ,(int)std::pow(2,gridLevel)};
  std::cout << "Number of Grid-Elements in each direction: " << nElements << std::endl;

  using CellGridType = YaspGrid<dim, EquidistantOffsetCoordinates<double, dim> >;
  CellGridType grid_CE(lower,upper,nElements);
  using GridView = CellGridType::LeafGridView;
  const GridView gridView_CE = grid_CE.leafGridView();

  //--- Choose a finite element space for Cell Problem
  using namespace Functions::BasisFactory;
  Functions::BasisFactory::Experimental::PeriodicIndexSet periodicIndices;

  //--- get PeriodicIndices for periodicBasis (Don't do the following in real life: It has quadratic run-time in the number of vertices.)
  for (const auto& v1 : vertices(gridView_CE))
    for (const auto& v2 : vertices(gridView_CE))
      if (equivalent(v1.geometry().corner(0), v2.geometry().corner(0)))
      {
        periodicIndices.unifyIndexPair({gridView_CE.indexSet().index(v1)}, {gridView_CE.indexSet().index(v2)});
      }

  //--- setup first order periodic Lagrange-Basis
  auto Basis_CE = makeBasis(
    gridView_CE,
    power<dim>(
      Functions::BasisFactory::Experimental::periodic(lagrange<1>(), periodicIndices),
      flatLexicographic()
      ));


  // -------------------------------------------------------
  /**
   * @brief This needs to be declared as Callable otherwise Class(param) wont work
   *        if _call__(self,x) is not implemented
   */
  Python::Callable MicrostructureClass  = pyModule.get("Microstructure");

  /**
   * @brief Create instance of class
   *        This needs to be a 'Python::Reference' and not 'Python::Callable' 
   *        since Callable only works if __call__(self,x) is implemented in the python module!!!
   * 
   */
  Python::Reference microstructure = MicrostructureClass(); //Setup a constant microstructure

  //Get (double) class member variable
  double gamma;
  microstructure.get("gamma").toC<double>(gamma);
  std::cout << "microstructure gamma:" << gamma << std::endl;

  //Get (int) class member variable
  int phases;
  microstructure.get("phases").toC<int>(phases);
  std::cout << "microstructure phases:" << phases << std::endl;


  //Get (std::string) class member variable
  std::string phaseType;
  // isotropic material setup
  int phase = 1;
  microstructure.get("phase" + std::to_string(phase) + "_type").toC<std::string>(phaseType);
  std::cout << "microstructure phaseType:" << phaseType << std::endl;

  Dune::FieldVector<double,2> materialParameters(0);
  microstructure.get("materialParameters_phase" + std::to_string(phase)).toC<Dune::FieldVector<double,2>>(materialParameters);
  std::cout << "Lame-Parameters (mu,lambda): (" << materialParameters[0] << "," << materialParameters[1] << ") " << std::endl;


  // orthotropic material setup
  Dune::FieldVector<double,9> materialParameters_orthotropic(0);
  phase = 2;

  microstructure.get("phase" + std::to_string(phase) + "_type").toC<std::string>(phaseType);
  std::cout << "microstructure phaseType:" << phaseType << std::endl;

  int axis;
  double angle;
  microstructure.get("phase" + std::to_string(phase) + "_axis").toC<int>(axis);
  microstructure.get("phase" + std::to_string(phase) + "_angle").toC<double>(angle);

  std::cout << "orthotropic material with material rotation axis: " << axis << std::endl;
  std::cout << "orthotropic material with material rotation angle: " << angle << std::endl;

  

  microstructure.get("materialParameters_phase" + std::to_string(phase)).toC<Dune::FieldVector<double,9>>(materialParameters_orthotropic);
  printvector(std::cout, materialParameters_orthotropic, "materialParameters_orthotropic: ", "--");


  // anisotropic material setup
  Dune::FieldMatrix<double,6,6> materialParameters_anisotropic(0); 
  phase = 4;
  microstructure.get("phase" + std::to_string(phase) + "_type").toC<std::string>(phaseType);
  std::cout << "microstructure phaseType:" << phaseType << std::endl;
  microstructure.get("materialParameters_phase" + std::to_string(phase)).toC<Dune::FieldMatrix<double,6,6>>(materialParameters_anisotropic);
  printmatrix(std::cout, materialParameters_anisotropic, "materialParameters_anisotropic: ", "--");



  //Get prestrain of phase 
  auto prestrain1 = Python::make_function<Dune::FieldMatrix<double,3,3>>(Python::Callable(microstructure.get("prestrain_phase" + std::to_string(phase))));
  auto localPrestrain1 = localFunction(Dune::Functions::makeGridViewFunction(prestrain1, gridView_CE));


  //Class Member function.
  auto indicatorFunction = Python::make_function<int>(Python::Callable(microstructure.get("indicatorFunction")));
  // auto indicatorFunction = Python::make_function<int>(c); // Kann auch direkt class instance benutzenn denn diese hat operator() (~ __call__(..))
  auto indicatorFunctionGVF = Dune::Functions::makeGridViewFunction(indicatorFunction, gridView_CE);
  auto localIndicatorFunction = localFunction(indicatorFunctionGVF);


  // Test evaluate the localFunctions on quadrature Points
  for (const auto& element : elements(gridView_CE))
  {
    localIndicatorFunction.bind(element);
    localPrestrain1.bind(element);

    const auto& quadRule = Dune::QuadratureRules<double,dim>::rule(element.type(), 2);
    for (const auto& quadPoint : quadRule)
    {
      auto evaluateIndicator = localIndicatorFunction(quadPoint.position());
      auto evaluatePrestrain = localPrestrain1(quadPoint.position());
      // std::cout << "localIndicatorFunction(quadPoint.position()): " << localIndicatorFunction(quadPoint.position()) << std::endl;
      // printmatrix(std::cout, localPrestrain1(quadPoint.position()), "localPrestrain1(quadPoint.position()): ", "--");
    }

  }


  using MaterialType = prestrainedMaterial<GridView>;
  std::shared_ptr<MaterialType> material = std::make_shared<MaterialType>(gridView_CE,microstructure,parameterSet,pyModule);

  //Setup a material object
  // auto material = prestrainedMaterial(gridView_CE,microstructure,parameterSet,pyModule);

  //Setup a corrector object & try to assemble/solve
  // auto correctorComputer = CorrectorComputer(Basis_CE, material, log, parameterSet);
  // correctorComputer.assemble();
  // correctorComputer.solve();
  std::shared_ptr<CorrectorComputer<decltype(Basis_CE),MaterialType> > correctorComputer = std::make_shared<CorrectorComputer<decltype(Basis_CE),MaterialType>>(Basis_CE, material, parameterSet);
  correctorComputer->assemble();
  correctorComputer->solve();


  //Setup EffectiveQuantitiesComputer object & get Quantities.
  // auto effectiveQuantitiesComputer = EffectiveQuantitiesComputer(correctorComputer,material);
  auto effectiveQuantitiesComputer = EffectiveQuantitiesComputer(correctorComputer);
  effectiveQuantitiesComputer.computeEffectiveQuantities();
  auto Qeff = effectiveQuantitiesComputer.getQeff();
  auto Beff = effectiveQuantitiesComputer.getBeff();
  printmatrix(std::cout, Qeff, "Matrix Qeff", "--");
  printvector(std::cout, Beff, "Beff", "--");


  std::cout << "Total time elapsed: " << globalTimer.elapsed() << std::endl;
  std::cout << "readmicrostructuretest passed." << std::endl;
  return 0;
}
