import os
import sys
import re
import codecs
import numpy as np
import matplotlib as mpl
import matplotlib.pyplot as plt
import matplotlib.colors as colors
from matplotlib import cm
from matplotlib.colors import ListedColormap, LinearSegmentedColormap

from scipy.optimize import minimize_scalar 
import importlib
#---------------------------------------------------------------------------

def xytokappaalpha(x,y):
    """
        Convert 2-dimensional cartesian coordinates [x,y]
        to [curvature,angle]
    """
    if y>0:
        return [np.sqrt(x**2+y**2), np.abs(np.arctan2(y,x))]
    else:
        return [-np.sqrt(x**2+y**2), np.abs(np.arctan2(y,x))]


def ReadEffectiveQuantities(QFilePath, BFilePath):
    """"
        Read Output Matrices (effective quantities) from file-paths.
        Micro-Problem .txt-output Files : 
        (default): '../outputs/Qmatrix.txt' , '../outputs/Bmatrix.txt'

        Note: These correspond to the coefficients of Qhom,Beff
        so Qhom is a 3x3-matrix while Beff is a Vector os size 3
    """
    # -- Read Matrix Qhom
    X = []
    with codecs.open(QFilePath, encoding='utf-8-sig') as f:
        for line in f:
            s = line.split()
            X.append([float(s[i]) for i in range(len(s))])
    Qhom = np.array([[X[0][2], X[1][2], X[2][2]],
                  [X[3][2], X[4][2], X[5][2]],
                  [X[6][2], X[7][2], X[8][2]] ])

    # -- Read Beff (as Vector)
    X = []
    with codecs.open(BFilePath, encoding='utf-8-sig') as f:
        for line in f:
            s = line.split()
            X.append([float(s[i]) for i in range(len(s))])
    Beff = np.array([X[0][2], X[1][2], X[2][2]])   # Beff is written as Coefficients to .txt-file
    return Qhom, Beff




def SetParameterMaterialFunction(parameterFile, parameterName, parameterValue):
    """"
        Overwrite parameter 'parameterName' in the 'parameterFile'
        with new value: 'parameterValue.
    """
    with open(parameterFile +'.py', 'r') as file:
        filedata = file.read()
        filedata = re.sub('(?m)'+str(parameterName)+'\s?=.*',str(parameterName)+' = '+str(parameterValue),filedata)
        f = open(parameterFile +'.py','w')
        f.write(filedata)
        f.close()




def get_local_minimizer_on_axes(Q,B):
    """
        Routine that computes minimizers along the axis (alpha=0 , pi/2.0).
        Along the axis the energy is a 1-D function in kappa.

        *This relies on the sciPy Optimization package.
    """
    invoke_function=lambda kappa: energy(kappa,0,Q,B)
    result_0 = minimize_scalar(invoke_function, method="golden")
    invoke_function=lambda kappa: energy(kappa,np.pi/2,Q,B)
    result_90 = minimize_scalar(invoke_function, method="golden")
    return np.array([[result_0.x,0],[result_90.x,np.pi/2]])



#-------- Energy evaluations & Plots:

def energy(kappa,alpha,Q,B)  :
    """
        Energy evaluation for given effective quantities Q and B 
        with input:
        * curvature: kappa 
        * angle: alpha
    """
    # Compute cylindrical minimizer coefficients from angle & curvature
    G = kappa*np.array([np.cos(alpha)**2, np.sin(alpha)**2, np.sqrt(2)*np.cos(alpha)*np.sin(alpha)])
    # Subtract the effective prestrain
    G = G - B;
    # Compute the energy
    energy =  (np.transpose(G)).dot((Q.dot(G)));
    return energy


def createEnergyValues(r,theta,Q,B):
    """
        Compute energy values for a given meshgrid of curvature(r) and angle(theta) values
        associated to the effective quantities Q and B.

        the curvature array 'r' only features positive values. However the energy is also 
        evaluated for the corresponding negative values.
    """
    # Get number of sample points.
    N = r.shape[0]

    E=np.zeros(np.shape(r))
    for k in range(0,N): 
        for l in range(0,N):     
            if theta[k,l]<np.pi:
                E[k,l]=energy(r[k,l],theta[k,l],Q,B)
            else:
                E[k,l]=energy(-r[k,l],theta[k,l],Q,B)

    return E




def getMinimizers(r,theta,E):
    """"
        Compute global minimizer (by Sampling method)
        as well as the minimizer on the opposite sign - curvature scale

        E: meshgrid evaluation of energy.
    """
    kappa=0
    kappa_pos=0
    kappa_neg=0


    N = E.shape[0]
    # Compute global Minimizer 
    [kmin,lmin]=np.unravel_index(E.argmin(),(N,N))
    kappamin=r[kmin,lmin]
    alphamin=theta[kmin,lmin]
    Emin = E[kmin,lmin]

    # Get minimizer in Positive curvature region
    N_mid=int(N/2)
    [kmin,lmin]=np.unravel_index(E[:N_mid,:].argmin(),(N_mid,N))
    kappamin_pos=r[kmin,lmin]
    alphamin_pos=theta[kmin,lmin]
    Emin_pos=E[kmin,lmin]
    # Get minimizer in Negative curvature region
    [kmin,lmin]=np.unravel_index(E[N_mid:,:].argmin(),(N_mid,N))
    kappamin_neg=r[kmin+N_mid,lmin]
    alphamin_neg=theta[kmin+N_mid,lmin]
    Emin_neg=E[kmin+N_mid,lmin]

    print('Emin:', Emin)
    print('Emin_pos:', Emin_pos)
    print('Emin_neg:', Emin_neg)
    print('E.min(): ',E.min())
    print('E.max(): ',E.max())

    minimizer_sign = 'positive'

    if(abs(Emin_pos - Emin_neg)<1e-5):
        print('Both minimizers have nearly the same energy - likely two minimizers.')
        energyDifference = 0
    elif(Emin_pos < Emin_neg):
        print('Global Minimizer has positive curvature.')
        energyDifference = Emin_neg-Emin_pos
    else:
        print('Global Minimizer has negative curvature.')
        minimizer_sign = 'negative'
        energyDifference = Emin_pos-Emin_neg


    if (minimizer_sign == 'positive'):
        alpha_globalmin = alphamin
        kappa_globalmin = kappamin
        kappa_localmin = (-1)*kappamin_neg 
        alpha_localmin = alphamin_neg
    else :
        alpha_globalmin = alphamin
        kappa_globalmin = (-1)*kappamin
        kappa_localmin = kappamin_pos
        alpha_localmin = alphamin_pos

    print('kappa_globalmin: ', kappa_globalmin)
    print('alpha_globalmin: ', alpha_globalmin)
    print('kappa_localmin: ',  kappa_localmin)
    print('alpha_localmin: ',  alpha_localmin)
    # print('energy (global minimizer)', energy(kappa_globalmin,alpha_globalmin,Q,B))
    # print('energy (local minimizer)', energy(kappa_localmin,alpha_localmin,Q,B))
    # print('energy in origin (0,0):' ,energy(0,0,Q,B))

    return [kappa_globalmin,alpha_globalmin,kappa_localmin,alpha_localmin,energyDifference,minimizer_sign];






def createEnergyLandscapeFigure(r,theta,E,resultPath , parameterName='',parameterValue='',title_string='',normalize=True, displayMinimizer=True, plotContourLines=True):
    """
        Create an energy landscape plot.
    """
    print('Create energy landscape ...')



    # Setup Plot style.
    plt.style.use("seaborn")
    plt.style.use('energylandscape.mplstyle')
    # Set correct Figure size (width/height) to make the figure fit the text without scaling.
    textwidth = 6.26894      # textwidth in inch
    width = textwidth * 0.5
    height = textwidth/1.618 # The golden ratio.

    #Get Minimzers from sampled energy values.
    # [kappa_globalmin,alpha_globalmin,kappa_localmin,alpha_localmin,energyDifference] = getMinimizers(r,theta,E,Q,B) 
    [kappa_globalmin,alpha_globalmin,kappa_localmin,alpha_localmin,energyDifference,minimizer_sign] = getMinimizers(r,theta,E) 

    #Get maximal curvature parameter used for sampling
    curvatureLength = r[0][-1]

    # Normalize minimum energy to zero:
    if normalize:
        E=E-E.min()
        print('E.min(): (after normalization): ',E.min())
        print('E.max(): (after normalization): ',E.max())


    # Create a figure & axes object
    fig, ax1 = plt.subplots(ncols=1,nrows=1, figsize=(width,height), subplot_kw=dict(polar=True),gridspec_kw={'hspace': 0.4}, zorder=1)
     # Adjust placement inside figure
    fig.subplots_adjust(left=.1, bottom=0.01, right=.9, top=0.9) 

    """
        Define new custom colormap (Test)
        define top and bottom colormaps 
    """
    # top = cm.get_cmap('Oranges_r', 128) # r means reversed version
    # bottom = cm.get_cmap('Blues', 128)# combine it all
    # newcolors = np.vstack((top(np.linspace(0, 1, 128)),
    #                     bottom(np.linspace(0, 1, 128))))
    # orange_blue = colors.ListedColormap(newcolors, name='OrangeBlue')
    # bi_pride = LinearSegmentedColormap.from_list("", ["#D60270", "#9B4F96", "#0038A8"])
    # testColor = LinearSegmentedColormap.from_list("", ["#380282","#0343DF","#FF7F50",  "#FE420F"])

    """
        Plot with 'contourf':
        Its important to hide the Contour LineStroke (remove white lines between contour levels) using set_edgecolor("face")
    """
    # levs=np.geomspace(0.01,E.max(),300)  # Return numbers spaced evenly on a log scale (a geometric progression).
    # pcm=ax1.contourf(theta, r, E, levs, norm=colors.PowerNorm(gamma=0.5,vmin=E.min(),vmax=E.max()), cmap='brg', zorder=0)
    # for c in pcm.collections:
    #   c.set_edgecolor("face")

    """
        Create plot with 'pcolormesh'
        * Reducing the image-quality/file-size by using rasterized option:
        * Looks much smoother without 'rasterize' option however file-size gets out of hand..: 
        * Together with the 'dpi=..' option of figsave we can control the file-size
        * Advantage: Text etc. is still vectorized without having too large files.

        Problem: How to get the Colorbarticks  in a uniform spacing when using PowerNorm (unclear)
    """
    pcm=ax1.pcolormesh(theta, r, E, norm=colors.PowerNorm(gamma=0.4,vmin=E.min(),vmax=E.max()) ,cmap='plasma', edgecolors='face', zorder=0, rasterized=True, alpha=0.8) 



    print('0.025*E.max():',0.025*E.max())
    print('energyDifference: ', energyDifference)

    """"
        Create additional contour-line plot in the range between
        zero and the energyDifference between both minimizers.
    """
    if plotContourLines:
        contour_levs = np.linspace(0.0001,energyDifference*1.0,6) # contour only from zero (Minimizer) to energyDifference between global/local Minimizer.
        contour_levs = np.linspace(0.0001,1,3) # contour only from zero (Minimizer) to energyDifference between global/local Minimizer.
        # contour_levs = np.linspace(0.0001,energyDifference*1.0,2) # contour only from zero (Minimizer) to energyDifference between global/local Minimizer.
        print('contour_levs:', contour_levs)
        contour_levs = np.sort(contour_levs, axis=None)  
        # contour_levs = np.linspace(0.0001,energyDifference*1.0,1) # contour only from zero (Minimizer) to energyDifference between global/local Minimizer.
        print('contour_levs (sorted):', contour_levs)
        contours = plt.contour(theta, r, E, contour_levs, norm=colors.PowerNorm(gamma=0.4,vmin=E.min(),vmax=E.max()), colors='ivory',linewidths=0.65, linestyles="dashed", alpha=0.8, zorder=2)
        contours_min = plt.contour(theta, r, E, [contour_levs[1]], norm=colors.PowerNorm(gamma=0.4,vmin=E.min(),vmax=E.max()), colors='yellow',linewidths=0.65, linestyles="solid", alpha=0.9, zorder=2)

    # Set Title.
    ax1.set_title(title_string,pad=6)
    # ax1.set_title(r"ratio $r=0.12$",pad=6)
    # ax1.set_title(r""+parameterName + ": " + str(parameterValue),pad=20)


    # Create a string of the minimal (angle,curvature) values used to label the global minimizer.
    curvatureSignString = ''
    if minimizer_sign == 'negative':
        curvatureSignString = '-'

    Minimizer_string = ': (' + str(np.round(np.rad2deg(alpha_globalmin),decimals=1)) + '°,' + curvatureSignString  + str(np.round(abs(kappa_globalmin),decimals=2))+ ')'


    # Mark Global Minimizer.
    if displayMinimizer:
        gmin = ax1.plot(alpha_globalmin, abs(kappa_globalmin),
                        markerfacecolor='cornflowerblue',
                        markeredgecolor='white',            # marker edgecolor
                        marker='o',                         # each marker will be rendered as a circle
                        markersize=6,                       # marker size
                        markeredgewidth=1.0,                  # marker edge width
                        linewidth=0,
                        zorder=3,
                        alpha=1,                           # Change opacity
                        label = Minimizer_string)   
        

    plot_minimizerInOppositeCurvatureRegion = False     #Note: this is not necessarily a local minimizer!
        

    if plot_minimizerInOppositeCurvatureRegion:
            ax1.plot(alpha_localmin, abs(kappa_localmin),   # We need to take abs(kappa) since negative curvatures are still plotted with a positive kappa!
                        markerfacecolor='green',    
                        markeredgecolor='white',            # marker edgecolor
                        marker='D',                         # each marker will be rendered as a circle
                        markersize=4,                       # marker size
                        markeredgewidth=0.5,                  # marker edge width
                        linewidth=0,                      # line width
                        zorder=3,
                        alpha=1)
            

            # #--- annotate local minimizer
            # ax1.annotate( '('+ str(np.round(360-np.rad2deg(alpha_localmin),decimals=1)) + '°' + ',' + str(np.round(kappa_localmin,decimals=2))+ ')' ,
            #             xy = (alpha_localmin,abs(kappa_localmin)),
            #             # xytext=(-1.6, np.round(kappa_localmin,decimals=2)+0.9),
            #             # xytext=(-1.6, np.round(kappa_localmin,decimals=2)),
            #             xytext=(alpha_localmin,abs(kappa_localmin)),
            #             color='ivory',
            #             # zorder=5,
            #             fontsize='8')

    print('Global Minimizer (alphamin, kappamin):' , '('+ str(np.round(np.rad2deg(alpha_globalmin),decimals=1)) + ',' + str(np.round(abs(kappa_globalmin),decimals=2))+ ')' ) 
    print('Local Minimizer opposite curvature (alphamin, kappamin):' , '('+ str(np.round(360-np.rad2deg(alpha_localmin),decimals=1)) + ',' + str(np.round(abs(kappa_localmin),decimals=2))+ ')' ) 

    #### Colorbar for Contourf-Plot:
    # colorbarticks=np.geomspace(0.00001,E.max(),10)
    # cbar = plt.colorbar(pcm, ax=[ax1], extend='max', ticks=colorbarticks, pad=0.1, orientation="horizontal")

    # Colorbar for pcolormesh-Plot:
    cbar = fig.colorbar(pcm, ax=ax1, pad=0.1,  orientation="horizontal") 


    """
        Set Tick labels
        x: Angle 
        y: Radius
    
    """
    anglelabel=["0°","45°", "90°", "135°", "180°","-135°","-90°","-45°"]
    # anglelabel=["0°","45°", "90°", "135°",r"$\pm180$° ","-135°","-90°","-45°"]
    ax1.set_xticks(np.array([.0,1/4,2/4,3/4,1,5/4,6/4,7/4])*np.pi)
    ax1.set_xticklabels(anglelabel)
    # ax1.xaxis.grid(True, color='white')
    ax1.xaxis.grid(True)
    # We want to grid-axis to be on top of the contour lines to hide confusing contours on the grid-axis.
    ax1.set_axisbelow(False)
    # Adjust pad of xticklabels:
    ax1.tick_params(pad=0.5)
    # rotate radius axis labels.
    ax1.set_rlabel_position(225)
    # Adjust radial ticks.
    radiusTickLength = 0.5
    ax1.yaxis.set_major_locator(plt.FixedLocator(np.arange(radiusTickLength,curvatureLength,radiusTickLength)))
    # (Alternative used fixed values)
    # ax1.set_yticks(np.array([1,2,3]))
    # ax1.set_yticklabels(["1","2","3"], zorder=20, color="white")
    ax1.yaxis.set_tick_params(color='white', pad = 1.0)
    ax1.yaxis.grid(True)

    # Create Minor ticks manually as polar coordinates does not have the option.
    tick = [ax1.get_rmax(),ax1.get_rmax()*0.97]
    for t  in np.deg2rad(np.arange(0,360,5)):
        ax1.plot([t,t], tick, lw=0.1, color="k")

    if displayMinimizer:
        # Create a Legend.
        legend = ax1.legend(bbox_to_anchor=[0.75, 0.15],loc='lower center',borderaxespad=0.1,frameon=True,bbox_transform=fig.transFigure,handletextpad=0.1)
        frame = legend.get_frame()
        frame.set_edgecolor('black')

    # Save Figure as png.
    # fig.savefig(resultPath + '/energy_landscape_'+ parameterName + '_' +str(parameterValue) + '.png', dpi=100)
    # Save Figure as pdf.
    fig.set_size_inches(width, height)
    fig.savefig(resultPath + '/energy_landscape_'+ parameterName + '_' +str(parameterValue) + '.pdf', dpi=1500) #dpi on
    return